extensions [ gis table csv time rnd ]

;__includes [ "Example-Project-B_3Reaches-3Species/parameters-ExampleB.nls" ]
__includes [ "Example-Project-B_3Reaches-3Species/parameters-ExampleB.nls" "InSTREAM7.3_test-procedures.nls" ]

globals
[
  ;; Model run control parameters -- values set in set-parameters
  start-date                 ; Date that simulation starts, string in format: "10/1/2000"
  end-date                   ; Last date simulated, string in format:  "9/30/2005"

  ;; Observer input files and parameters -- values set in set-parameters
  GIS-file-name              ; Name of GIS file with cell polygons for all reaches
  initial-population-file    ; Name of file with initial trout populations
  background-image-file      ; Name of an optional image (BMP, JPG, GIF, or PNG) displayed in background of view; "None" if not used.
  GIS-property-for-cell-ID   ; Name of GIS file's property for cell ID
  GIS-property-for-cell-reach-name ; Name of GIS file's property for  reach name
  GIS-property-for-cell-area       ; Name of GIS file's property for cell area
  GIS-property-for-cell-dist-escape  ; Name of GIS file's property for distance to escape cover
  GIS-property-for-cell-num-hiding-places ; Name of GIS file's property for number of hiding places
  GIS-property-for-cell-frac-vel-shelter ; Name of GIS file's property for fraction of cell with velocity shelter
  GIS-property-for-cell-frac-spawn ; Name of GIS file's property for fraction of cell with spawning gravel

  ;; Year shuffler parameters
  shuffle-years?             ; If true, then shuffle input years, with input years defined by shuffle-day
  shuffle-day                ; If years are shuffled, then draw a new year on this day (a string in M/d format)
                             ; Typically will be the first day of the water year
  shuffle-rand-seed          ; Random number seed for year shuffling, which is independent of random-number-seed

  ;; Secondary output file switches -- values set in set-parameters
  individual-fish-output?    ; Set to true to get output on each individual fish
  frac-trout-to-output       ; The fraction of individual fish to report in individual-fish-output
  cell-output?               ; Set to true to get output on each habitat cell
  debug-growth?              ; Set to true to get test output for growth; code in procedure growth-rate-for.
  debug-survival?            ; Set to true to get test output for survival; code in survive.
  debug-redd-survival?       ; Set to true to get test output for redd survival; code in redd-survive
  debug-survival-logistics?  ; Set to true to get debug output for survival logistic functions; code in build-logistic-functions.
  debug-cell-variables?      ; Set true to get output of all cell variables, except depth and velocity tables; coded in read-hydraulics.
  debug-cell-hydraulics?     ; Set true to get output of cell depth and velocity tables; coded in read-hydraulics.
  debug-cell-resources?      ; Set true to get output of cell resource variables and their depletion; coded in update-habitat
  debug-spawning-readiness?  ; Set true to get output on whether trout meet criteria for spawning; coded in spawn

  ;; Display parameters -- values set in set-parameters
  world-resolution           ; Display resolution: centimeters per patch
  trout-size-scale           ; Ratio of display size to fish length
  max-shade-depth            ; Upper limit of depth for color shading, cm
  max-shade-velocity         ; Upper limit of velocity for color shading, cm/s
  max-shade-light            ; Upper limit of light intensity for color shading, W/m2

  ;; Output parameters -- values set in set-parameters
  file-output-frequency      ; Minimum number of time units between file output updates - must be an integer
  file-output-units          ; Time units for file-output-frequency: "hours", "days", "weeks", "months", or "years"
  census-days                ; List of days of year (M/d format) on which "is-census?" reports TRUE
  census-years-to-skip       ; Number of years at start of simulation before "is-census?" can be TRUE
  age-class-list             ; List defining the age classes used to break out results.
  write-frames?              ; If true, writes Interface to a graphics file each tick, for making an animation

  ;; Input testing parameter -- values set in set-parameters
  max-days-between-time-series-inputs ; maximum number of days between rows in time-series input files
                                      ; used to define missing input -- use 1.0 for daily input

  ;; Light parameters -- values set in set-parameters
  latitude             ; Site latitude (deg)
  light-correction     ; Multiplier for atmospheric light losses
  twilight-angle       ; parameter for definition of twilight: degrees that sun is below horizon
                       ; 6 deg. is "civil twilight" and higher values should not be used.
  light-at-night       ; Light intensity at night (W/m2)

  ; Observer variables
  ;; Time variables
  sim-time                   ; A LogoTime variable with simulation date & time that current time step ends on
  step-length                ; Length of current time step (d): sim-time minus prev-time
  formatted-sim-time         ; A string with formatted sim-time
  prev-time                  ; LogoTime variable for start of current step == end of previous step
  sim-end-time               ; A LogoTime variable with time that simulation ends
  sim-date                   ; A LogoTime variable for the *date* that current time step ends on
  julian-date                ; Julian date of current time (start of time step)
  light-phase                ; An output variable equal to "night" "dawn" "day" or "dusk".
  census-start-time          ; A LogoTime variable with the date on which "is-census?" can start being true
  census-day-code            ; A LogoTime day variable with the day of year for census output

  ;; Light variables
  sunlight-irradiance    ; Mean sunlight irradiance (intensity) above water for time step (W/m2)
  day-length             ; Time of sunlight in day (d)
  twilight-length        ; Duration of morning or evening twilight (d)
  sunrise-time           ; Time of sunrise (a LogoTime variable)
  sunset-time            ; Time of sunset (a LogoTime variable)
  declination            ; Used in light intensity calculations

  ;; BehaviorSpace outputs, updated in update-BehaviorSpace-outputs. Add any variables to be reported from BehaviorSpace
  ;; Modified to provide separate results for each of 3 species
  RT-age-0-abund       ; Abundance of age 0 rainbow trout
  RT-age-1-abund       ; Abundance of age 1 rainbow trout
  RT-age-2+-abund      ; Abundance of rainbow trout age 2 and older
  RT-age-0-length      ; Mean length of age 0 rainbow trout; blank if there are none
  RT-age-1-length      ; Mean length of age 1 rainbow trout; blank if there are none
  RT-age-2+-length     ; Mean length of age 2 and older rainbow trout; blank if there are none
  BT-age-0-abund       ; Abundance of age 0 brown trout
  BT-age-1-abund       ; Abundance of age 1 brown trout
  BT-age-2+-abund      ; Abundance of brown trout age 2 and older
  BT-age-0-length      ; Mean length of age 0 brown trout; blank if there are none
  BT-age-1-length      ; Mean length of age 1 brown trout; blank if there are none
  BT-age-2+-length     ; Mean length of age 2 and older brown trout; blank if there are none
  CT-age-0-abund       ; Abundance of age 0 cutthroat trout
  CT-age-1-abund       ; Abundance of age 1 cutthroat trout
  CT-age-2+-abund      ; Abundance of cutthroat trout age 2 and older
  CT-age-0-length      ; Mean length of age 0 cutthroat trout; blank if there are none
  CT-age-1-length      ; Mean length of age 1 cutthroat trout; blank if there are none
  CT-age-2+-length     ; Mean length of age 2 and older cutthroat trout; blank if there are none

  ;; Internal data structures
  species-list               ; A list of species names
  trout-logistics-table    ; A table of logistic functions for survival, by species
  polygons-dataset    ; The cell polygons, from GIS input
  cell-ID-table       ; A *table* linking cell IDs to corresponding cells
  model-patches       ; An agentset of the patches that are within the modeled space
  daily-time-step-list ; List of the times at which time steps occur on current day.
  daily-step-length-list  ; List of the lengths of time steps on current day, used in habitat and activity selection.
  fish-events-list      ; List of events that can be written to the fish events output file.
  redd-events-list      ; List of events that can be written to the redd events output file.
  redd-summary-list     ; List of output that can be written to the redd summary output file.
  cmax-interpolation-table  ; A table that contains the CMax temperature functions. Initialized in set-parameters and checked in check-interpolation-tables.
  spawn-depth-interpolation-table  ; A table that contains the spawning depth suitability functions.
  spawn-vel-interpolation-table  ; A table that contains the spawning velocity suitability functions.
  debug-output-list      ; A list used in secondary output files.
  year-shuffler-table    ; A table relating simulate year to shuffled year -- see set-up-time

  ;; Other global variables
  outfile-name-base           ; Text string for creating unique output file names for each model run; set in set-globals
  brief-pop-outfile-name      ; String with name of brief population output file; set in build-output-files
  detailed-pop-outfile-name   ; String with name of detailed population output file; set in build-output-files
  fish-events-outfile-name    ; String with name of fish events output file; set in build-output-files
  redd-events-outfile-name    ; String with name of redd events output file; set in build-output-files
  redd-summary-outfile-name   ; String with name of redd summary output file
  fish-outfile-name           ; String with name of fish output file; set in build-output-files
  cell-outfile-name           ; String with name of cell output file; set in build-output-files
  next-output-time            ; A LogoTime for when the next file output is scheduled.
  shade-variable              ; Controls patch shading. Equal to "velocity", "depth", "light" or "off"

  ; Reach parameters -- values set in set-parameters (which is in the parameter file)
  ; These are lists with values for each reach
  ; New parameters must be declared here, initialized in set-parameters, and added to the procedure test-parameters
  reach-names                 ; Text names of reaches-- must match values in GIS file
  reach-drift-concs           ; Concentration of drift food, g/cm3
  reach-search-prods          ; Production of search food, g/cm2/d
  reach-shelter-speed-fracs   ; Fraction by which velocity shelter reduces swimming speed
  reach-prey-energy-densities ; Energy content of fish prey, J/g
  reach-drift-regen-distances ; Distance over which drift is regenerated (cm)
  reach-shadings              ; Multiplier for shading: fraction of sunlight *not* blocked by shading
  reach-fish-pred-mins        ; Minimum probability of surviving fish predation
  reach-terr-pred-mins        ; Minimum probability of surviving terrestrial predation
  reach-light-turbid-coefs    ; Multiplier in relation between light attentuation coefficient and turbidity (cm-1 NTU-1)
  reach-light-turbid-consts   ; Constant in relation between light attentuation coefficient and turbidity (cm-1)
  reach-max-spawn-flows       ; Maximum flow (m3/s) at which trout can spawn
  reach-shear-As              ; Multiplier (s/m3) in the relation between flow and shear stress, for redd scour
  reach-shear-Bs              ; Exponent in the relation between flow and shear stress, for redd scour

  ;; Reach initialization parameters -- values set in set-parameters
  time-series-input-files     ; Name of time series input file
  depth-file-names            ; Depth lookup table for cells
  velocity-file-names         ; Velocity lookup table for cells

  ; Trout and redd parameters -- values set in set-parameters (which is in the parameter file)
  ; These are lists with values for each species
  ; New parameters must be declared here, initialized in set-parameters, and added to the procedure test-parameters
  trout-display-color
  trout-capture-R1
  trout-capture-R9
  trout-cmax-A
  trout-cmax-B
  trout-emerge-length-min
  trout-emerge-length-mode
  trout-emerge-length-max
  trout-react-dist-A
  trout-react-dist-B
  trout-turbid-threshold
  trout-turbid-min
  trout-turbid-exp
  trout-light-threshold
  trout-light-min
  trout-light-exp
  trout-energy-density
  trout-fitness-horizon
  trout-fitness-length
  trout-max-speed-A
  trout-max-speed-B
  trout-max-speed-C
  trout-max-speed-D
  trout-max-speed-E
  trout-move-radius-max
  trout-move-radius-L1
  trout-move-radius-L9
  trout-pisciv-length
  trout-resp-A
  trout-resp-B
  trout-resp-C
  trout-resp-D
  trout-search-area
  trout-spawn-egg-viability
  trout-spawn-fecund-mult
  trout-spawn-fecund-exp
  trout-spawn-suitability-tol
  trout-spawn-start-day
  trout-spawn-end-day
  trout-spawn-max-flow-change
  trout-spawn-max-temp
  trout-spawn-min-age
  trout-spawn-min-length
  trout-spawn-min-cond
  trout-spawn-min-temp
  trout-spawn-prob
  trout-spawn-wt-loss-fraction
  trout-weight-A
  trout-weight-B
  mort-condition-S-at-K5
  mort-fish-pred-D1
  mort-fish-pred-D9
  mort-fish-pred-hiding-factor
  mort-fish-pred-I1
  mort-fish-pred-I9
  mort-fish-pred-L1
  mort-fish-pred-L9
  mort-fish-pred-P1
  mort-fish-pred-P9
  mort-fish-pred-T1
  mort-fish-pred-T9
  mort-high-temp-T1
  mort-high-temp-T9
  mort-redd-dewater-surv
  mort-redd-hi-temp-T1
  mort-redd-hi-temp-T9
  mort-redd-lo-temp-T1
  mort-redd-lo-temp-T9
  mort-redd-scour-depth
  mort-strand-survival-when-dry
  mort-terr-pred-D1
  mort-terr-pred-D9
  mort-terr-pred-H1
  mort-terr-pred-H9
  mort-terr-pred-hiding-factor
  mort-terr-pred-I1
  mort-terr-pred-I9
  mort-terr-pred-L1
  mort-terr-pred-L9
  mort-terr-pred-V1
  mort-terr-pred-V9
  redd-devel-A
  redd-devel-B
  redd-devel-C
  redd-area

  ; Superindividual parameters -- values set in set-parameters
  trout-superind-max-rep       ; The maximum number of trout represented by a superindividual
  trout-superind-max-length     ; The length (cm) at which superindividuals turn into individuals
]

breed [ reaches reach ]    ; The habitat reaches
breed [ cells cell ]       ; The habitat cells
breed [ trout a-trout ]    ; The fish
breed [ redds redd ]       ; The redds (trout nests)

patches-own
[
  ; Static variables
  patches-cell            ; The cell that the patch belongs to
  patches-cells-ID    ; The ID code of the cell that the patch is in.
  patches-reach       ; The reach that the patch is in.

  ; Dynamic state variables
  depth                   ; In cm
  velocity                ; In cm/s
  light                   ; Light intensity at cell's mid-depth, W/m2
]

reaches-own
[
  ;; Reach setup parameters
  reach-name                    ; The name of the reach, text string
  reach-time-series-input-file        ; Name of time series input file
  reach-depth-file-name               ; Depth lookup table for cells
  reach-velocity-file-name            ; Velocity lookup table for cells

  ;; Reach-scale habitat parameters
  reach-drift-conc           ; Concentration of drift food, g/cm3
  reach-search-prod          ; Production of search food, g/cm2/d
  reach-shelter-speed-frac   ; Fraction by which velocity shelter reduces swimming speed
  reach-prey-energy-density  ; Energy content of fish prey, J/g
  reach-drift-regen-distance ; Distance over which drift is regenerated (cm)
  reach-shading              ; Multiplier for shading: fraction of sunlight *not* blocked by shading
  reach-fish-pred-min        ; Minimum probability of surviving fish predation
  reach-terr-pred-min        ; Minimum probability of surviving terrestrial predation
  reach-light-turbid-coef    ; Multiplier in relation between light attentuation coefficient and turbidity (cm-1 NTU-1)
  reach-light-turbid-const   ; Constant in relation between light attentuation coefficient and turbidity (cm-1)
  reach-max-spawn-flow       ; Maximum flow (m3/s) at which trout can spawn -- Green River estimate
  reach-shear-A              ; Multiplier (s/m3) in the relation between flow and shear stress, for redd scour -- OFF
  reach-shear-B              ; Exponent in the relation between flow and shear stress, for redd scour
  reach-pisciv-trout-density ; Density (fish/cm2) of trout long enough to be piscivorous

  ;; Time series reach variables
  reach-flow                    ; Reach flow, m3/s
  reach-temperature             ; Reach temperature, C
  reach-turbidity               ; Reach turbidity, NTU

  ;; Internal reach data data structures
  reaches-cells                 ; An agentset of the cells in the reach
  reach-time-series-input       ; LogoTimeSeries object holding flow, temperature, turbidity input

  ;; Intermediate reach variables to reduce computations
  reach-cmax-temp-func        ; A list of cmax temperature functions for each species
  reach-max-swim-temp-term    ; A list of max swim speed temperature term values for each species
  reach-resp-temp-term        ; A list of respiration temperature term values for each species
  reach-is-flow-peak?         ; true if flow peak triggers redd shear stress
  reach-scour-param           ; Reach shear stress parameter for redd scour
]

cells-own        ; Cells are turtles that represent the polygonal habitat units
[
  ; Static state variables read from GIS shapefile
  cell-ID                      ; A unique identifier string, read from shapefile
  cell-area                    ; Cell area (cm2); read in from shapefile
  cell-num-hiding-places       ; Number of hiding places in cell (integer)
  cell-dist-escape             ; A characteristic distance to escape cover (cm)
  cell-frac-vel-shelter        ; Fraction of cell providing velocity shelter (0 - 1.0)
  cell-frac-spawn              ; Fraction of cell providing spawning gravel (0 - 1.0)
  cell-reach-name              ; Name of the cell's reach

  ; Calculated static state variables
  cells-patches                ; The NetLogo patches inside the cell
  adjacent-cells               ; The cells that represent adjacent (intersecting) polygons; these include self
  cell-area-approx             ; Cell area (cm2), calculated from the number of patches, for testing only
  max-dist-to-adjacent-cell    ; Distance (cm) from cell to furthest member of adjacent-cells

  ; Dynamic state variables
  cell-available-drift          ; Availability of drift food, g/d
  cell-available-search         ; Availability of search food, g/d
  cell-available-vel-shelter  ; Available area of velocity shelter, cm2
  cell-available-hiding-places ; Available number of hiding places, integer

  ; Hydraulic calculation variables
  depth-lookup     ; Lookup table for depth
  velocity-lookup  ; Lookup table for velocity
  flow-at-wetting  ; Lowest flow at which cell is wet

  cell-logistic-table
]

trout-own
[
  trout-species           ; A text string, defined in set-parameters.
  trout-spp-index         ; An integer for the species-list item for the trout's species
  trout-age               ; Age in years, incremented each 1 January.
  trout-length            ; Length (cm).
  trout-weight            ; Weight (g).
  trout-sex               ; A text string equal to "female" or "male"
  trout-condition         ; Condition factor, 0-1.
  trout-activity          ; A text string equal to "drift", "search", or "hide".
  trout-superind-rep     ; The (integer) number of trout represented by this trout.
  growth-memory-list      ; A list of growth values experienced in previous time steps of current 24-hr period.
  consumption-memory-list ; A list of food consumption (g) experienced in previous time steps of current 24-hr period.
  survival-memory-list    ; A list of survival probabilities experienced in previous time steps of current 24-hr period.
  trout-in-shelter?       ; A boolean for whether trout is drift-feeding in velocity shelter.
  trout-spawned-this-season? ; A date-time variable for when the fish last spawned

  trout-logistic-table

  ; Intermediate variables to reduce calcs., updated in update-trout
  trout-resp-std-wt-term  ; resp-A * wt ^ resp-B
  trout-max-speed-len-term ; max-swim-A * L + max-swim-B
  trout-cmax-wt-term      ; cmax-A * wt ^ (1+cmax-B)
  trout-prev-cons-this-day ; Food consumption (g) in prev. steps of day; updated in select-habitat-and-activity and used in drift-intake and search-intake procedures

  ; Output
  trout-output-list       ; A list written in grow for individual fish output file.
  why-not-spawn           ; A text variable reporting why trout did not spawn
  trout-tracked?          ; True if trout is reported in individual-fish output file.
]

redds-own
[
  redd-species            ; The redd's species (same as trout species)
  redd-spp-index      ; An integer for the species-list item for the trout's species
  redd-num-eggs           ; The number of viable eggs in the redd (integer)
  redd-frac-developed     ; Fraction of fully developed (0.0 to 1.0)
  redd-emerge-days        ; The number of days on which the redd has been producing new trout

  ; Output variables
  redd-date-created
  redd-date-emptied
  eggs-initial
  eggs-lo-temp
  eggs-hi-temp
  eggs-dewatering
  eggs-scour
  eggs-superimp
  eggs-emerged
]

to set-globals
  ; Observer procedure that initializes global variables.
  ; Parameters are initialized separately in set-parameters.
  ; DO NOT CHANGE any of these statements!

  ifelse behaviorspace-run-number > 0
  [ set outfile-name-base (word "r" behaviorspace-run-number "_" date-and-time ".csv") ]
  [ set outfile-name-base (word date-and-time ".csv") ]
  while [(position ":" outfile-name-base) != false]    ; Replace colons, which are not allowed in file names
  [
    let the-position position ":" outfile-name-base
    set outfile-name-base replace-item the-position outfile-name-base "-"
  ]
  while [(position " " outfile-name-base) != false]    ; Replace blanks, which are not cool in file names
  [
    let the-position position " " outfile-name-base
    set outfile-name-base replace-item the-position outfile-name-base "_"
  ]

  set brief-pop-outfile-name      "b-p-o-n" ; String with temporary name of brief population output file; reset in build-output-file
  set detailed-pop-outfile-name   "d-p-o-n" ; String with temporary name of detailed population output file; reset in build-output-file
  set fish-events-outfile-name    "f-e-o-n" ; String with temporary name of fish events output file; reset in build-output-file
  set redd-events-outfile-name    "r-e-o-n" ; String with temporary name of redd events output file; reset in build-output-file
  set redd-summary-outfile-name   "r-s-o-n" ; String with name of redd summary output file
  set fish-outfile-name           "f-o-n"   ; String with temporary name of fish output file; reset in build-output-file
  set cell-outfile-name           "c-o-n"   ; String with temporary name of cell output file; reset in build-output-file

  set daily-step-length-list (list 1.0)   ; List of the lengths of time steps on current day, used in habitat and activity selection.
                                          ; Must not be empty or zero; initialize to 1.0 d.
  set light-phase "At setup"
  set fish-events-list  (list)
  set redd-events-list  (list)
  set redd-summary-list (list)

end

to setup

  ca
  reset-ticks

  ifelse random-number-seed > 0
  [ random-seed random-number-seed ]
  [ set random-number-seed 0 ] ; To make it clear that negative values do not work.

  print "Starting to set and check parameters..."
   set-globals
   set-parameters
   test-parameters
   build-reaches
   build-logistic-functions
   check-interpolation-tables
  print "Finished setting and checking parameters"

  print "Starting to read GIS files and build cells..."
   build-cells
   stop-shading ; Initialize shade-variable and draw background if there is one.
  print "Finished building cells"

  print "Starting to read hydraulics..."
   ask reaches [ read-hydraulics ]
  print "Finished reading hydraulics"

  print "Starting to set up time and check for missing input..."
   set-up-time   ; Set up date and time-series input variables
   ask reaches [ check-for-missing-input ] ; This must be after time is set up
  print "Finished setting up time and checking input"

  print "Starting to creating initial fish..."
   build-initial-populations ; Keep this after set-up-time
  print "Finished creating fish"

  update-output              ; Output model state before start of run

end

to go
  ; The main model schedule

  ; Determine if it is time to stop.
  ; This is first so that BehaviorSpace reports correct time when
  ; reporting results only at end of simulation instead of every tick.
  if time:is-after? sim-time sim-end-time
  [
    show (word "Simulation finished in " timer " seconds.")
    stop
  ]

  tick
  if ticks = 1 [ reset-timer ]

  ; Advance the date and time and update time-series habitat variables.
  update-time-and-habitat

  ; Update trout variables
  update-trout

  update-logistics

  ; Select habitat and activity, survive, grow in descending size order.
  foreach sort-on [-1 * trout-length]  trout
  [ next-trout -> ask next-trout
    [
      select-activity-and-habitat

      survive

      grow
    ]
  ]

  ; Spawning and redd actions
  ; Spawn, which is executed only on first "day" light phase of date
  if light-phase = "day"
  [
    ; Do spawning in descending size order
    foreach sort-on [-1 * trout-length]  (trout with [trout-sex = "female"])
    [ next-female -> ask next-female [ spawn ] ]
  ]

  ; Update redd development
  ask redds [ develop-eggs ]

  ; Redd survival
  ask redds [ redd-survive ]

  ; Create new trout when redds are fully developed
  ; This is also done only once per day, on the first "day" light phase
  if light-phase = "day"
  [
    ask redds with [ redd-frac-developed >= 1.0 ] [ redd-emergence ]
  ]

  ; Update outputs
  update-output

end

to test-parameters
  ; A global procedure to test reach and trout parameters to make sure they are lists
  ; with the proper number of values. Parameters must be lists with at least as many
  ; values as there are reaches/species

    ; Reach habitat parameters
  let reach-params (list
  reach-drift-concs           ; Concentration of drift food, g/cm3
  reach-search-prods          ; Production of search food, g/cm2/d
  reach-shelter-speed-fracs   ; Fraction by which velocity shelter reduces swimming speed
  reach-prey-energy-densities ; Energy content of fish prey, J/g
  reach-drift-regen-distances ; Distance over which drift is regenerated (cm)
  reach-shadings              ; Multiplier for shading: fraction of sunlight *not* blocked by shading
  reach-fish-pred-mins        ; Minimum probability of surviving fish predation
  reach-terr-pred-mins        ; Minimum probability of surviving terrestrial predation
  reach-light-turbid-coefs    ; Multiplier in relation between light attentuation coefficient and turbidity (cm-1 NTU-1)
  reach-light-turbid-consts   ; Constant in relation between light attentuation coefficient and turbidity (cm-1)
  reach-max-spawn-flows       ; Maximum flow (m3/s) at which trout can spawn
  reach-shear-As              ; Multiplier (s/m3) in the relation between flow and shear stress, for redd scour
  reach-shear-Bs              ; Exponent in the relation between flow and shear stress, for redd scour
  time-series-input-files     ; Name of time series input file
  depth-file-names            ; Depth lookup table for cells
  velocity-file-names         ; Velocity lookup table for cells
    )

  let num-reaches length reach-names
  foreach reach-params [next-param ->
    if not ((is-list? next-param) and (length next-param >= num-reaches))
    [ error (word "A reach parameter is not in the required format of a list with one at least value per reach. Its value: " next-param) ]
  ]

    ; Trout and redd parameters -- values set in set-parameters
  let trout-params (list
  trout-display-color
  trout-capture-R1
  trout-capture-R9
  trout-cmax-A
  trout-cmax-B
  trout-emerge-length-min
  trout-emerge-length-mode
  trout-emerge-length-max
  trout-react-dist-A
  trout-react-dist-B
  trout-turbid-threshold
  trout-turbid-min
  trout-turbid-exp
  trout-light-threshold
  trout-light-min
  trout-light-exp
  trout-energy-density
  trout-fitness-horizon
  trout-fitness-length
  trout-max-speed-A
  trout-max-speed-B
  trout-max-speed-C
  trout-max-speed-D
  trout-max-speed-E
  trout-move-radius-max
  trout-move-radius-L1
  trout-move-radius-L9
  trout-pisciv-length
  trout-resp-A
  trout-resp-B
  trout-resp-C
  trout-resp-D
  trout-search-area
  trout-spawn-egg-viability
  trout-spawn-fecund-mult
  trout-spawn-fecund-exp
  trout-spawn-suitability-tol
  trout-spawn-start-day
  trout-spawn-end-day
  trout-spawn-max-flow-change
  trout-spawn-max-temp
  trout-spawn-min-age
  trout-spawn-min-length
  trout-spawn-min-cond
  trout-spawn-min-temp
  trout-spawn-prob
  trout-spawn-wt-loss-fraction
  trout-weight-A
  trout-weight-B
  mort-condition-S-at-K5
  mort-fish-pred-D1
  mort-fish-pred-D9
  mort-fish-pred-hiding-factor
  mort-fish-pred-I1
  mort-fish-pred-I9
  mort-fish-pred-L1
  mort-fish-pred-L9
  mort-fish-pred-P1
  mort-fish-pred-P9
  mort-fish-pred-T1
  mort-fish-pred-T9
  mort-high-temp-T1
  mort-high-temp-T9
  mort-redd-dewater-surv
  mort-redd-hi-temp-T1
  mort-redd-hi-temp-T9
  mort-redd-lo-temp-T1
  mort-redd-lo-temp-T9
  mort-redd-scour-depth
  mort-strand-survival-when-dry
  mort-terr-pred-D1
  mort-terr-pred-D9
  mort-terr-pred-H1
  mort-terr-pred-H9
  mort-terr-pred-hiding-factor
  mort-terr-pred-I1
  mort-terr-pred-I9
  mort-terr-pred-L1
  mort-terr-pred-L9
  mort-terr-pred-V1
  mort-terr-pred-V9
  redd-devel-A
  redd-devel-B
  redd-devel-C
  redd-area
  trout-superind-max-rep       ; The maximum number of trout represented by a superindividual
  trout-superind-max-length     ; The length (cm) at which superindividuals turn into individuals
    )

  let num-species length species-list
  foreach trout-params [next-param ->
    if not ((is-list? next-param) and (length next-param >= num-species))
    [ error (word "A trout parameter is not in the required format of a list with at least one value per reach. Its value: " next-param) ]
  ]

end

to build-reaches
  ; A global procedure to create the habitat reaches
  ; One reach is created for each value in the list reach-names

  foreach reach-names [ the-reach-name ->
    let the-index position the-reach-name reach-names
    create-reaches 1
    [
      ; Set reach parameters from the lists created in set-parameters
      ;; Static reach variables and parameters
      set reach-name item the-index reach-names  ; The name of the reach, text string
      ; reaches-cells                 ; An agentset of the cells in the reach; initialized in build-cells
      set reach-time-series-input-file item the-index time-series-input-files        ; Name of time series input file
      set reach-depth-file-name item the-index depth-file-names               ; Depth lookup table for cells
      set reach-velocity-file-name item the-index velocity-file-names            ; Velocity lookup table for cells

      ;; Reach-scale habitat parameters
      set reach-drift-conc item the-index reach-drift-concs               ; Concentration of drift food, g/cm3
      set reach-search-prod item the-index reach-search-prods             ; Production of search food, g/cm2/d
      set reach-shelter-speed-frac item the-index reach-shelter-speed-fracs     ; Fraction by which velocity shelter reduces swimming speed
      set reach-prey-energy-density item the-index reach-prey-energy-densities  ; Energy content of fish prey, J/g
      set reach-drift-regen-distance item the-index reach-drift-regen-distances ; Distance over which drift is regenerated (cm)
      set reach-shading item the-index reach-shadings                     ; Multiplier for shading: fraction of sunlight *not* blocked by shading
      set reach-fish-pred-min item the-index reach-fish-pred-mins         ; Minimum probability of surviving fish predation
      set reach-terr-pred-min item the-index reach-terr-pred-mins         ; Minimum probability of surviving terrestrial predation
      set reach-light-turbid-coef item the-index reach-light-turbid-coefs     ; Multiplier in relation between light attentuation coefficient and turbidity (cm-1 NTU-1)
      set reach-light-turbid-const item the-index reach-light-turbid-consts   ; Constant in relation between light attentuation coefficient and turbidity (cm-1)
      set reach-max-spawn-flow item the-index reach-max-spawn-flows       ; Maximum flow (m3/s) at which trout can spawn -- Green River estimate
      set reach-shear-A item the-index reach-shear-As                     ; Multiplier (s/m3) in the relation between flow and shear stress, for redd scour -- OFF
      set reach-shear-B item the-index reach-shear-Bs                     ; Exponent in the relation between flow and shear stress, for redd scour

      ;; Initialize dynamic reach variables
      set reach-flow         -999                ; Initialized to ensure prev-flow is calculated on first time step
      set reach-temperature  -999                ; Initialized to ensure prev-temp is calculated on first time step
      set reach-pisciv-trout-density -999           ; Density (fish/cm2) of trout long enough to be piscivorous; updated in update-logistics
      set reach-is-flow-peak? false          ; For redd scour
      set reach-scour-param -999             ; Dummy value to indicate that it is irrelevant

      ;; Tuck reach turtles into the upper left corner; their location has no effect
      setxy min-pxcor max-pycor
    ]
  ] ; foreach reach-names
end

to set-up-time
  ; Global procedure to set date and time-series variables

  ; First, create time variables for start, end dates
  set sim-time time:create-with-format (word start-date " 00") "M/d/yyyy HH"
  set sim-end-time time:create-with-format (word end-date " 23") "M/d/yyyy HH"

  if time:is-before? sim-end-time sim-time [ error (word "Simulation start date is after the simulation end date") ]

  set formatted-sim-time time:show sim-time "M/d/yyyy HH:mm"
  set sim-date (time:create-with-format start-date "M/d/yyyy")
  set sim-date time:plus sim-date -1 "day"  ; This is just a trick to make habitat update work on 1st tick.

  ; Set up census dates
  set census-start-time time:plus sim-time census-years-to-skip "years"  ; The date on which "is-census?" can start being true
  if not is-list? census-days [ error "Census-days not initialized as a list" ]
  let temp-census-day-list (list)  ; The following code is just to avoid any problems with leading zeros in census-days.
  foreach census-days [ next-day ->
    let a-day time:show (time:create-with-format next-day "M/d") "M/d"
    set temp-census-day-list lput a-day temp-census-day-list ]
  set census-days temp-census-day-list

  ; Build time-series input data structures for each reach
  ; and check that they have values for the simulated time period.
  ask reaches
  [
    set reach-time-series-input ts-load-with-format reach-time-series-input-file "M/d/yyyy H:mm" ; Format must not use HH:mm
    let input-date-nearest-start first ts-get reach-time-series-input sim-time "ALL"
    if abs (time:difference-between input-date-nearest-start sim-time "days") > 4
     [ error (word "Time series input file " reach-time-series-input-file
        " does not include input within 4 days of the simulation start time") ]
    let input-date-nearest-end first ts-get reach-time-series-input sim-end-time "ALL"
    if abs (time:difference-between input-date-nearest-end sim-end-time "days") > 4
     [ error (word "Time series input file " reach-time-series-input-file
        " does not include input within 4 days of the simulation end time") ]
  ]

  ; Initialize the list of daily step times
  set daily-time-step-list (list) ; List of the times at which time steps occur on current day.

  ; Schedule the first output, which is always on the first time step.
  set next-output-time time:copy sim-time

  ; Set up year shuffler if used
  if shuffle-years?
  [
    let shuffle-time time:create shuffle-day
    let start-year   time:get "year" sim-time
    let start-day    time:create (time:show sim-time "MM-dd")
    let start-water-year ifelse-value (time:is-before? start-day shuffle-time)
     [ start-year ]
     [ start-year + 1 ]

    let end-year time:get "year" sim-end-time

    let end-day    time:create (time:show sim-end-time "MM-dd")
    let end-water-year ifelse-value (time:is-before? end-day shuffle-time)
     [ end-year ]
     [ end-year + 1 ]

    if start-water-year = end-water-year [ error "You cannot shuffle years when simulation starts and ends in same year"]

    let last-date-required time:create-with-format (word shuffle-day "/" end-water-year " 00:00") "M/d/yyyy HH:mm"
    let first-date-required time:create-with-format (word shuffle-day "/" (start-water-year - 1) " 00:00") "M/d/yyyy HH:mm"
    show (word "First-date-required: " (time:show first-date-required "M/d/yyyy HH:mm"))
    show (word "Last-date-required: " (time:show last-date-required "M/d/yyyy HH:mm"))

    let a-water-year start-water-year
    let year-list (list)
    while [a-water-year <= end-water-year]
    [
      set year-list lput a-water-year year-list
      set a-water-year a-water-year + 1
    ]
    show year-list

    set year-shuffler-table table:make
    let shuffled-year-list (list)

    ; Shuffle the list of input water years, using either the main generator or a local one
    ifelse shuffle-rand-seed = 0
    [ set shuffled-year-list shuffle year-list ]
    [ with-local-randomness
      [
        random-seed shuffle-rand-seed
        set shuffled-year-list shuffle year-list
      ] ; with-local-randomness
    ] ; ifelse

    foreach year-list  ; for each year, relate it to a randomly drawn year
    [ next-year ->
      let year-position position next-year year-list
      let shuffled-year (item year-position shuffled-year-list)
      table:put year-shuffler-table next-year (shuffled-year - next-year)
      ; Use shuffled-year-list instead of one-of year-list to make sure the same year is not drawn more than once
    ]
    ; show year-shuffler-table

    ask reaches
    [
      ; Make sure every reach has input through all water years, which is required only for year-shuffling
      let input-date-nearest-start first (ts-get reach-time-series-input first-date-required "ALL")
      if abs (time:difference-between input-date-nearest-start first-date-required "days") > 4
      [ error (word "Time series input file " reach-time-series-input-file
        " does not include input starting at " (time:show first-date-required "M/d/yyyy") ", which is required for year shuffler") ]
      let input-date-nearest-end first (ts-get reach-time-series-input last-date-required "ALL")
      if abs (time:difference-between input-date-nearest-end last-date-required "days") > 4
      [ error (word "Time series input file " reach-time-series-input-file
        " does not include input through " (time:show last-date-required "M/d/yyyy") ", which is required for year shuffler") ]

      ; Create a list of the times in the input, over the simulated range
      let the-input-times ts-get-range reach-time-series-input sim-time sim-end-time "LOGOTIME"

      ; Create a new time series
      let a-ts ts-create ["flow" "temperature" "turbidity" "data-time"]

      ; Loop through the input times and add a row to the new time series for each
      foreach the-input-times [ next-time ->
        let the-year time:get "year" next-time
        let the-day time:create (time:show next-time "MM-dd")
        let the-water-year ifelse-value (time:is-before? the-day shuffle-time)
        [ the-year ]
        [ the-year + 1 ]

        let shuffle-year-offset table:get year-shuffler-table the-water-year
        let shuffled-time time:plus next-time shuffle-year-offset "year"
        let the-flow ts-get reach-time-series-input shuffled-time "flow"
        let the-temperature ts-get reach-time-series-input shuffled-time "temperature"
        let the-turbidity ts-get reach-time-series-input shuffled-time "turbidity"
        let the-data-time time:show shuffled-time "M/d/yyyy HH:mm"
        set a-ts ts-add-row a-ts (sentence next-time the-flow the-temperature the-turbidity the-data-time)

      ] ; foreach the-input-times
      set reach-time-series-input a-ts
      ts-write reach-time-series-input (word "Shuffled-input-" reach-name outfile-name-base)
    ] ; ask reaches

  ]

  ; Initialize cell depths and velocities so build-initial-populations can put fish in wet cells.
  ask reaches
  [
    set reach-flow ts-get reach-time-series-input sim-time "flow"
    update-hydraulics-for reach-flow
  ]

end

to update-time-and-habitat
  ; Observer procedure to set sim-time to end of current time step and
  ; calculate when current day's time steps occur. Also calls update-habitat

  ; First record the end of previous time step as start of this one
  set prev-time time:copy sim-time

  ; Next see if the list of today's step times is empty;
  ; this occurs when the current time step ends on a new day.
  ifelse empty? daily-time-step-list
  [
    set sim-date time:plus sim-date 1 "day"
    set julian-date (time:get "dayofyear" sim-date)
    update-day-length  ; Calculate new day length, twilight length
    set declination 23.45 * sin (0.9863 * (284 + julian-date)) ; This is used in update-light

    ; Calculate when twilight begins, ends, as fractions of a day after midnight.
    ; Used below to set time steps
    let start-morning-twilight ((1.0 - day-length) / 2) - twilight-length   ; current step ends when morning twilight starts
    let end-morning-twilight start-morning-twilight + (2 * twilight-length)  ; 2nd step ends twilight-length after sunrise
    let start-evening-twilight start-morning-twilight + day-length          ; 3rd step ends twilight-length before sunset
    let end-evening-twilight end-morning-twilight + day-length             ; 4th step ends when evening twilight ends

    ; Make sure each phase has non-zero length, a problem at high latitudes near the solstice
    if start-morning-twilight < 0.001   ; night ends less than 10 minutes after midnight (0.001 days = 1.4 minutes)
    [
      show "Warning: Dawn calculated to start before 00:01; set to 00:01"
      set start-morning-twilight 0.001
    ]
    if end-morning-twilight > 0.499   ; Dawn ends less than 10 minutes before noon
    [
      show "Warning: Dawn calculated to end after 11:59; set to 11:59"
      set end-morning-twilight 0.499
    ]
    if start-evening-twilight < 0.501   ; Dusk starts less than 10 minutes after noon
    [
      show "Warning: Dusk calculated to start before 12:01; set to 12:01"
      set start-evening-twilight 0.501
    ]
    if end-evening-twilight > 0.999   ; Dusk ends less than 10 minutes before midnight
    [
      show "Warning: Dusk calculated to end after 23:59; set to 23:59"
      set end-evening-twilight 0.999
    ]

    let midnight-string time:show sim-date "yyyy-MM-dd"
    set midnight-string (word midnight-string " 00:00:00.0")
    let midnight-time time:create midnight-string

    ; Sunrise and sunset times are used in update-light
    set sunrise-time time:plus midnight-time (0.5 - (day-length / 2)) "days"
    set sunset-time time:plus midnight-time (0.5 +  (day-length / 2)) "days"

    ; Set end of current time step to start of morning twilight.
    set sim-time time:plus midnight-time start-morning-twilight "days"
    set formatted-sim-time time:show sim-time "M/d/yyyy HH:mm"

    ; Build list of times at which remainder of the day's time steps end.
    set daily-time-step-list (list (time:plus midnight-time end-morning-twilight "days"))
    set daily-time-step-list lput (time:plus midnight-time start-evening-twilight "days") daily-time-step-list
    set daily-time-step-list lput (time:plus midnight-time end-evening-twilight "days") daily-time-step-list

    ; Update habitat variables when this is first step of day
    set step-length time:difference-between prev-time sim-time "days"
    set light-phase "night"
    update-habitat
  ]
  [
    ; This is not the first time step of the day so get next time
    ; from the list then delete that time from list
    set sim-time time:copy (first daily-time-step-list)
    set formatted-sim-time time:show sim-time "M/d/yyyy HH:mm"
    set daily-time-step-list but-first daily-time-step-list

    ; Update habitat when this is not the first step of day
    set step-length time:difference-between prev-time sim-time "days"
    ifelse light-phase = "night"
    [ set light-phase "dawn" ]
    [ ifelse light-phase = "dawn"
      [ set light-phase "day" ]
      [ ifelse light-phase = "day"
        [ set light-phase "dusk" ]
        [ if light-phase = "dusk"
          [ error (word "Update of light-phase should not have value of " light-phase) ]
        ] ; if light-phase
      ] ; ifelse "day"
    ] ; ifelse "dawn"
    update-habitat
  ]

  ; At too-high latitudes, step-length can be zero, which causes errors
  if step-length <= 0.0 [ error "Step-length is zero, possibly because latitude is too high" ]

end

to update-habitat
  ; Observer procedure to set flow, temperature, depth, velocity, light, food of patches

  ; Calculate time at middle of step
  let mid-step-time time:plus prev-time (step-length / 2) "days"
  ; Calculate surface light
  update-light

  ; Have reaches update
  ask reaches
  [

    ; First, get flow, temperature, and turbidity for the time step;
    ; these are the input values having time nearest the mid-time of the time step.
    ; Also update intermediate variables that depend on flow, temperature.

    ; Check whether flow or temperature changed since last time step to avoid
    ; unnecessary updates
    let prev-flow reach-flow
    let prev-temperature reach-temperature

    ; Reset redd scour variables
    set reach-is-flow-peak? false
    set reach-scour-param -999 ; Dummy value to indicate that it is irrelevant

    set reach-flow ts-get reach-time-series-input mid-step-time "flow"
    set reach-temperature ts-get reach-time-series-input mid-step-time "temperature"
    set reach-turbidity ts-get reach-time-series-input mid-step-time "turbidity"
    ; show (word formatted-sim-time " Flow: " reach-flow " Temperature: " reach-temperature " Turbidity: " reach-turbidity)

    if (reach-flow != prev-flow)
    [
      ; Interpolate the depth and velocity from flow
      update-hydraulics-for reach-flow

      ; Re-color the patches if necessary; this is computationally expensive
      if shade-variable = "depth" or shade-variable = "velocity"
      [ shade-reaches-patches ]

      ; Update redd scour variables, only if the flow changed and there are redds in the reach
      if any? redds with [patches-reach = myself] [ update-redd-scour-vars ]
    ]

    if (reach-temperature != prev-temperature)
    [
      ; Update intermediate variables that depend on temperature
      let an-index 0
      let spp-list-length length species-list
      set reach-cmax-temp-func (list)
      set reach-max-swim-temp-term (list)
      set reach-resp-temp-term (list)
      while [an-index < spp-list-length]
      [
        let the-species item an-index species-list
        set reach-cmax-temp-func lput (interpolate cmax-interpolation-table the-species reach-temperature) reach-cmax-temp-func
        set reach-max-swim-temp-term lput (((item an-index trout-max-speed-C) * reach-temperature * reach-temperature) +
          ((item an-index trout-max-speed-D) * reach-temperature) +
          (item an-index trout-max-speed-E)) reach-max-swim-temp-term
        set reach-resp-temp-term lput (exp ((item an-index trout-resp-C) * reach-temperature * reach-temperature)) reach-resp-temp-term

        set an-index an-index + 1
      ]
    ]

    ; Update light in cells - This must be done every time step
    ; Then calculate light at mid-depth of each cell, as a patch variable so it can be displayed
    ; This sets light even for dry cells
    ; Calculate k here because it varies only among reaches
    let k ((reach-light-turbid-coef * reach-turbidity) + reach-light-turbid-const)
    ask reaches-cells
    [
      let cells-light sunlight-irradiance * ([reach-shading] of myself) * exp (-1 * k * (depth / 2))
      ask cells-patches  [ set light cells-light ]
    ]

    ; Update depleted cell variables
    ask reaches-cells
    [
      set cell-available-vel-shelter (cell-area * cell-frac-vel-shelter)
      set cell-available-hiding-places cell-num-hiding-places
      set cell-available-drift 86400 * cell-area * depth * velocity * ([reach-drift-conc] of myself) / ([reach-drift-regen-distance] of myself)
      set cell-available-search cell-area * ([reach-search-prod] of myself)
    ]

    ; If cells are colored by light, they must be updated each time step
    if shade-variable = "light" [ shade-reaches-patches ]

  ] ; ask reaches

  ; Temporary light test output
  ; User must manually delete the test file
  ;  file-open "cell-light-test.csv"
  ;  ask cells [ file-print csv:to-row (list
  ;    "depth:" depth "turbidity: " turbidity "sunlight:" sunlight-irradiance "shading:" reach-shading "cell-light:" light)
  ;  ]
  ;  file-close

  ; Optional debug output.
  if debug-cell-resources?
  [
    let cells-output-file-name (word "Debug-cell-resources-" outfile-name-base)
    if not file-exists? cells-output-file-name
    [
      file-open cells-output-file-name
      file-print "Debug output for cell resources"
      file-print (word "Created: " date-and-time)
      file-print "Time step,Reach,Cell ID,Resource,Change,New value"
      file-close
      set debug-output-list (list)
    ]
    ; Print output from previous time step, produced in "deplete-*-by"
    if not empty? debug-output-list
    [
      file-open cells-output-file-name
      foreach debug-output-list
        [ next-line -> file-print csv:to-row next-line ]
      file-close
    ]

    ; Clear the list
    set debug-output-list (list)
    ; Have cells save their status after being updated
    ask cells
    [
      let the-reach-name [reach-name] of patches-reach
      set debug-output-list fput (list ticks the-reach-name cell-ID "vel-shelter" "reset" cell-available-vel-shelter) debug-output-list
      set debug-output-list fput (list ticks the-reach-name cell-ID "hide-places" "reset" cell-available-hiding-places) debug-output-list
      set debug-output-list fput (list ticks the-reach-name cell-ID "drift" "reset" cell-available-drift) debug-output-list
      set debug-output-list fput (list ticks the-reach-name cell-ID "search" "reset" cell-available-search) debug-output-list
    ]
  ] ; if debug-cell-resources?

end

to update-trout
  ; An observer procedure to update trout variables each time step

  ; Update memory lists used by trout in habitat and activity selection.
  update-trout-memories

  ; Calculate intermediate variables to reduce calculations
  ask trout
  [
    set trout-resp-std-wt-term (item trout-spp-index trout-resp-A) * (trout-weight ^ (item trout-spp-index trout-resp-B))
    set trout-max-speed-len-term ((item trout-spp-index trout-max-speed-A) * trout-length) + (item trout-spp-index trout-max-speed-B)
    set trout-cmax-wt-term (item trout-spp-index trout-cmax-A) * (trout-weight ^ (item trout-spp-index trout-cmax-B))
  ]

  ; Update age if this is the first step of a new calendar year
  if (time:get "year" sim-time) > (time:get "year" prev-time)
  [
    ask trout [ set trout-age trout-age + 1 ]
  ]

  ; Update variable for whether trout has spawned this season, only on first day time step
  if light-phase = "day"
  [
    ; Create a temporary variable with the current *day*, for comparison to spawning date parameters
    let the-day time:create (time:show sim-time "MM-dd")

    ask trout with [ trout-spawned-this-season? ] ; Use "with" to reduce calculations
    [
      if time:is-equal? the-day (item trout-spp-index trout-spawn-start-day)
      [ set trout-spawned-this-season? false ]
    ]
  ]


end

to update-trout-memories
  ; Observer procedure to maintain memory lists of trout.
  ; daily-step-length-list is a list of the time step lengths of the 24-hr
  ; period that ends at the end of the current time step. However, the list
  ; does not include the length of the current step, which is step-length.
  ; New values are added to the start of the list, so the last value is for the
  ; time step furthest in the past.
  ; last-step-adjuster adjusts the last value to make the total, plus current
  ; step length, total to 1.0 days.
  ; The current step length is added to daily-step-length-list at the end of
  ; this procedure, so elsewhere in the procedure the list starts with length of previous step.
  ; Here, "last step" means the last step on the list, which is the oldest (least recent) one.

  let last-step-adjuster 0.0
  let num-steps-dropped 0

  ; Reduce the list of past step lengths to make its sum as close as possible to 1.0 day.
  while [ abs ((sum daily-step-length-list + step-length) - 1.0) > abs ((sum (but-last daily-step-length-list) + step-length) - 1.0) ]
  [
    set num-steps-dropped num-steps-dropped + 1
    set daily-step-length-list but-last daily-step-length-list
  ]

  ; Calculate the last step adjustment, used in update-fitness-memories-with
  let last-step-length last daily-step-length-list
  set last-step-adjuster ((1.0 - step-length - sum daily-step-length-list) / last-step-length) + 1

  ; Update trout memory lists
  ask trout [ update-fitness-memories-with num-steps-dropped last-step-adjuster ]

  ; Temporary test output. The file must be deleted by hand.
;  file-open "StepLengthListTest.csv"
;  if ticks = 1
;  [
;    file-print "time,step-length,last-step-adjuster,list"
;  ]
;  file-print csv:to-row (sentence formatted-sim-time step-length last-step-adjuster daily-step-length-list)
;  file-close

  ; Finally, add the current step length to the list
  set daily-step-length-list fput step-length daily-step-length-list

end

to update-fitness-memories-with [ a-num-steps-dropped a-last-step-adjuster ]
  ; A trout procedure to update memory lists for sub-daily time steps.
  ; New values are added to these lists at the end of select-activity-and-habitat.
  ; Rely on NetLogo for error catching

  ; If no steps are to be dropped, then stop
  if a-num-steps-dropped = 0 [ stop ]

  ; First, drop too-old items from the end of the lists
  repeat a-num-steps-dropped
  [
    set growth-memory-list but-last growth-memory-list
    set consumption-memory-list but-last consumption-memory-list
    set survival-memory-list but-last survival-memory-list
  ]

  ; Then adjust oldest item on lists if needed to represent exactly 24 hrs
  if a-last-step-adjuster != 1.0
  [
    set growth-memory-list lput (last growth-memory-list * a-last-step-adjuster) (but-last growth-memory-list)
    set consumption-memory-list lput (last consumption-memory-list * a-last-step-adjuster) (but-last consumption-memory-list)
    set survival-memory-list lput (last survival-memory-list ^ a-last-step-adjuster) (but-last survival-memory-list)
  ]

end

to update-logistics
  ; Update the logistic functions used in survival and drift feeding.
  ; Not all these need updating each tick, but the computation time is small.

  ask trout
  [
    table:clear trout-logistic-table
    table:put trout-logistic-table "fish-pred-length" (evaluate-logistic "fish-pred-length" trout-species trout-length)
    table:put trout-logistic-table "terr-pred-length" (evaluate-logistic "terr-pred-length" trout-species trout-length)
    table:put trout-logistic-table "hab-sel-radius"   (evaluate-logistic "hab-sel-radius"   trout-species trout-length)
  ]

  ask reaches
  [

    ; Calculate density of piscivorous trout, used in fish-pred-density cell logistic
    let pisciv-trout trout with [(patches-reach = myself) and (trout-length > item trout-spp-index trout-pisciv-length)]
    let not-dry-cells (reaches-cells with [ depth > 0.0 ])
    set reach-pisciv-trout-density ifelse-value (any? not-dry-cells)
      [ (count pisciv-trout) / (sum [cell-area] of not-dry-cells) ]
      [ 0.0 ]

    let temperature reach-temperature
    let pisciv-trout-density reach-pisciv-trout-density

    ; It isn't necessary to update logistics in cells that are dry AND have no redds but
    ; the execution time seems to be negligible.
    ask reaches-cells
      [
        foreach species-list [ next-species ->
          let species-logistic-table (table:get cell-logistic-table next-species)
          table:clear species-logistic-table
          table:put species-logistic-table "high-temperature" (evaluate-logistic "high-temperature" next-species temperature)
          table:put species-logistic-table "fish-pred-depth" (evaluate-logistic "fish-pred-depth" next-species depth)
          table:put species-logistic-table "fish-pred-density" (evaluate-logistic "fish-pred-density" next-species pisciv-trout-density)
          table:put species-logistic-table "fish-pred-light" (evaluate-logistic "fish-pred-light" next-species light)
          table:put species-logistic-table "fish-pred-temperature" (evaluate-logistic "fish-pred-temperature" next-species temperature)
          table:put species-logistic-table "terr-pred-depth" (evaluate-logistic "terr-pred-depth" next-species depth)
          table:put species-logistic-table "terr-pred-velocity" (evaluate-logistic "terr-pred-velocity" next-species velocity)
          table:put species-logistic-table "terr-pred-light" (evaluate-logistic "terr-pred-light" next-species light)
          table:put species-logistic-table "terr-pred-dist-escape" (evaluate-logistic "terr-pred-dist-escape" next-species cell-dist-escape)
          table:put species-logistic-table "redd-lo-temp" (evaluate-logistic "redd-lo-temp" next-species temperature)
          table:put species-logistic-table "redd-hi-temp" (evaluate-logistic "redd-hi-temp" next-species temperature)
        ] ; foreach species-list
      ] ; ask reaches-cells
  ]; ask reaches

end

to update-redd-scour-vars
  ; Reach procedure to determine whether the flow has peaked and, if so,
  ; what the reach-average shear stress is.
  ; This is called from update-habitat only on time steps when the flow has changed and there are some redds.

  let flow-list ts-get-range reach-time-series-input (time:plus sim-time -2.0 "days") (time:plus sim-time 2.0 "days") "flow"
  ifelse reach-flow >= (max flow-list)
  [ ; Flow peaked, so calculate the Haschenberger model's shear parameter
    set reach-is-flow-peak? true
    let reach-shear-stress reach-shear-A * (reach-flow ^ reach-shear-B)
    set reach-scour-param 3.33 * exp (-1.52 * (reach-shear-stress / 0.045))
  ]
  [ ; Flow did not peak
    set reach-is-flow-peak? false
    set reach-scour-param -999 ; Dummy value to indicate that it is irrelevant
  ]

end

to check-interpolation-tables
  ; Observer procedure to check the interpolation tables for CMax, spawning suitability.
  ; It must be executed after set-parameters, which creates the tables.

  ; First, make sure the number of species in the table is correct
  if table:length cmax-interpolation-table != length species-list
  [ error (word "Length of cmax interpolation table: " (table:length cmax-interpolation-table)
    " is not equal to number of species: " (length species-list)) ]

  ; Second, make sure all species are in table and keys are in ascending order with no duplicates
  foreach species-list [ next-species ->
    if not table:has-key? cmax-interpolation-table next-species
    [ error (word "Cmax temperature parameters not provided for species: " next-species
      " Check cmax-interpolation-table input in set-parameters.") ]

    let species-cmax-table table:get cmax-interpolation-table next-species
    let key-list (table:keys species-cmax-table)
    let an-index 1
    while [ an-index <= (length key-list) - 1 ]
    [
      if (item an-index key-list) <= (item (an-index - 1) key-list)
      [
        error (word "Error in CMax temperature function:\n"
          "temperatures are not in ascending order with no duplicates.\n"
          "Check cmax-interpolation-table input for species " next-species " in set-parameters")
      ]
      set an-index an-index + 1
    ]
  ]

end

to find-small-cells
  ; Observer procedure to identify any cells smaller in area than one patch.
  ; Such cells are likely to raise an error in build-cells

  ; First read in the parameters to get the GIS file name and world resolution
  set-parameters

  ; Read in the shapefile of cell polygons
  set polygons-dataset gis:load-dataset GIS-file-name

  let patch-area world-resolution * world-resolution

  show (word "Patch area is " patch-area " cm2")

  foreach (gis:feature-list-of polygons-dataset )
  [ the-polygon ->
    let cells-area (gis:property-value the-polygon GIS-property-for-cell-area) * 100 * 100    ; Convert area from m2 to cm2
    if cells-area < patch-area
    [
      show (word "Cell " (gis:property-value the-polygon GIS-property-for-cell-ID) " has area of " cells-area " cm2")
    ]
  ]

  show "find-small-cells finished"

end

to build-cells
  ; Observer procedure to read GIS data and build habitat cells

  ; Create the table linking cell IDs to cells
  set cell-ID-table table:make

  ; Read in the shapefile of cell polygons
  set polygons-dataset gis:load-dataset GIS-file-name

  ; Un-comment the next statement to see all the habitat cell variables in the shapefile.
  ; show gis:property-names polygons-dataset

  ; Re-size the world according to the polygon range and display resolution
  let space-envelope gis:envelope-of polygons-dataset
  let gis-x-range ((item 1 space-envelope) - (item 0 space-envelope))
  let gis-y-range ((item 3 space-envelope) - (item 2 space-envelope))
  let display-x-range ceiling (gis-x-range) / (world-resolution / 100) ; keep units in m
  let display-y-range ceiling (gis-y-range) / (world-resolution / 100) ; keep units in m

  resize-world 0 display-x-range 0 display-y-range

  print (word "  Spatial extents: " precision gis-x-range 2 " m east-west, " precision gis-y-range 2 " m north-south")

  ; Convert GIS polygons to World space
  gis:set-world-envelope space-envelope

  gis:set-drawing-color white
  gis:draw polygons-dataset 1

  ; Copy polygon cell ID into its patch variable
  gis:apply-coverage polygons-dataset GIS-property-for-cell-ID patches-cells-ID
  ; Cell IDs must be strings!
  ; apply-coverage sets patches-cells-ID to NaN for cells outside the GIS coverage.
  ; is-number? returns true for NaN.
  ; The "or" term of following statement excludes values of NaN
  if any? patches with [is-number? patches-cells-ID and (patches-cells-ID > 0.0 or patches-cells-ID <= 0.0)]
    [ error "Build-cells: some model patches have non-string values of patches-cells-ID" ]

  set model-patches patches with [ is-string? patches-cells-ID ]
  print (word "  Number of patches in model: " count model-patches)

  ; Each polygon creates a cell
  ; Cells set cells-patches to patches they contain
  foreach (gis:feature-list-of polygons-dataset )
  [ the-polygon ->
    create-cells 1
    [
      ; Initialize cell variables so we can test whether they were input
      set cell-ID -99
      set cell-num-hiding-places -99
      set cell-dist-escape -99
      set cells-patches no-patches

      ; Read in cell variables
      set cell-ID gis:property-value the-polygon GIS-property-for-cell-ID   ; Text cell ID code
      set cell-reach-name gis:property-value the-polygon GIS-property-for-cell-reach-name  ; The name of the cell's reach
      set cell-area (gis:property-value the-polygon GIS-property-for-cell-area) * 100 * 100    ; Convert area from m2 to cm2
      set cell-dist-escape gis:property-value the-polygon GIS-property-for-cell-dist-escape * 100    ; Convert distance from m to cm
      set cell-num-hiding-places gis:property-value the-polygon GIS-property-for-cell-num-hiding-places      ; Number of hiding places in cell
      set cell-frac-vel-shelter gis:property-value the-polygon GIS-property-for-cell-frac-vel-shelter      ; Fraction of cell with velocity shelter
      set cell-frac-spawn gis:property-value the-polygon GIS-property-for-cell-frac-spawn      ; Fraction of cell with spawning gravel

      let cell-centroid gis:centroid-of the-polygon
      let centroid-location gis:location-of cell-centroid
      setxy item 0 centroid-location item 1 centroid-location
      set shape "circle"
      set size 0.5

      set cell-logistic-table table:make
      foreach species-list [ next-species ->
        table:put cell-logistic-table next-species (table:make)
      ]

      ; Put the cell in the cell-ID lookup table
      table:put cell-ID-table cell-ID self
    ]
  ] ; foreach polygons-dataset

  ; Set the cells' agentsets of the patches in them. This code is *much* faster than
  ; simply having cells do this: set cells-patches model-patches with [ patches-cells-ID = [cell-ID] of myself ]
  ask model-patches
  [
    set patches-cell table:get cell-ID-table patches-cells-ID
    ask patches-cell
    [
      set cells-patches (patch-set cells-patches myself)
    ]
  ]
;    set cells-patches model-patches with [ patches-cells-ID = [cell-ID] of myself ]
;
;    ; Assign the reach to the cell's patches, and check for a reach name not in the reach parameters
;    let my-reach one-of reaches with [reach-name = [cell-reach-name] of myself]
;    if my-reach = nobody
;    [
;      error (word "Cell " cell-ID " has unknown reach name: " cell-reach-name)
;    ]
;    ask cells-patches
;    [
;      set patches-cell myself
;      set patches-reach my-reach
;    ]

  ask cells
  [
  ; Check for cells that have no patches within them, which would cause big errors.
    if count cells-patches = 0
    [ ; Try to take over the patch the cell is on, which is OK if that cell has > 1 patch
      let this-patch patch-here
      let this-patches-cell [patches-cell] of this-patch
      ifelse count ([cells-patches] of this-patches-cell) > 1
      [ ; steal the patch from the neighbor cell
        set cells-patches (patch-set this-patch)
        let the-cell self
        ask cells-patches
        [
          set patches-cell myself
          set patches-cells-ID [cell-ID] of myself
        ]
        ask this-patches-cell [set cells-patches cells-patches with [self != this-patch]]
      ]
      [ ; Cannot steal a patch so raise error
        error (word "Cell " cell-ID " has no patches. Reduce the value of world-resolution.")
      ]
    ] ; if count cells-patches

    ; APPROXIMATION FOR AREA - can be used to test shapefile values
    ; set cell-area-approx (count cells-patches) * world-resolution * world-resolution

    ; So cell can use patch variables, make sure it is in one of its patches
    ; (Does not always happen because centroid of irregular polygon may not be in polygon)
    if patches-cells-ID != cell-ID [ move-to min-one-of cells-patches [distance myself]]
  ]

  ; Build agentsets of cells in each reach, and assign reach variable of patches
  ask reaches
  [
    let my-name reach-name
    let this-reach self
    set reaches-cells cells with [cell-reach-name = my-name]
    if not any? reaches-cells
    [
      error (word "Reach " reach-name " has no cells. Check that reach names are consistent among input files and parameters.")
    ]
    ask reaches-cells
    [
      ask cells-patches [set patches-reach this-reach]
    ]
  ] ; ask reaches

  ; Build agentsets of adjacent cells by looping through the polygons
  ; and finding intersecting ones
  foreach gis:feature-list-of polygons-dataset
  [
    this-poly ->
    let polys-cell-ID gis:property-value this-poly GIS-property-for-cell-ID
    let polys-cell table:get cell-ID-table polys-cell-ID

    ; initialize the agentset of adjacent cells, which DOES include the current cell
    ask polys-cell [set adjacent-cells no-turtles]

    foreach gis:feature-list-of polygons-dataset
    [
      another-poly ->
      if gis:intersects? this-poly another-poly
      [
        let adjacent-cell-ID gis:property-value another-poly GIS-property-for-cell-ID
        ask polys-cell
        [
          set adjacent-cells (turtle-set adjacent-cells (table:get cell-ID-table adjacent-cell-ID))
        ]
      ]
    ]

  ]
  ; show "Done finding adjacent cells"

  ; Now set the cell's variable for max distance to an adjacent cell and do some error checking
  ask cells
  [
    set max-dist-to-adjacent-cell world-resolution * (max [distance myself] of adjacent-cells)
    if count adjacent-cells < 2 ; There should be at least one adjacent cell in addition to this one
     [ error (word "Cell " cell-ID " has no cells adjacent to it")]
    if max-dist-to-adjacent-cell < 10
     [ error "In build-cells, furthest adjacent cell is less than 10 cm away" ]
    if patches-cells-ID != cell-ID
     [ error (word "Cell is not in of one of its patches, cell ID " cell-ID ". Try smaller value of world-resolution")]
    if is-number? cell-ID
     [ error "Cell ID not initialized or initialized to a number instead of string"]
    if (cell-dist-escape = -99)
     [ error (word "cell-dist-to-hide not initialized at cell ID " cell-ID)]
    if not is-number? cell-dist-escape
     [ error (word "cell-dist-to-hide has non-number value at cell ID " cell-ID)]
    if (cell-dist-escape < 0.0)
     [ error (word "cell-dist-to-hide has negative number at cell ID " cell-ID)]
  ]

end

to read-hydraulics
  ; Reach procedure to read in depth and velocity lookup tables for each cell.

  ; For some reason we can get file reading errors without file-close-all here
  file-close-all

  ; Depth first

  ; First, make sure file name and file exist
  if empty? reach-depth-file-name [ error (word "Missing depth file name for reach " reach-name) ]
  if not file-exists? reach-depth-file-name
    [ error (word "Depth file " reach-depth-file-name " not found") ]

  ; Open file and skip an arbitrary number of header lines
  file-open reach-depth-file-name
  let next-row csv:from-row file-read-line
  while [(is-string? first next-row) and (first (first next-row) = ";")]
  [
    set next-row csv:from-row file-read-line
  ]

  ; Read in the number of flows
  let num-depth-flows first next-row
  if not is-number? num-depth-flows [ error (word "Error reading number of flows in depth file " reach-depth-file-name) ]

  ; Read in the flow list, and strip off any non-numbers at the beginning
  ; (Non-numbers result from blank spreadsheet cells)
  let reach-depth-flow-list csv:from-row file-read-line
  while [not is-number? first reach-depth-flow-list] [ set reach-depth-flow-list but-first reach-depth-flow-list ]

  ; Check for duplicate or non-increasing flows
  let index 0
  while [index < (length reach-depth-flow-list - 1)]
  [
    if (item index reach-depth-flow-list) >= (item (index + 1) reach-depth-flow-list)
     [error (word "Flow " (item index reach-depth-flow-list) " in depth input file " reach-depth-file-name " is not smaller than the next flow")]
    set index index + 1
  ]

  ; Read in the depth table for each cell
  while [not file-at-end?]
  [
    set next-row csv:from-row file-read-line
    let next-cell (word first next-row)       ; Cell IDs must be strings
    ; Make sure the cell ID is in the table created from GIS file
    if not table:has-key? cell-ID-table next-cell
    [
      error (word "Depths file " reach-depth-file-name " has a row with cell ID not in GIS file: " next-cell)
    ]
    ask (table:get cell-ID-table next-cell) ; This is the cell with the row's ID number
    [
      ; make a copy of the flow list so we can modify it for this cell
      let flow-list reach-depth-flow-list
      let cell-depth-list map [ a-depth -> a-depth * 100 ] but-first next-row   ; convert units from m to cm
      if length cell-depth-list != num-depth-flows
       [ error (word "Cell " cell-ID "does not have correct number of depths in lookup table") ]

      ; If all depths are zero (cell is never wet) skip ahead; cell will be deleted below
      if max cell-depth-list > 0
      [
        ; Add a lookup table entry for the highest flow at which depth is zero
        ; or the depth at which flow is zero
        ; Do this only if the input does not include zero flow
        if (first reach-depth-flow-list) > 0.0
        [
          ; First handle cells that are wet at all input flows
          ifelse (first cell-depth-list) > 0.0
          [
            let new-Q 0.0
            let new-D 0.0
            let Q0 (item 0 flow-list)
            let Q1 (item 1 flow-list)
            let D0 (item 0 cell-depth-list)
            let D1 (item 1 cell-depth-list)

            let a-slope (D1 - D0) / (Q1 - Q0)
            let a-Y-int D1 - (a-slope * Q1) ; This is the extrapolated depth at zero flow
            ; if depth decreases from 1st to 2nd flow, set depth at zero to first
            ifelse a-slope <= 0.0
            [ set new-D D0 ]
            [
              ; if depth at zero flow is negative, find depth when Q is zero
              ifelse a-Y-int <= 0.0
              [ set new-Q -1 * (a-Y-int / a-slope) ]
              [ set new-D a-Y-int ]
            ]
            ; Add a new lookup point for zero flow/depth
            set flow-list fput new-Q flow-list
            set cell-depth-list fput new-D cell-depth-list
          ] ; if first depth > 0.0

          [ ; cell is not wet at first flow
            ; First strip off flows with zero depth, except the highest flow with zero depth
            ; Make sure there are at least two values on the lists
            while [(length flow-list > 2) and (first cell-depth-list) = 0.0 and (item 1 cell-depth-list) = 0.0]
            [
              set flow-list but-first flow-list
              set cell-depth-list but-first cell-depth-list
            ]

            ; If there are more than two positive depths, find the flow at which cell becomes wet.
            ; If depth is >0 only at the highest flow, then leave the 2nd highest flow on the list
            ; and it becomes flow-at-wetting
            if length flow-list > 2
            [
              ; Now extrapolate the flow at which depth reaches zero, from the two lowest non-zero flows
              let Q1 (item 1 flow-list)
              let Q2 (item 2 flow-list)
              let D1 (item 1 cell-depth-list)
              let D2 (item 2 cell-depth-list)

              let a-slope (D2 - D1) / (Q2 - Q1)
              let a-Y-int D2 - (a-slope * Q2)
              let an-X-int -999                   ; Do this to handle zero or negative slopes, e.g., when depths are all zero
              if a-slope > 0.0 [set an-X-int -1 * a-Y-int / a-slope] ; This is the extrapolate flow at zero depth
              ; If depth increases from 1st to 2nd flow, and extrapolated depth at zero flow
              ; is greater than the input with zero depth, then replace the flow at zero depth
              if an-X-int > (first flow-list)
              [
                ; Replace the flow at zero depth
                set flow-list fput an-X-int (but-first flow-list)
              ]
            ] ; if length flow-list > 2
          ] ; Cell is not wet at first flow

       ] ; if input list doesn't include zero
     ] ; max cell-depth-list > 0

     ; Set flow-at-wetting and create a table of flow vs. depth
     set flow-at-wetting first flow-list
     set depth-lookup table:make
     let the-index 0
     repeat (length flow-list)
     [
       table:put depth-lookup (item the-index flow-list) (item the-index cell-depth-list)
       set the-index the-index + 1
     ]


    ]

  ]
  file-close

  ; Velocity second

  ; First, make sure file exists
  if not file-exists? reach-velocity-file-name
    [ error (word "Velocity file " reach-velocity-file-name " not found") ]

  ; Open file and skip an arbitrary number of header lines
  file-open reach-velocity-file-name
  set next-row csv:from-row file-read-line
  while [(is-string? first next-row) and (first (first next-row) = ";")]
  [
    set next-row csv:from-row file-read-line
  ]

  ; Read in the number of flows
  let num-vel-flows first next-row
  if not is-number? num-vel-flows [ error (word "Error reading number of flows in velocity file " reach-velocity-file-name) ]

  ; Read in the flow list, and strip off any non-numbers at the beginning
  ; (Non-numbers result from blank spreadsheet cells)
  let reach-velocity-flow-list csv:from-row file-read-line
  while [not is-number? first reach-velocity-flow-list] [ set reach-velocity-flow-list but-first reach-velocity-flow-list ]

  ; Read in the velocity table for each cell
  while [not file-at-end?]
  [
    set next-row csv:from-row file-read-line
    let next-cell (word first next-row)       ; Convert cell ID to a string
    ; Make sure the cell ID is in the table created from GIS file
    if not table:has-key? cell-ID-table next-cell
    [
      error (word "Velocities file " reach-velocity-file-name " has a row with cell ID not in GIS file: " next-cell)
    ]
    ask (table:get cell-ID-table next-cell) ; This is the cell with the ID in the row just read
    [
      ; make a copy of the flow list so we can modify it for this cell
      let flow-list reach-velocity-flow-list
      let cell-vel-list map [ a-vel -> a-vel * 100 ] but-first next-row   ; convert units from m to cm
      if length cell-vel-list != num-vel-flows
       [ error (word "Cell " cell-ID "does not have correct number of velocities in lookup table") ]

      ; Strip off flows below the flow at zero depth,
      ; then add zero vel. at flow of zero depth
      ; Make sure list keeps at least two items
      ; Do this only if the input does not include zero flow
      if (first flow-list) > 0.0
      [
        while [(length flow-list > 1) and (first flow-list) <= flow-at-wetting]
        [
          set flow-list but-first flow-list
          set cell-vel-list but-first cell-vel-list
        ]

        set flow-list fput flow-at-wetting flow-list
        set cell-vel-list fput 0.0 cell-vel-list
     ] ; if input list doesn't include zero

     ; Create a table of flow vs. velocity
     set velocity-lookup table:make
     let the-index 0
     repeat (length flow-list)
     [
       table:put velocity-lookup (item the-index flow-list) (item the-index cell-vel-list)
       set the-index the-index + 1
     ]
    ]

  ]
  file-close

  ; Some error checking
  ask reaches-cells
  [
    if depth-lookup = 0
    [ error (word "Depth lookup table not created for cell " cell-ID " of reach " [reach-name] of myself) ]
    if velocity-lookup = 0
    [ error (word "Velocity lookup table not created for cell " cell-ID " of reach " [reach-name] of myself) ]
    if table:length depth-lookup < 2
    [ error (word "Cell " cell-ID " has fewer than 2 values in its depth lookup table; check " [reach-depth-file-name] of myself) ]
    if table:length velocity-lookup < 2
    [ error (word "Cell " cell-ID " has fewer than 2 values in its velocity lookup table; check " [reach-velocity-file-name] of myself) ]
  ]

  ; Drop cells that are never wet
  let always-dry-cells reaches-cells with [max (table:values depth-lookup) <= 0.0]
  if any? always-dry-cells
  [
    Show (word "Deleting " (count always-dry-cells) " cells that are never wet in reach " reach-name)
    ask always-dry-cells
    [
      ask cells-patches [set pcolor black]
      die
    ]
    ; This update of model-patches is not really necessary
    set model-patches model-patches with [patches-cell != nobody]
  ]

  ; Optional debug output. This is done once.
  if debug-cell-variables?
  [
    let debug-cells-file-name (word "Debug-cell-vars-" reach-name "-" outfile-name-base)
    ; Create a list of lists that will be output, add header info.
    let out-list (list (list "Debug output for habitat cells"))
    set out-list lput (list "Created:" date-and-time) out-list
    let next-line-list (list "cell-ID" "area (m2)" "dist-to-hide" "num-hiding-places" "frac-vel-shelter"
      "frac-spawn" "num-adjacent-cells" "max-dist-to-adjacent")
    set out-list lput next-line-list out-list

;    foreach sort-on [read-from-string cell-ID] reaches-cells [ next-cell -> ask next-cell ; Cells write output, sorted by cell ID.
    foreach sort-on [cell-ID] reaches-cells [ next-cell -> ask next-cell ; Cells write output, sorted by cell ID.
      [
        set next-line-list (sentence cell-ID (cell-area / 10000) cell-dist-escape cell-num-hiding-places cell-frac-vel-shelter
          cell-frac-spawn (count adjacent-cells) max-dist-to-adjacent-cell)
        set out-list lput next-line-list out-list
      ] ; ask next-cell
    ] ; foreach

    ; Now dump it all to the output file
    csv:to-file debug-cells-file-name out-list
  ] ; if debug-cell-variables?

  ; Optional debug output. This is done once.
  if debug-cell-hydraulics?
  [
    let debug-hyd-file-name (word "Debug-cell-hyd-" reach-name "-" outfile-name-base)
    ; Create a list of lists that will be output, add header info.
    let out-list (list (list "Debug output for cell hydraulics"))
    set out-list lput (list "Created:" date-and-time) out-list

    foreach sort-on [cell-ID] reaches-cells [ next-cell -> ask next-cell ; Cells write output, sorted by cell ID.
      [
        set out-list lput (list "") out-list
        set out-list lput (list "Cell:" cell-ID) out-list
        set out-list lput (list "Depth-Flow" "Depth") out-list
        let depth-flows table:keys depth-lookup
        let num-rows length depth-flows
        let the-index 0
        while [the-index < num-rows]
        [
          let the-flow (item the-index depth-flows)
          set out-list lput (list the-flow (table:get depth-lookup the-flow)) out-list
          set the-index the-index + 1
        ]

        set out-list lput (list "") out-list
        set out-list lput (list "Cell:" cell-ID) out-list
        set out-list lput (list "Vel-Flow" "Vel") out-list
        let vel-flows table:keys velocity-lookup
        set num-rows length vel-flows
        set the-index 0
        while [the-index < num-rows]
        [
          let the-flow (item the-index vel-flows)
          set out-list lput (list the-flow (table:get velocity-lookup the-flow)) out-list
          set the-index the-index + 1
        ]

      ] ; ask next-cell
    ] ; foreach

    ; Now dump it all to the output file
    csv:to-file debug-hyd-file-name out-list
  ] ; if debug-cell-hyd?

end

to update-day-length
  ; Calculate day length and twilight length
  ; Method from: Glarner, H. "Length of day and twilight", http://www.gandraxa.com/length_of_day.xml
  ; Accessed 25 Oct 2018.
  ; Results tested to values produced by US Naval Observatory:
  ; Simplifications:
  ;   Day of year starting at winter solstice = Julian day + 10 (assumes solstice is always 21 Dec.)
  ; Should work when julian-date is greater than 365

  ; See the test code in the following procedure

  ; First calculate day length without twilight periods, using eq. 14 of Glarner.
  ; Calc. (j x Day) and convert it from radians (per Glarner) to degrees (for NetLogo)
  let day-term (0.0172 * (julian-date + 10)) * 180 / pi
  ; "pre-m" is the uncorrected value of m, also used to calculate twilight, below
  let pre-m 1 - (tan (latitude) * tan (23.439 * cos (day-term)))
  let m pre-m

  ; if m is negative, then sun never appears (polar winter)
  if m < 0.0 [ set m 0.0 ]
  ; if m > 2, sun never sets (polar summer)
  if m > 2.0 [ set m 2.0 ]

  ; Calc. day length (fraction of 1 day) using eqs. 7 and 8 of Glarner
  set day-length (acos (1 - m)) / 180

  ; Now calculate day length with the two twilight periods, using Eq. 27 of Glarner
  ; Eq. 19 of Glarner
  let i (tan twilight-angle) / (cos latitude)

  ; Eq. 23
  let n pre-m + i

  ; if n is negative, then twlight never starts (polar winter)
  if n < 0.0 [ set n 0.0 ]
  ; if n > 2, twlight never ends (polar summer)
  if n > 2.0 [ set n 2.0 ]

  ; Eq. 27
  let day-length-with-twilight (acos (1 - n)) / 180
  ; Calculate twilight length as difference in day length with, without twilight.
  ; Remember that twilight happens twice a day but we want the length of one twilight.
  set twilight-length (day-length-with-twilight - day-length) / 2

  ; The time stepping approach does not accommodate day lengths or twilight lengths of zero
  if day-length = 0.0 or day-length = 1.0
  [ error "Day length calculated to be 0.0 or 1.0; InSTREAM does not accommodate polar sites with no sunrise/sunset" ]
  if twilight-length = 0.0
  [ error "Twilight length calculated to be 0.0, which is not allowed. Make sure twilight-angle is > 0.0" ]

  ; show (word "Lat: " site-latitude " Julian date: " julian-date " Day length: " day-length " Twilight length: " twilight-length)

end

to update-light
  ; Observer procedure to calculate average surface light intensity for a time step.
  ; It is a general procedure that considers cases that may not occur in this model.
  ; It considers how much of the time period is during day and night.
  ; It assumes Julian date, sunrise-time, and sunset-time are already calculated,
  ; that sim-time is the end of the time step, and prev-time is the start of the step.
  ; Twilight is neglected here, and light during any time between sunset and
  ; sunrise is equal to the parameter "light-at-night".

  let tomorrows-sunrise-time time:plus sunrise-time 24 "hours" ; an approximation
  let tomorrows-sunset-time time:plus sunset-time 24 "hours" ; an approximation

  ; Temporary test output

;  if ticks = 1
;  [
;    if file-exists? "LightTestOutput.csv" [ file-delete "LightTestOutput.csv" ]
;    file-open "LightTestOutput.csv"
;    file-print "prev-time,sim-time,light-phase,step-length,sunrise-time,sunset-time,day-length,twilight-length,case,light-in-day,light-intensity"
;    file-close
;  ]
;  file-open "LightTestOutput.csv"
;  file-type csv:to-row (list
;    time:show prev-time "yyyy-MM-dd HH:mm"
;    time:show sim-time "yyyy-MM-dd HH:mm"
;    light-phase step-length
;    time:show sunrise-time "yyyy-MM-dd HH:mm"
;    time:show sunset-time "yyyy-MM-dd HH:mm"
;    day-length twilight-length)

  ; Calculate mean solar intensity between start and end of time step
  if step-length >= 1.0 ; step length is >= 24 hours, so use the daily mean intensity
  [
    set sunlight-irradiance (light-intensity-between sunrise-time sunset-time) * day-length
    + (light-at-night * (1.0 - day-length))
;      file-print (word ",full-day," (light-intensity-between sunrise-time sunset-time) "," sunlight-irradiance)
;      file-close
    stop
  ]

  if time:is-before? prev-time sunrise-time ; time steps starting before sunrise
  [
    if time:is-before? sim-time sunrise-time ; whole time step is at night
    [
      set sunlight-irradiance light-at-night
;      file-print (word ",all-night,0," sunlight-irradiance)
;      file-close
      stop
    ]

    if time:is-before? sim-time sunset-time ; step goes from night into day, which will include morning twilight steps
    [
      set sunlight-irradiance (((time:difference-between prev-time sunrise-time "days") * light-at-night) +
        ((light-intensity-between sunrise-time sim-time) * (time:difference-between sunrise-time sim-time "days")))
        / step-length
;      file-print (word ",night-to-day," (light-intensity-between sunrise-time sim-time) "," sunlight-irradiance)
;      file-close
      stop
    ]

    if time:is-between? sim-time sunset-time tomorrows-sunrise-time ; step goes from night through day to night
    [
      set sunlight-irradiance (((light-intensity-between sunrise-time sunset-time) * day-length)
       + ((step-length - day-length) * light-at-night))
       / step-length
;      file-print (word ",night-to-night," (light-intensity-between sunrise-time sunset-time) "," sunlight-irradiance)
;      file-close
      stop
    ]

    ; We should not get here
    error "Did not calculate sunlight intensity with prev-time < sunrise"

  ]

  if time:is-between? prev-time sunrise-time sunset-time; time steps starting during day
  [
    if time:is-before? sim-time sunset-time ; step is entirely within daytime
    [
      set sunlight-irradiance light-intensity-between prev-time sim-time
;      file-print (word ",all-day," (light-intensity-between prev-time sim-time) "," sunlight-irradiance)
;      file-close
      stop
    ]

    if time:is-between? sim-time sunset-time tomorrows-sunrise-time ; step goes from day to night
    [
      set sunlight-irradiance (((light-intensity-between prev-time sunset-time) * (time:difference-between prev-time sunset-time "days"))
       + (light-at-night * time:difference-between sunset-time sim-time "days"))
       / step-length
;      file-print (word ",day-to-night," (light-intensity-between prev-time sunset-time) "," sunlight-irradiance)
;      file-close
      stop
    ]

    if time:is-after? sim-time tomorrows-sunrise-time ; step goes from day to night to day
    [
      set sunlight-irradiance (light-intensity-between prev-time sunset-time) * (time:difference-between prev-time sunset-time "days")
      set sunlight-irradiance sunlight-irradiance + ((1 - day-length) * light-at-night)
      set sunlight-irradiance sunlight-irradiance
       + ((light-intensity-between tomorrows-sunrise-time sim-time) * (time:difference-between tomorrows-sunrise-time sim-time "days"))
      set sunlight-irradiance sunlight-irradiance / step-length
;      file-print (word ",day-to-day," (light-intensity-between prev-time sunset-time) "," sunlight-irradiance)
;      file-close
      stop
    ]

    ; We should not get here
    error "Did not calculate sunlight intensity with prev-time > sunrise"
  ]

  if (time:is-after? prev-time sunset-time) ; time steps starting after sunset
  [
    if time:is-before? sim-time tomorrows-sunrise-time ; step is all at night
    [
      set sunlight-irradiance light-at-night
;      file-print (word ",all-night-b,0," sunlight-irradiance)
;      file-close
      stop
    ]

    if time:is-between? sim-time tomorrows-sunrise-time tomorrows-sunset-time ; step is from night to day
      [
        set sunlight-irradiance (light-at-night * time:difference-between prev-time tomorrows-sunrise-time "days")
        set sunlight-irradiance sunlight-irradiance
         + ((light-intensity-between tomorrows-sunrise-time sim-time) * (time:difference-between tomorrows-sunrise-time sim-time "days"))
        set sunlight-irradiance sunlight-irradiance / step-length
;        file-print (word ",night-to-day-b," (light-intensity-between tomorrows-sunrise-time sim-time) "," sunlight-irradiance)
;        file-close
        stop
      ]

    if time:is-after? sim-time tomorrows-sunset-time ; step is from night to following night
    ; Approximate that tomorrow's day length is the same as today's
      [
        set sunlight-irradiance (light-intensity-between tomorrows-sunrise-time tomorrows-sunset-time) * day-length
        set sunlight-irradiance sunlight-irradiance + ((step-length - day-length) * light-at-night)
        set sunlight-irradiance sunlight-irradiance / step-length
;        file-print (word ",night-to-night-b" (light-intensity-between tomorrows-sunrise-time tomorrows-sunset-time) "," sunlight-irradiance)
;        file-close
        stop
      ]

    error "Did not calculate sunlight intensity with prev-time > sunset"
  ]

end

to-report light-intensity-between [a-start-time an-end-time]
  ; Observer procedure that calculates mean solar light intensity (W/m2) above water
  ; between two times. These times must not be at night.
  ; Method is modified from eq. 1.10.4 of:
  ; Duffie, J. A. and W. A. Beckman. 2013. Solar engineering of thermal processes.
  ; 4th edition. John Wiley & Sons, Inc., Hoboken, New Jersey.

  let start-hour (time:get "hour" a-start-time) + (time:get "minute" a-start-time / 60)
  let end-hour (time:get "hour" an-end-time) + (time:get "minute" an-end-time / 60)
  let sunset-end-hour (time:get "hour" sunset-time) + (time:get "minute" sunset-time / 60) ; For error-checking

  ; Some defensive programming
  if start-hour = end-hour [ report 0.0 ]  ; Can happen due to rounding in prev. lines but only near sunrise or sunset
  if start-hour > end-hour ; should not happen because times are all within 1 day
  [ error (word "start-hour: " start-hour " not before end-hour: " end-hour)]

  if time:is-before? a-start-time sunrise-time
  [ error (word "a-start-time " (time:show a-start-time "yyyy-MM-dd HH:mm") " is before sunrise: " (time:show sunrise-time "yyyy-MM-dd HH:mm"))]

  if end-hour > sunset-end-hour
  [ error (word "end-hour: " end-hour " is after sunset hour: " sunset-end-hour)] ; an-end-time can be after current day's sunset

  let w1 (-180 + (15 * start-hour))
  let w2 (-180 + (15 * end-hour))

  let term1 light-correction * 16404 / (pi * (end-hour - start-hour))  ; 16404 = solar constant * 12
  let term2 1 + 0.033 * cos ((360 * julian-date) / 365)
  let term3a (cos latitude) * (cos declination) * (sin w2 - sin w1)
  let term3b (pi * (w2 - w1) / 180) * (sin latitude) * (sin declination)
  ; show (word "start-hour: " start-hour " end-hour: " end-hour "1: " term1 " 2: " term2 " 3a: " term3a " 3b: " term3b)

  report term1 * term2 * (term3a + term3b)

end

to update-hydraulics-for [a-flow]
  ; Reach procedure to interpolate depth, velocity of patches from flow

  ask reaches-cells
  [
    let cell-depth 0.0
    let cell-velocity 0.0

    ; First, see if flow is at or above the flow at which depth first exceeds zero
    ; Using ">=" instead of ">" allows this to work when flow is zero
    if a-flow >= flow-at-wetting
    [
      set cell-depth interpolate-hydraulics depth-lookup a-flow
      set cell-velocity interpolate-hydraulics velocity-lookup a-flow
    ]

    ask cells-patches
    [
      set depth cell-depth
      set velocity cell-velocity
    ]

  ]

  ; Now color the patches
  shade-reaches-patches

end

to-report interpolate-hydraulics [a-table a-flow ]
  ; An observer procedure to interpolate a Y value from a table, given an input X value.
  ; Used for cell depth and velocity.

  let the-flows table:keys a-table
  let the-max-index (length the-flows) - 1

  ; Otherwise, find the first flow in the table above the input flow
  let the-index 1
  while [((item the-index the-flows) < a-flow) and (the-index < the-max-index) ]
  [set the-index the-index + 1]

  let t-1 item (the-index - 1) the-flows
  let t-2 item (the-index) the-flows
  let f-1 table:get a-table t-1
  let f-2 table:get a-table t-2
  let the-ratio (a-flow - t-1) / (t-2 - t-1)
  report f-1 + (the-ratio * (f-2 - f-1))

end

to build-initial-populations
  ; Observer procedure to create initial fish population from
  ; characteristics in fish initialization file

  ; Read in the file
  if  not file-exists? initial-population-file [ error (word "Initial population file " initial-population-file " not found") ]
  let the-input csv:from-file initial-population-file

  ; First make an error-checking pass through each row
  foreach the-input
  [
    next-row ->
    let first-val first next-row
    ; Skip blank rows and those starting with ";"
    if (not empty? first-val) and (first first-val != ";")
    [
      if length next-row != 7
      [ error (word "Fish initialization file " initial-population-file
        " has wrong number of values in row: " next-row) ]
      let the-species  item 0 next-row
      let the-reach    item 1 next-row
      let the-age      item 2 next-row
      let the-num-fish item 3 next-row
      let the-len-min  item 4 next-row
      let the-len-mode item 5 next-row
      let the-len-max  item 6 next-row

      ; Error checking and warnings
      if not member? the-species species-list
      [ show (word "Warning: Fish initialization file includes unused species: " the-species) ]
      if not member? the-reach reach-names
      [ show (word "Warning: Fish initialization file includes unused reach: " the-reach) ]
      if not is-number? the-age
      [ error (word "Fish initialization file " initial-population-file
        " has invalid fish age in row: " next-row) ]
      if not is-number? the-num-fish
      [ error (word "Fish initialization file " initial-population-file
        " has invalid number of fish in row: " next-row) ]
      if not is-number? the-len-min
      [ error (word "Fish initialization file " initial-population-file
        " has invalid length minimum in row: " next-row) ]
      if not is-number? the-len-mode
      [ error (word "Fish initialization file " initial-population-file
        " has invalid length mode in row: " next-row) ]
      if not is-number? the-len-max
      [ error (word "Fish initialization file " initial-population-file
        " has invalid length maximum in row: " next-row) ]
      if (the-len-min > the-len-max) or (the-len-min > the-len-mode) or (the-len-mode > the-len-max)
      [ error (word "Fish initialization file " initial-population-file
        " has invalid length parameters in row: " next-row) ]

    ] ; if not-empty?
  ] ; foreach the-input

  ; Create the fish for each row
  let trouts-reach no-turtles
  let wet-patches no-patches
  let total-num-created 0
  foreach the-input
  [
    next-row ->
    let first-val first next-row
    ; Skip blank rows and those starting with ";"
    if (not empty? first-val) and (first first-val != ";")
    [
      let the-species  item 0 next-row
      let the-reach    item 1 next-row
      let the-age      item 2 next-row
      let the-num-fish item 3 next-row
      let the-len-min  item 4 next-row
      let the-len-mode item 5 next-row
      let the-len-max  item 6 next-row

      if (member? the-species species-list) and (member? the-reach reach-names)
      [
        set total-num-created total-num-created + the-num-fish

        if the-reach != trouts-reach
        [
          set trouts-reach one-of reaches with [reach-name = the-reach]
          set wet-patches model-patches with [ (patches-reach = trouts-reach) and (depth > 0.0) ]
          if not any? wet-patches [error (word "Reach " the-reach " has no wet cells at initialization") ]
        ]

        create-trout the-num-fish
        [
          let the-length random-triangular the-len-min the-len-mode the-len-max
          ; Use this separate procedure to keep all fish initialization code in one place
          initialize-trout-with the-species the-age the-length 1
          move-to one-of wet-patches
          save-event "initialized"
        ] ; create-trout
        ] ; if member?
    ] ; if not-empty?
  ] ; foreach the-input

  print (word "  Number of trout initialized: " total-num-created)

end

to initialize-trout-with [ a-species an-age a-length a-superind-rep ]
  ; Trout reporter to initialize a newly created trout. Used for both model intialization
  ; and emergence of new trout from redds

  ; Set state variables
  set trout-species a-species
  set trout-spp-index position a-species species-list
  set trout-age an-age
  set trout-sex one-of [ "female" "male" ]
  set trout-length a-length
  set trout-condition 1.0
  set trout-weight trout-condition * ((item trout-spp-index trout-weight-A) * (trout-length ^ (item trout-spp-index trout-weight-B)))
  set trout-superind-rep a-superind-rep
  set trout-spawned-this-season? false
  set why-not-spawn "male"   ; This will be changed when females try to spawn
  set trout-tracked? (random-float 1.0 < frac-trout-to-output)  ; This reports TRUE or FALSE

  ; Set display variables
  set color (item trout-spp-index trout-display-color) - 2 + random-float 5  ; Add some variability among individuals for visibility
  set size trout-length * trout-size-scale

  ; Set memory lists and habitat selection variables.
  ; Memory lists are initialized with one value for each entry on daily-step-length-list.
  ; This is important for creating new trout in mid-simulation as in redd emergence
  set growth-memory-list (list)     ; A list of growth values experienced in previous time steps of current 24-hr period.
  set consumption-memory-list (list) ; A list of food consumption values experienced in previous time steps of current 24-hr period.
  set survival-memory-list (list)   ; A list of survival probabilities experienced in previous time steps of current 24-hr period.
  repeat (length daily-step-length-list)
  [
    set growth-memory-list lput 0.0 growth-memory-list
    set consumption-memory-list lput 0.0 consumption-memory-list
    set survival-memory-list lput 1.0 survival-memory-list
  ]

  set trout-in-shelter? false        ; A boolean for whether trout is drift-feeding in velocity shelter.
  set trout-activity "hide"          ; This is arbitrary but necessary for detailed population output file to work upon setup.

  set trout-logistic-table table:make

end

to select-activity-and-habitat
  ; A trout procedure that includes deciding whether to feed or hide,
  ; deciding which cell to occupy, growing, and depleting food.
  ; This must be executed in order of size, large to small!

  ; Un-set trout variable for whether it's using velocity shelter
  set trout-in-shelter? false

  ; Identify potential destination cells, which include the trout's current cell.
  ; Convert radius from cm to patch-widths
  ; Remember that the fish doing this may not be exactly at its cell while
  ; distances are measured from the cell center; so "cells in-radius" does not work
  let my-cell patches-cell
  let hab-sel-radius (item trout-spp-index trout-move-radius-max) * (table:get trout-logistic-table "hab-sel-radius") ; in cm
  ; In following statement, "distance my-cell" is used instead of "in-radius" because
  ; (a) distance must be measured from the cell, not the trout, and (b) it's faster.
  let dest-cells cells with [(distance my-cell) <= (hab-sel-radius / world-resolution )]

  ; If necessary, add adjacent cells to destinations
  if hab-sel-radius < [max-dist-to-adjacent-cell] of my-cell
  [ set dest-cells (turtle-set dest-cells ([adjacent-cells] of my-cell)) ]

  ; Exclude dry cells
  set dest-cells dest-cells with [depth > 0]

  ; A test output for destination cells. The user must delete the file between runs.
;  if not file-exists? "DestCellsTestOutput.csv"
;  [
;    file-open "DestCellsTestOutput.csv"
;    file-print csv:to-row (list "Fish" "FishCellX" "FishCellY" "FishLength" "HabSelRadius" "MaxDistToAdjacentCell" "CellX" "CellY" "CellDepth")
;    file-close
;  ]
;
;  let the-fish who
;  let fish-cell-x [xcor] of my-cell
;  let fish-cell-y [ycor] of my-cell
;  let the-fish-length trout-length
;  let the-dist-to-adjacent [max-dist-to-adjacent-cell] of my-cell
;
;  file-open "DestCellsTestOutput.csv"
;  ask dest-cells
;  [
;    file-print csv:to-row (list the-fish fish-cell-x fish-cell-y the-fish-length hab-sel-radius the-dist-to-adjacent xcor ycor depth)
;  ]
;  file-close
  ; End of test output code

  ; If there are no non-dry destinations, then the current cell is the only potential destination
  if not any? dest-cells [ set dest-cells (turtle-set my-cell) ]

  ; Second, calculate growth and survival over all preceding time steps (if any) of current 1-day period
  let prev-growth-this-day 0.0
  set trout-prev-cons-this-day 0.0     ; This is a trout variable so it can be used in drift-intake and search-intake procedures
  let prev-survival-this-day 1.0
  if step-length < 1.0
  [
    set prev-growth-this-day sum growth-memory-list           ; This sums the values on the list
    set trout-prev-cons-this-day sum consumption-memory-list        ; This sums the values on the list
    set prev-survival-this-day reduce * survival-memory-list  ; This calculates the product of values on list
  ]

;  ; A test output for memory lists. The user must delete the file between runs.
;  ; File headers are not feasible because number of columns varies.
;  file-open "MemoryListTestOutput.csv"
;  let output-list (sentence formatted-sim-time who "step-length:" step-length "daily-step-length-list:" daily-step-length-list)
;  set output-list (sentence output-list "growth:" growth-memory-list "consumption" consumption-memory-list "survival:" survival-memory-list)
;  set output-list (sentence output-list "prev-growth:" prev-growth-this-day "prev-cons:" trout-prev-cons-this-day "prev-survival:" prev-survival-this-day)
;  file-print csv:to-row output-list
;  file-close
;  ; End of test output code

  ; Third, identify best cell for each potential activity
  ; Hiding does not require availability of a hiding place.
  let best-hide-cell max-one-of dest-cells [fitness-for myself "hide" prev-growth-this-day prev-survival-this-day]
  let hiding-fitness [fitness-for myself "hide" prev-growth-this-day prev-survival-this-day] of best-hide-cell

  let best-drift-feed-cell max-one-of dest-cells [fitness-for myself "drift" prev-growth-this-day prev-survival-this-day]
  let drift-fitness [fitness-for myself "drift" prev-growth-this-day prev-survival-this-day] of best-drift-feed-cell

  let best-search-feed-cell max-one-of dest-cells [fitness-for myself "search" prev-growth-this-day prev-survival-this-day]
  let search-fitness [fitness-for myself "search" prev-growth-this-day prev-survival-this-day] of best-search-feed-cell

  ; Finally select best activity and move to best cell
  ; and update memory variables.
  ; Step lengths are factored in when memory variables are recorded.
  ; Trout state variables should be updated *before* trout moves to cell so they reflect conditions
  ; before the trout depletes cell resources
  ifelse (hiding-fitness > drift-fitness) and (hiding-fitness > search-fitness)
  [ ; Trout decides to hide.
    set trout-activity "hide"
    let growth-this-step (growth-rate-for self best-hide-cell "hide") * step-length
    set growth-memory-list fput growth-this-step growth-memory-list
    set consumption-memory-list fput 0.0 consumption-memory-list   ; Food consumption is zero if hiding
    let survival-this-step (non-starve-survival-for self best-hide-cell "hide") ^ step-length
    set survival-memory-list fput survival-this-step survival-memory-list
    use-cell-and-activity best-hide-cell "hide" ; Move to cell, deplete cell resources.
  ]
  [ ; Trout decides to feed
    ifelse (drift-fitness > search-fitness)
    [  ; Drift feeding is best
      let growth-this-step (growth-rate-for self best-drift-feed-cell "drift") * step-length
      let consumption-this-step (drift-intake-for self best-drift-feed-cell) * step-length
      set trout-activity "drift"
      set growth-memory-list fput growth-this-step growth-memory-list
      set consumption-memory-list fput consumption-this-step consumption-memory-list
      let survival-this-step (non-starve-survival-for self best-drift-feed-cell trout-activity) ^ step-length
      set survival-memory-list fput survival-this-step survival-memory-list
      use-cell-and-activity best-drift-feed-cell trout-activity ; Move to cell, deplete cell resources, set trout-in-shelter?.
    ]
    [  ; Search feeding is best
      set trout-activity "search"
      let growth-this-step (growth-rate-for self best-search-feed-cell "search") * step-length
      set growth-memory-list fput growth-this-step growth-memory-list
      let consumption-this-step (search-intake-for self best-search-feed-cell) * step-length
      set consumption-memory-list fput consumption-this-step consumption-memory-list
      let survival-this-step (non-starve-survival-for self best-search-feed-cell trout-activity) ^ step-length
      set survival-memory-list fput survival-this-step survival-memory-list
      use-cell-and-activity best-search-feed-cell trout-activity ; Move to cell, deplete cell resources.
    ]
  ]

  ; A test output for habitat selection. The user must delete the file between runs.
  ; "ask best-hide-cell" can be replaced with "ask dest-cells" to get more complete
  ; (and extremely large) output.
;  if not file-exists? "HabSelTestOutput.csv"
;  [
;    file-open "HabSelTestOutput.csv"
;    file-type "Time,Light phase,Step length,Fish,Fish-length,Fish-weight,Prev-growth,Prev-survival,Cell,"
;    file-print "Hide growth this step,Feed growth this step,Hide survival this step,Feed survival this step,Hide fitness,Feed fitness,Selected activity"
;    file-close
;  ]
;
;  let the-fish self
;  let output-list-a (sentence formatted-sim-time light-phase step-length who trout-length trout-weight prev-growth-this-day prev-survival-this-day)
;
;  file-open "HabSelTestOutput.csv"
;  ask best-hide-cell
;  [
;    let output-list (sentence output-list-a "best-hide-cell"
;     (growth-rate-for the-fish self "hide") ; hide growth
;     (growth-rate-for the-fish self "feed") ; feed growth; this must be after hide.
;     (non-starve-survival-for the-fish self "hide") ; hide survival
;     (non-starve-survival-for the-fish self ([temporary-feed-mode] of the-fish)) ; feed survival; temp-feed-mode was set in growth-rate-for.
;     (fitness-for the-fish "hide" prev-growth-this-day prev-survival-this-day)
;     (fitness-for the-fish "feed" prev-growth-this-day prev-survival-this-day)
;      [trout-activity] of the-fish
;     ) ; set output-list
;    file-print csv:to-row output-list
;  ]
;  ask best-feed-cell
;  [
;    let output-list (sentence output-list-a "best-feed-cell"
;     (growth-rate-for the-fish self "hide") ; hide growth
;     (growth-rate-for the-fish self "feed") ; feed growth; this must be after hide.
;     (non-starve-survival-for the-fish self "hide") ; hide survival
;     (non-starve-survival-for the-fish self ([temporary-feed-mode] of the-fish)) ; feed survival; temp-feed-mode was set in growth-rate-for.
;     (fitness-for the-fish "hide" prev-growth-this-day prev-survival-this-day)
;     (fitness-for the-fish "feed" prev-growth-this-day prev-survival-this-day)
;      [trout-activity] of the-fish
;     ) ; set output-list
;    file-print csv:to-row output-list
;  ]
;  file-close
  ; End of test output code

end

to grow
  ; A trout procedure to update weight, length, and condition from growth (g).
  ; Executed by the main model schedule (go procedure).
  ; After select-activity-and-habitat is executed, the growth (g) for the current
  ; time step is the first item on growth-memory-list.

  let a-growth first growth-memory-list

  ; Update weight
  set trout-weight trout-weight + a-growth
  if trout-weight < 0.0 [ set trout-weight 0.0 ]

  ; Update length and condition
  ; This reproduces the method in length-with-growth for efficiency
  let healthy-weight (item trout-spp-index trout-weight-A) * (trout-length ^ (item trout-spp-index trout-weight-B))
  ifelse trout-weight > healthy-weight      ; True if fish grows in length
  [
    set trout-length length-for-weight trout-weight
    set trout-condition 1.0
    set size trout-size-scale * trout-length ; Update display size
  ]
  [
    ; Fish lost weight, so no change in length
    set trout-condition (trout-weight / healthy-weight)
  ]

  ; Save the state of the fish here for output later, if individual fish output is on.
  if individual-fish-output? and trout-tracked?
  [
    set trout-output-list (list behaviorspace-run-number formatted-sim-time light-phase ([reach-name] of patches-reach)
          who trout-species trout-age trout-length trout-weight trout-condition trout-superind-rep trout-activity
          (first growth-memory-list) ; Growth in current time step.
          (first survival-memory-list) ; Survival probability for current time step.
      patches-cells-ID (([xcor] of patches-cell) * world-resolution) (([ycor] of patches-cell) * world-resolution) ([reach-temperature] of patches-reach) depth velocity
          light ([cell-num-hiding-places] of patches-cell) ([cell-available-hiding-places] of patches-cell))
  ]

  ; Finally, separate superindividuals if they are big enough
  if (trout-superind-rep > 1) and (trout-length >= (item trout-spp-index trout-superind-max-length))
  [
    separate-superindividual
  ]

end

to-report weight-from-length-and-condition [ a-length a-condition ]
  ; Trout procedure to calculate weight from length and condition

  if a-length < 0 [ error "weight-from-length-and-condition received a negative length"]
  if a-condition < 0 or a-condition > 1.0
   [ error (word "weight-from-length-and-condition received an illegal condition: " a-condition)]

  report a-condition * (item trout-spp-index trout-weight-A) * (a-length ^ (item trout-spp-index trout-weight-B))

end

to-report length-for-weight [ a-weight ]
  ; A trout procedure to calculate length of a healthy fish for the weight

  if a-weight < 0 [ error "length-for-weight received a negative weight"]

  report (a-weight / (item trout-spp-index trout-weight-A)) ^ (1 / (item trout-spp-index trout-weight-B))

end

to-report condition-with-growth [ a-growth ]
  ; A trout procedure to calculate what trout-condition would be with a change in weight.

  ; Report 1.0 if condition is already 1.0 and growth is positive.
  if (a-growth > 0.0) and (trout-condition = 1.0)
  [ report 1.0 ]

  let new-weight trout-weight + a-growth
  if new-weight <= 0.0 [ report 0.0 ]
  let healthy-weight (item trout-spp-index trout-weight-A) * (trout-length ^ (item trout-spp-index trout-weight-B))
  ifelse new-weight > healthy-weight
  [ report 1.0 ] ; Fish will reach condition of 1.0.
  [ report new-weight / healthy-weight ] ; Fish remains below healthy weight for length so condition < 1.0.

end

to-report length-with-growth [ a-growth ]
  ; A trout procedure to calculate what length would be with a change in weight.

  ; Report current length if growth is negative.
  if (a-growth <= 0.0) [ report trout-length ]

  let new-weight trout-weight + a-growth
  let healthy-weight (item trout-spp-index trout-weight-A) * (trout-length ^ (item trout-spp-index trout-weight-B))
  ifelse new-weight > healthy-weight
  [ report length-for-weight new-weight ] ; Fish will reach condition of 1.0 and grow in length.
  [ report trout-length ] ; Fish remains below healthy weight for length so does not get longer.

end

to-report fitness-for [a-fish an-activity prev-growth prev-survival]
  ; A cell procedure to calculate expected fitness for the fish if it executes the activity in the cell.
  ; Prev-growth and prev-survival are the growth and non-starvation survival for previous time steps in current day.

  ; Start with some defensive programming
  if (an-activity != "drift") and (an-activity != "search") and (an-activity != "hide")
    [ error (word "Fitness-for received an illegal value of an-activity: " an-activity) ]

  ; Set fitness to zero if fish would swim faster than max swimming speed
  let fish-max-speed max-swim-speed-for a-fish self
  if (an-activity = "search") and (velocity > fish-max-speed) [ report 0.0 ]

  if (an-activity = "drift") and (drift-swim-speed-for a-fish self > fish-max-speed) [ report 0.0 ]

  ; Calculate growth and survival in this cell.
  let growth-this-step (growth-rate-for a-fish self an-activity) * step-length

  let survival-this-step (non-starve-survival-for a-fish self an-activity) ^ step-length

  let daily-growth prev-growth + growth-this-step
  let daily-survival prev-survival * survival-this-step

  ; Now calculate the fitness measure for this daily growth and survival.
  let the-species ([trout-species] of a-fish)
  let the-spp-index ([trout-spp-index] of a-fish)
  let time-horizon item the-spp-index trout-fitness-horizon

  let mean-starv-survival-to-horizon [mean-condition-survival-with daily-growth time-horizon] of a-fish

  let fitness (daily-survival * mean-starv-survival-to-horizon) ^ time-horizon

  let maturity-length item the-spp-index trout-fitness-length
  if ([trout-length] of a-fish < maturity-length)
  [ ; Expected length is less than fitness-length
    let length-at-horizon [length-with-growth (daily-growth * time-horizon)] of a-fish
    set fitness fitness * ifelse-value (length-at-horizon < maturity-length)
    [ length-at-horizon / maturity-length ] ; fish will remain below maturity length
    [ 1.0 ] ; Expected length exceeds fitness-length
  ]

  ; A test output for fitness evaluation. The user must delete the file between runs.
  ; This generates a very large output file; use it only with very few fish.
;  if not file-exists? "FitnessTestOutput.csv"
;  [
;    file-open "FitnessTestOutput.csv"
;    file-type "Time,Light-phase,Step length,Velocity,DriftSwimSpeed,MaxSwim,Fish,Fish-length,Fish-weight,Prev-growth,Prev-survival,Activity,"
;    file-print "Growth this step,Survival this step,(blank),Daily-growth,Daily-survival,Mean-starv-to-horizon,fitness"
;    file-close
;  ]
;
;  file-open "FitnessTestOutput.csv"
;  let output-list (sentence formatted-sim-time light-phase step-length velocity (drift-swim-speed-for a-fish self)
;    fish-max-speed
;    [who] of a-fish
;    [trout-length] of a-fish
;    [trout-weight] of a-fish
;    prev-growth prev-survival an-activity
;    growth-this-step survival-this-step " "
;    daily-growth daily-survival
;    mean-starv-survival-to-horizon fitness)
;  file-print csv:to-row output-list
;  file-close
  ; End of test output code

  report fitness

end

to use-cell-and-activity [ a-cell an-activity ]
  ; A trout procedure to move to and conduct the activity in a cell
  ; Procedure moves fish to somewhere within a cell's patches,
  ; updates fish variable for activity, and
  ; depletes any resources used by the fish

  move-to one-of [cells-patches] of a-cell

  set trout-activity an-activity

  let superind-rep trout-superind-rep ; Create a local variable that can be used in any context
  if superind-rep < 1 [error (word "trout-superind-rep is " superind-rep) ]

  ; Deplete food consumed by fish. Fish's current value of fish-intake
  ; is set in select-activity-and-habitat to its consumption if feeding in best cell.
  ifelse an-activity = "drift"
  [ ; deplete drift consumption and velocity shelter
    ask a-cell
    [
      deplete-drift-by ((drift-intake-for myself self) * superind-rep)
      if cell-available-vel-shelter >= (([trout-length] of myself) ^ 2) * superind-rep
        [
          deplete-vel-shelter-by (([trout-length] of myself) ^ 2) * superind-rep
          ask myself [ set trout-in-shelter? true ]
        ] ; if cell-available
    ] ; ask a-cell
  ] ; if an-activity = "drift"
  [ ifelse an-activity = "search"
    [ ; deplete search consumption
      ask a-cell [ deplete-search-by ((search-intake-for myself self) * superind-rep) ]
    ]
    [
      ifelse an-activity = "hide"
      [ ; deplete hiding cover, if available.
        ask a-cell [ deplete-hiding-places-by superind-rep ]
      ]
      [
        error (word "use-cell-and-activity received an illegal activity " an-activity)
      ]
    ] ; ifelse an-activity = hide
  ] ; ifelse an-activity = search

end

to-report mean-condition-survival-with [a-daily-growth T]
  ; A trout reporter to calculate mean survival of low condition until the time horizon
  ; a-daily-growth is growth (g/d) and T is the number of days to the time horizon.
  ; This procedure implements the linear model of low condition survival.
  ; The equation s = (2 * s5) + (2 * k) - (2 * k * s5) - 1 implements the linear model.

  let k trout-condition

  ; If trout has condition = 1 and positive growth, survival is 1.0
  if (k = 1.0) and (a-daily-growth >= 0.0) [ report 1.0 ]

  let s5 item trout-spp-index mort-condition-S-at-K5

  let s-now (2 * s5) + (2 * k) - (2 * k * s5) - 1 ; Survival at current condition

  if abs a-daily-growth < 0.0001
  [ ; Condition is steady
    report s-now
  ]

  ; Calculate daily change in condition
  let healthy-weight (item trout-spp-index trout-weight-A) * (trout-length ^ (item trout-spp-index trout-weight-B))
  let d-k a-daily-growth / healthy-weight
  let s-k0 (2 * s5) - 1 ; Survival at condition of zero

  ; Now calculate first moment of all remaining trajectories in K.
  ifelse d-k < 0
  [ ; Condition is decreasing
    let t-at-k-0 k / (-1 * d-k)
    ifelse t-at-k-0 <= T
    [ ; Condition reaches zero before time horizon
      let area1 t-at-k-0 * (s-now + s-k0) / 2
      let area2 (T - t-at-k-0) * s-k0
      report ((area1 + area2) / T)
    ]
    [ ; Condition remains above zero to time horizon
      let k-at-t k + (T * d-k)
      let s-at-t (2 * s5) + (2 * k-at-t) - (2 * k-at-t * s5) - 1
      report ((s-now + s-at-t) / 2)
    ]
  ] ; Condition is decreasing
  [ ; Condition is increasing or steady
    let t-at-k-1 (1.0 - k) / d-k
    ifelse t-at-k-1 <= T
    [ ; Condition reaches one before time horizon
      let area1 t-at-k-1 * (s-now + 1.0) / 2
      let area2 (T - t-at-k-1) * 1.0
      report ((area1 + area2) / T)
    ]
    [ ; Condition remains below 1.0 to time horizon
      let k-at-t k + (T * d-k)
      let s-at-t (2 * s5) + (2 * k-at-t) - (2 * k-at-t * s5) - 1
      report ((s-now + s-at-t) / 2)
    ]
   ] ; Condition is increasing

end

to deplete-drift-by [ a-consumption ]
  ; A cell procedure to reduce drift food availability by a fish's consumption.

  if (cell-available-drift - a-consumption) < -0.00000001  ; Use a small difference instead of "=" to avoid errors due to roundoff.
  [
    error (word "Error in deplete-drift-by: drift consumption of " a-consumption
      "\n exceeds availability of " cell-available-drift)
  ]
  set cell-available-drift cell-available-drift - a-consumption

  if debug-cell-resources?
  [
     set debug-output-list lput (list ticks cell-reach-name cell-ID "drift" a-consumption cell-available-drift) debug-output-list
  ]

end

to deplete-search-by [ a-consumption ]
  ; A cell procedure to reduce search food availability by a fish's consumption.

  if (cell-available-search - a-consumption) < -0.00000001  ; Use a small difference instead of "=" to avoid errors due to roundoff.
  [
    error (word "Error in deplete-search-by: search consumption of " a-consumption
      "\n exceeds availability of " cell-available-search)
  ]
  set cell-available-search cell-available-search - a-consumption

  if debug-cell-resources?
  [
     set debug-output-list lput (list ticks cell-reach-name cell-ID "search" a-consumption cell-available-search) debug-output-list
  ]

end

to deplete-vel-shelter-by [ a-consumption ]
  ; A cell procedure to reduce velocity shelter availability by a fish's consumption.
  ; The area used by last trout may exceed available area.

  set cell-available-vel-shelter cell-available-vel-shelter - a-consumption
  if cell-available-vel-shelter < -0.00000001  ; Use a small difference instead of "=" to avoid errors due to roundoff.
  [
    error (word "Error in deplete-vel-shelter-by: shelter consumption of " a-consumption
      "\n exceeds availability of " cell-available-vel-shelter)
  ]

  if debug-cell-resources?
  [
     set debug-output-list lput (list ticks cell-reach-name cell-ID "vel-shelter" a-consumption cell-available-vel-shelter) debug-output-list
  ]

end

to deplete-hiding-places-by [ a-num-places ]
  ; A cell procedure to reduce the number of available hiding places.
  ; It is OK for a trout to hide when there are no available hiding places,
  ; in which case cell-available-hiding-places remains unchanged.

  if cell-available-hiding-places >= a-num-places
  [
    set cell-available-hiding-places cell-available-hiding-places - a-num-places
  ]

  if debug-cell-resources?
  [
     set debug-output-list lput (list ticks cell-reach-name cell-ID "hide-places" a-num-places cell-available-hiding-places) debug-output-list
  ]

end

to-report growth-rate-for [ a-fish a-cell an-activity ]
  ; An observer procedure to calculate the daily growth (g/d) that a fish would get for a cell and activity.
  ; an-activity has valid values of "drift", "search", or "hide".
  ; Growth is reported as a *daily* rate of mass increase.

  ; Test output for growth -- this produces extremely large output files.
  ; This optional output code should normally BE COMMENTED OUT because this procedure is called many times.
  ; There is also optional output code at the END of this procedure that must be
  ; un-commented or commented out along with the following statements.
;  if debug-growth?
;  [
;    let growth-output-file-name (word "Debug-growth-" outfile-name-base)
;    if not file-exists? growth-output-file-name
;    [
;      file-open growth-output-file-name
;      file-print (word "InSTREAM-7 growth test output file, Created " date-and-time)
;      file-type "trout-species,activity,trout-length,trout-weight,temperature,depth,velocity,light,turbidity,"
;      file-type "drift-available,search-available,vel-shelter-available,max-swim-speed,drift-conc,search-prod,"
;      file-print "CMax-temp-term,CMax,drift-intake,search-intake,growth-rate"
;      file-close
;    ]
;
;    let the-reach [patches-reach] of a-cell
;
;    file-open growth-output-file-name
;    file-type csv:to-row (list
;      [trout-species] of a-fish
;      an-activity
;      [trout-length] of a-fish
;      [trout-weight] of a-fish
;      [reach-temperature] of the-reach
;      [depth] of a-cell
;      [velocity] of a-cell
;      [light] of a-cell
;      [reach-turbidity] of the-reach
;      [cell-available-drift] of a-cell
;      [cell-available-search] of a-cell
;      [cell-available-vel-shelter] of a-cell
;      max-swim-speed-for a-fish a-cell
;      [reach-drift-conc] of the-reach
;      [reach-search-prod] of the-reach
;      item ([trout-spp-index] of a-fish) ([reach-cmax-temp-func] of the-reach)
;      c-stepmax-for a-fish a-cell
;      drift-intake-for a-fish a-cell
;      search-intake-for a-fish a-cell
;      )
;  ]
  ; End of test output code except for final result, below

  let the-growth -999
  let fish-energy-density item ([trout-spp-index] of a-fish) trout-energy-density
  let the-reach [patches-reach] of a-cell

  ifelse (an-activity = "drift")
  [
    ; Calculate food intake rate
    let drift-intake drift-intake-for a-fish a-cell

    ; Determine swimming speed, which depends on whether velocity shelter is available.
    let drift-swim-speed  ifelse-value ([cell-available-vel-shelter] of a-cell > ([trout-length] of a-fish) ^ 2)
      [ ([velocity] of a-cell) * ([reach-shelter-speed-frac] of the-reach) ]
      [ [velocity] of a-cell ]

    ; Calculate respiration, which depends on swim speed
    let drift-resp respiration-for a-fish a-cell drift-swim-speed

    ; Do the energy balance and convert J energy intake to g growth
    let drift-net-energy-intake (drift-intake * ([reach-prey-energy-density] of the-reach)) - drift-resp
    set the-growth drift-net-energy-intake / fish-energy-density
  ] ; an-activity = "drift"
  [
    ifelse (an-activity = "search")
    [
      ; Calculate food intake rate
      let search-intake search-intake-for a-fish a-cell

      ; Calculate respiration, which depends on swim speed; for search feeding,
      ; swim speed is the cell's velocity
      let search-resp respiration-for a-fish a-cell ([velocity] of a-cell)

      ; Do the energy balance and convert J energy intake to g growth
      let search-net-energy-intake (search-intake * ([reach-prey-energy-density] of the-reach)) - search-resp
      set the-growth search-net-energy-intake / fish-energy-density
    ] ; an-activity = "search"
    [
      ifelse (an-activity = "hide")
      [ ; growth when hiding

        ; Food intake rate is zero when hiding
        ; Swimming speed is zero when hiding
        ; Net energy intake is therefore -1 * respiration at zero swimming speed.

        ; Calculate respiration, which depends on swim speed
        let hide-resp respiration-for a-fish a-cell 0.0

        ; Do the energy balance and convert J energy intake to g growth
        let hide-net-energy-intake (0.0 - hide-resp)
        set the-growth hide-net-energy-intake / fish-energy-density
      ] ; an-activity = "hide"
      [
        error (word "growth-rate-for was passed an illegal value of activity: " an-activity)
      ] ; error
    ] ; an-activity != "search"
  ]  ; an-activity != "drift"

  ; Optional test output is first so it produces full results no matter which activity is input.
  ; Test output for growth.
;  if debug-growth?
;  [
;    file-print (word "," the-growth)
;    file-close
;  ]

  report the-growth


end

to-report drift-intake-for [ a-fish a-cell ]
  ; An observer procedure to calculate the *daily* gross rate of
  ; food intake (g) from drift feeding

  let the-species [trout-species] of a-fish
  let the-spp-index [trout-spp-index] of a-fish
  let the-reach [patches-reach] of a-cell

  ; (1) Calculate the trout length function for detection distance
  let detect-dist (item the-spp-index trout-react-dist-A +
    ((item the-spp-index trout-react-dist-B) * [trout-length] of a-fish))

  ; (2) Calculate the turbidity function for detection distance
  let turbidity [reach-turbidity] of the-reach
  let turbid-function ifelse-value (turbidity <= (item the-spp-index trout-turbid-threshold))
  [ 1.0 ]
  [ (item the-spp-index trout-turbid-min) + (1.0 - (item the-spp-index trout-turbid-min)) *
    exp ((item the-spp-index trout-turbid-exp) * (turbidity - (item the-spp-index trout-turbid-threshold)))
  ]

  ; (3) Calculate the light function.
  let light-function ifelse-value ([light] of a-cell >= (item the-spp-index trout-light-threshold))
  [ 1.0 ]
  [ (item the-spp-index trout-light-min) + (1.0 - (item the-spp-index trout-light-min)) *
    exp ((item the-spp-index trout-light-exp) * ((item the-spp-index trout-light-threshold) - [light] of a-cell))
  ]

  ; (4) Calculate detection distance and capture area
  set detect-dist detect-dist * turbid-function * light-function

  let capture-height [depth] of a-cell
  if detect-dist < capture-height [ set capture-height detect-dist ]
  let capture-area (2 * detect-dist * capture-height)

  ; (5) Calculate capture success, which depends on cell velocity, not swim speed.
  let the-max-swim-speed max-swim-speed-for a-fish a-cell
  let capture-success (evaluate-logistic "capture-success" the-species (([velocity] of a-cell) / the-max-swim-speed))

  ; (6) Calculate drift intake as a daily rate, g/d
  let gross-intake capture-area * ([reach-drift-conc] of the-reach) * ([velocity] of a-cell) * 86400 * capture-success ; 86400 seconds per day

  ; (7) Calculate maximum intake and use it to limit intake
  let c-stepmax c-stepmax-for a-fish a-cell
  if gross-intake > c-stepmax [ set gross-intake c-stepmax ]

  ; (8) Limit gross intake by food availability in cell, considering superindividual size
  let avail-drift ([cell-available-drift] of a-cell) / ([trout-superind-rep] of a-fish)
  if gross-intake > avail-drift  [ set gross-intake avail-drift ]

  ; (9) Report gross intake
  report gross-intake

end

to-report search-intake-for [ a-fish a-cell ]
  ; An observer procedure to calculate the *daily* gross rate of
  ; food intake (g) from search feeding

  ; (1) Calculate velocity term
  let the-max-swim-speed max-swim-speed-for a-fish a-cell
  let the-vel-ratio ((the-max-swim-speed - [velocity] of a-cell) / the-max-swim-speed)
  ; if this ratio is negative, then there is no food intake
  if the-vel-ratio < 0.0 [ report 0 ]

  ; (2) Calculate search intake
  let gross-intake ([reach-search-prod] of ([patches-reach] of a-cell)) * (item ([trout-spp-index] of a-fish) trout-search-area) * the-vel-ratio

  ; (3) Calculate maximum intake and use it to limit intake
  let c-stepmax c-stepmax-for a-fish a-cell
  if gross-intake > c-stepmax [ set gross-intake c-stepmax ]

  ; (4) Limit gross intake by food availability in cell
  let avail-search ([cell-available-search] of a-cell) / ([trout-superind-rep] of a-fish)
  if gross-intake > avail-search [ set gross-intake avail-search ]

  ; (5) Report gross intake
  report gross-intake

end

to-report respiration-for [ a-fish a-cell a-swim-speed ]
  ; An observer procedure to calculate total respiration (J/d) for a fish
  ; Some equation terms are updated in update-habitat and update-trout

  ; Standard respiration
  let resp-temp-term [reach-resp-temp-term] of ([patches-reach] of a-cell)
  let resp-standard ([trout-resp-std-wt-term] of a-fish) *
   (item ([trout-spp-index] of a-fish) resp-temp-term)

  ; Activity respiration depends on square of ratio of swimming speed to max swim speed
  ; If this ratio is large (> ~25) then "exp" becomes too large to calculate so instead
  ; report an arbitrary large number.
  let the-swim-speed-ratio a-swim-speed / (max-swim-speed-for a-fish a-cell)
  if the-swim-speed-ratio > 20 [ report 999999 ]
  let resp-activity-func exp ((item ([trout-spp-index] of a-fish) trout-resp-D) * (the-swim-speed-ratio ^ 2))

  report resp-standard * resp-activity-func

end

to-report c-stepmax-for [ a-fish a-cell ]
  ; An observer procedure to calculate maximum consumption for the current time step, as a daily rate (g/d)

  let cmax-temp-func [reach-cmax-temp-func] of ([patches-reach] of a-cell)
  let c-max ([trout-cmax-wt-term] of a-fish) * (item ([trout-spp-index] of a-fish) cmax-temp-func)
  let c-stepmax ( c-max - ([trout-prev-cons-this-day] of a-fish) ) / step-length

  if c-stepmax < 0.0 [ set c-stepmax 0.0 ]

  report c-stepmax

end

to-report interpolate [a-table a-species an-input ]
  ; An observer procedure to interpolate a Y value from a table, given an input X value.
  ; Used for CMax temperature function, and spawning suitability for depth and velocity

  let the-interpolation-table table:get a-table a-species

  let the-keys table:keys the-interpolation-table
  if an-input < (first the-keys)
  [
    error (word "Error in interpolate: the input " an-input
      " is less than lowest X value in the table, for species " a-species
      ". The table X values: " the-keys)
  ]

  ; Check if input X value is above the highest X value in the table
  if an-input >= last the-keys
  [
    report table:get the-interpolation-table (last the-keys)
  ]

  ; Otherwise, find the first X value in the table above the input X
  let the-index 1
  while [(item the-index the-keys) < an-input] [set the-index the-index + 1]

  let t-1 item (the-index - 1) the-keys
  let t-2 item (the-index) the-keys
  let f-1 table:get the-interpolation-table t-1
  let f-2 table:get the-interpolation-table t-2
  let the-ratio (an-input - t-1) / (t-2 - t-1)
  report f-1 + (the-ratio * (f-2 - f-1))

end

to-report drift-swim-speed-for [ a-fish a-cell ]
  ; An observer procedure to report actual swimming speed of a *drift-feeding* fish at a cell.
  ; THIS WILL NOT WORK after other (smaller) fish select their habitat; it assumes that no velocity shelter
  ; has been used by smaller fish.

  let shelter-speed-frac [reach-shelter-speed-frac] of ([patches-reach] of a-cell)
  report ifelse-value ([cell-available-vel-shelter] of a-cell > ([trout-length] of a-fish) ^ 2)
  [ ([velocity] of a-cell) * shelter-speed-frac ]
  [ [velocity] of a-cell ]

end

to-report max-swim-speed-for [ a-fish a-cell ]
  ; An observer procedure to report maximum sustainable swimming speed of a fish at a cell.

  let the-spp-index [trout-spp-index] of a-fish
  let max-swim-temp-term [reach-max-swim-temp-term] of ([patches-reach] of a-cell)

  report ([trout-max-speed-len-term] of a-fish) * (item the-spp-index max-swim-temp-term)

end

to survive
  ; The main survival submodel procedure, called from the go procedure.
  ; A trout procedure that determines whether a trout lives or dies each time step.
  ; The order in which mortality types are evaluated is defined in the model description.

  ; Optional test output is first so it is executed even if fish dies.

  ; Test output for survival.
  if debug-survival?
  [
    let debug-survival-file-name (word "Debug-survival-" outfile-name-base)
    if not file-exists? debug-survival-file-name
    [
      file-open debug-survival-file-name
      file-print (word "InSTREAM-7 survival test output file, Created:," date-and-time  ",Values are *daily* survival")
      file-type "trout ID,species,reach,trout-length,trout-condition,trout-activity,trout-in-shelter?,temperature,"
      file-type "pisciv-density,depth,velocity,dist-to-hide,avail-hiding-places,light,max-swim-speed,"
      file-print "s-temperature,s-strand,s-condition,s-terr-pred,s-fish-pred"
      file-close
    ]

    file-open debug-survival-file-name
    file-print csv:to-row (list
      who
      trout-species
      [reach-name] of patches-reach
      trout-length
      trout-condition
      trout-activity
      trout-in-shelter?
      [reach-temperature] of patches-reach
      [reach-pisciv-trout-density] of patches-reach
      depth
      velocity
      [cell-dist-escape] of patches-cell
      [cell-available-hiding-places] of patches-cell
      light
      max-swim-speed-for self patches-cell
      survival-high-temperature-for self patches-cell
  ;    survival-high-velocity-for self patches-cell trout-activity
      survival-stranding-for self patches-cell
      survival-condition-for self patches-cell
      survival-terr-pred-for self patches-cell trout-activity
      survival-fish-pred-for self patches-cell trout-activity
    )
    file-close
  ]
  ; End of test output code


  if (random-float 1.0) > (survival-high-temperature-for self patches-cell) ^ step-length
  [ ; Trout died of high temperature
    save-event "died of high temperature"
    die
  ]

;  if (random-float 1.0) > (survival-high-velocity-for self patches-cell trout-activity) ^ step-length
;  [ ; Trout died of high velocity  REMOVED
;    save-event "died of high velocity"
;    die
;  ]

  if (random-float 1.0) > (survival-stranding-for self patches-cell) ^ step-length
  [ ; Trout died of stranding
    save-event "died of stranding"
    die
  ]

  if (random-float 1.0) > (survival-condition-for self patches-cell) ^ step-length
  [ ; Trout died of poor condition
    save-event "died of poor condition"
    die
  ]

  if (random-float 1.0) > (survival-terr-pred-for self patches-cell trout-activity) ^ step-length
  [ ; Trout died of terrestrial predation
    save-event "died of terr predation"
    die
  ]

  if (random-float 1.0) > (survival-fish-pred-for self patches-cell trout-activity) ^ step-length
  [ ; Trout died of fish predation
    save-event "died of fish predation"
    die
  ]

end

to-report non-starve-survival-for [ a-fish a-cell an-activity ]
  ; The observer procedure that calculates overall daily survival probability
  ; except for condition (starvation). Used in habitat and activity selection.
  ; Valid values of an-activity are "hide", "drift", and "search"

  report (survival-high-temperature-for a-fish a-cell) *
;   (survival-high-velocity-for a-fish a-cell an-activity) *  REMOVED
   (survival-stranding-for a-fish a-cell) *
   (survival-fish-pred-for a-fish a-cell an-activity) *
   (survival-terr-pred-for a-fish a-cell an-activity)

end

to-report survival-high-temperature-for [ a-fish a-cell ]
  ; An observer (or trout, cell) procedure to report daily survival probability for high temperature.
  ; Temperature is a global variable.

  report table:get (table:get [cell-logistic-table] of a-cell ([trout-species] of a-fish)) "high-temperature"

end

;to-report survival-high-velocity-for [ a-fish a-cell an-activity ]
;  ; An observer (or trout, cell) procedure to report daily survival probability for high velocity.
;  ; Survival depends on swimming velocity in cell (which depends on whether shelter is available)
;  ; and on sustainable swim speed of the fish.
;
;  ; This mortality source removed from model 4/26/2021
;
;  ; High velocity survival is assumed 100% for hiding fish.
;  if an-activity = "hide" [ report 1.0 ]
;
;  ifelse (an-activity = "drift")
;  [
;    let the-ratio (drift-swim-speed-for a-fish a-cell) / (max-swim-speed-for a-fish a-cell)
;    report evaluate-logistic "high-velocity" ([trout-species] of a-fish) the-ratio
;  ]
;  [
;    ifelse (an-activity = "search")
;    [
;      let the-ratio ([velocity] of a-cell) / (max-swim-speed-for a-fish a-cell)
;      report evaluate-logistic "high-velocity" ([trout-species] of a-fish) the-ratio
;    ]
;    [
;      error (word "survival-high-velocity-for received the illegal activity " an-activity)
;    ]
;  ]
;
;end

to-report survival-stranding-for [ a-fish a-cell ]
  ; An observer (or trout, cell) procedure to report daily survival probability for stranding.

  report ifelse-value ([depth] of a-cell <= 0.0)
  [ ; value if cell is dry
    item ([trout-spp-index] of a-fish) mort-strand-survival-when-dry
  ]
  [ ; value if cell is not dry
    1.0
  ]

end

to-report survival-condition-for [ a-fish a-cell ]
  ; An observer (or trout, cell) procedure to report daily survival probability for poor condition.
  ; This survival does not depend on the cell. It implements the linear condition mortality model.
  ; The equation is algebraically equivalent to a line with survival = mort-condition-S-at-K5 at K = 0.5
  ; and survival = 1.0 at K = 1.0

  let the-condition [trout-condition] of a-fish
  if the-condition = 1.0 [ report 1.0 ]

  let s5 item ([trout-spp-index] of a-fish) mort-condition-S-at-K5
  report (2 * s5) + (2 * the-condition) - (2 * the-condition * s5) - 1

end

to-report survival-fish-pred-for [a-fish a-cell an-activity]
  ; An observer procedure to report daily survival probability for fish predation.

  let the-species [trout-species] of a-fish
  let min-surv [reach-fish-pred-min] of ([patches-reach] of a-cell)
  let hide-factor ifelse-value (an-activity = "hide")
   [ item ([trout-spp-index] of a-fish) mort-fish-pred-hiding-factor ]
   [ 0.0 ]

  let the-fish-table [trout-logistic-table] of a-fish
  let the-cell-table [table:get cell-logistic-table the-species] of a-cell
  let relative-risk
    (1.0 - (table:get the-fish-table "fish-pred-length")) *
    (1.0 - (table:get the-cell-table "fish-pred-depth")) *
    (1.0 - (table:get the-cell-table "fish-pred-density")) *
    (1.0 - (table:get the-cell-table "fish-pred-light")) *
    (1.0 - (table:get the-cell-table "fish-pred-temperature")) *
    (1.0 - hide-factor)

  report min-surv + ((1 - min-surv) * (1.0 - relative-risk))

end

to-report survival-terr-pred-for [a-fish a-cell an-activity]
  ; An observer procedure to report daily survival probability for terrestrial predation.

  let the-species [trout-species] of a-fish
  let min-surv [reach-terr-pred-min] of ([patches-reach] of a-cell)
  let in-hiding-place? (an-activity = "hide") and
   ([cell-available-hiding-places] of a-cell) >= ([trout-superind-rep] of a-fish)
  let hide-factor ifelse-value in-hiding-place?
   [ item ([trout-spp-index] of a-fish) mort-terr-pred-hiding-factor ]
   [ 0.0 ]

  let the-fish-table [trout-logistic-table] of a-fish
  let the-cell-table [table:get cell-logistic-table the-species] of a-cell
  let relative-risk
    (1.0 - (table:get the-fish-table "terr-pred-length")) *
    (1.0 - (table:get the-cell-table "terr-pred-depth")) *
    (1.0 - (table:get the-cell-table "terr-pred-velocity")) *
    (1.0 - (table:get the-cell-table "terr-pred-light")) *
    (1.0 - (table:get the-cell-table "terr-pred-dist-escape")) *
    (1.0 - hide-factor)

  report min-surv + ((1 - min-surv) * (1.0 - relative-risk))

end

to redd-survive
  ; The redd survival submodel procedure, called from the go procedure.
  ; A redd procedure that determines how many of the redd's eggs survive each time step.
  ; The order in which mortality types are evaluated is defined in the model description.

  ; Optional test output is first so it is executed even if all eggs die.

  ; Test output for redd survival.
  if debug-redd-survival?
  [
    let debug-redd-survival-file-name (word "Debug-redd-survival-" outfile-name-base)
    if not file-exists? debug-redd-survival-file-name
    [
      file-open debug-redd-survival-file-name
      file-print (word "InSTREAM-7 redd survival test output file, Created:," date-and-time  ",Values are *daily* survival rates")
      file-type "redd,species,reach,temperature,flow,depth,is-flow-peak?,scour-param,"
      file-print "s-lo-temperature,s-hi-temperature,s-dewater,s-scour"
      file-close
    ]

    file-open debug-redd-survival-file-name
    file-print csv:to-row (list
      who
      redd-species
      [reach-name] of patches-reach
      [reach-temperature] of patches-reach
      [reach-flow] of patches-reach
      depth
      [reach-is-flow-peak?] of patches-reach
      [reach-scour-param] of patches-reach
      redd-survival-lo-temp-for self patches-cell
      redd-survival-hi-temp-for self patches-cell
      redd-survival-dewater-for self patches-cell
      ifelse-value ([reach-is-flow-peak?] of patches-reach) [redd-survival-scour-for self patches-cell] [ 1.0 ]
    )
    file-close
  ]
  ; End of test output code

  ; Low temperature mortality
  let lo-temp-mortality-rate (1.0 - (redd-survival-lo-temp-for self patches-cell)) * redd-num-eggs * step-length
  let eggs-died random-poisson lo-temp-mortality-rate
  if eggs-died > 0
  [ ; Some eggs died of of low temperature
    if eggs-died > redd-num-eggs [ set eggs-died redd-num-eggs ]
    set eggs-lo-temp eggs-lo-temp  + eggs-died
    set redd-num-eggs redd-num-eggs - eggs-died
    save-event (list "eggs-died-of-low-temp" eggs-died)
    if redd-num-eggs <= 0 [redd-die]
  ]

  ; High temperature mortality
  let hi-temp-mortality-rate (1.0 - (redd-survival-hi-temp-for self patches-cell)) * redd-num-eggs * step-length
  set eggs-died random-poisson hi-temp-mortality-rate
  if eggs-died > 0
  [ ; Some eggs died of of high temperature
    if eggs-died > redd-num-eggs [ set eggs-died redd-num-eggs ]
    set eggs-hi-temp eggs-hi-temp  + eggs-died
    set redd-num-eggs redd-num-eggs - eggs-died
    save-event (list "eggs-died-of-hi-temp" eggs-died)
    if redd-num-eggs <= 0 [redd-die]
  ]

  ; Dewatering mortality
  let dewater-mortality-rate (1.0 - (redd-survival-dewater-for self patches-cell)) * redd-num-eggs * step-length
  ; For dewatering, exclude mortality if depth > 0; the Poisson distribution can return positive values when
  ; mortality rate is zero.
  if dewater-mortality-rate > 0.00000001  ; Use a small number instead of zero because of roundoff.
  [
    set eggs-died random-poisson dewater-mortality-rate
    if eggs-died > 0
    [ ; Some eggs died of of dewatering
      if eggs-died > redd-num-eggs [ set eggs-died redd-num-eggs ]
      set eggs-dewatering eggs-dewatering  + eggs-died
      set redd-num-eggs redd-num-eggs - eggs-died
      save-event (list "eggs-died-of-dewatering" eggs-died)
      if redd-num-eggs <= 0 [redd-die]
    ]
  ] ; if dewater-mortality-rate > 0.0

  ; Scour mortality
  if [reach-is-flow-peak?] of patches-reach
  [ ; Flow peaked so execute scour
    if random-float 1.0 > (redd-survival-scour-for self patches-cell)
    [ ; Redd has been scoured so all eggs die
      set eggs-scour redd-num-eggs
      save-event (list "eggs-died-of-scour" redd-num-eggs) ; Careful here! Statement is different from previous redd mortalities
      set redd-num-eggs 0
      redd-die
    ]
  ]


end

to-report redd-survival-lo-temp-for [ a-redd a-cell ]
  ; An observer (or trout, cell) procedure to report redd daily survival rate for low temperature.

  report table:get (table:get [cell-logistic-table] of a-cell ([redd-species] of a-redd)) "redd-lo-temp"

end

to-report redd-survival-hi-temp-for [ a-redd a-cell ]
  ; An observer (or trout, cell) procedure to report redd daily survival rate for high temperature.

  report table:get (table:get [cell-logistic-table] of a-cell ([redd-species] of a-redd)) "redd-hi-temp"

end

to-report redd-survival-dewater-for [ a-redd a-cell ]
  ; An observer (or trout, cell) procedure to report redd daily survival rate for dewatering.
  ; Survival is simply a parameter.

  report ifelse-value ([depth] of a-redd > 0.0)
    [ 1.0 ] ; Redd is submerged so no mortality
    [ item ([redd-spp-index] of a-redd) mort-redd-dewater-surv ] ; Depth is zero so survival is < 1.0

end

to-report redd-survival-scour-for [ a-redd a-cell ]
  ; An observer (or redd, cell) procedure to report redd's probability of surviving a scour event.
  ; It implements the scour model of Haschenberger (1999).
  ; First check to make sure the redd scour parameter was initialized in update-redd-scour-vars.
  let scour-param [reach-scour-param] of ([patches-reach] of a-cell)
  if scour-param < 0.0 [ error (word "reach-scour-param has value of " scour-param ", indicating it was not updated") ]

  let x scour-param * (item ([redd-spp-index] of a-redd) mort-redd-scour-depth)
  report ifelse-value (x < 100)  ; Round survival off to 1.0 if it is very high
  [ 1.0 - (exp ( -1.0 * x))]
  [ 1.0 ]

end

to be-superimposed-by [ the-spawner ]
  ; A redd procedure that executes superimposition mortality
  ; First, determine whether superimposition occurs

  let the-disturbed-area item ([trout-spp-index] of the-spawner) redd-area
  let gravel-area ([cell-frac-spawn] of patches-cell) * ([cell-area] of patches-cell)
  let superimposition-prob ifelse-value (gravel-area > 0.0)
  [ the-disturbed-area / gravel-area ]
  [ 0.0 ]
  if random-float 1.0 < superimposition-prob
  [ ; Superimposition occurs
    ; Determine and record how many eggs die, and determine whether the redd dies
    let eggs-died random (redd-num-eggs + 1)
    set eggs-superimp eggs-superimp + eggs-died
    set redd-num-eggs redd-num-eggs - eggs-died
    save-event (list "eggs-died-of-superimp" eggs-died)
    if redd-num-eggs <= 0 [redd-die]
  ]

end

to build-logistic-functions
  ; An observer procedure to create logistic functions used in survival and other trout functions.
  ; Logistic functions are stored in a table; they can differ among species.
  ; Create the table.
  set trout-logistics-table table:make

  ; Build a table of functions for each species.
  foreach species-list [ next-species ->
   let the-species-table table:make
   let the-spp-index position next-species species-list

   ; Capture success for drift feeding
   create-logistic-with-table-and-params the-species-table
     "capture-success" (item the-spp-index trout-capture-R1) 0.1 (item the-spp-index trout-capture-R9) 0.9

   ; Habitat selection radius
   create-logistic-with-table-and-params the-species-table
     "hab-sel-radius" (item the-spp-index trout-move-radius-L1) 0.1 (item the-spp-index trout-move-radius-L9) 0.9

   ; High temperature
   create-logistic-with-table-and-params the-species-table
     "high-temperature" (item the-spp-index mort-high-temp-T1) 0.1 (item the-spp-index mort-high-temp-T9) 0.9

   ; Terrestrial predation
   create-logistic-with-table-and-params the-species-table
     "terr-pred-length" (item the-spp-index mort-terr-pred-L1) 0.1 (item the-spp-index mort-terr-pred-L9) 0.9
   create-logistic-with-table-and-params the-species-table
     "terr-pred-depth" (item the-spp-index mort-terr-pred-D1) 0.1 (item the-spp-index mort-terr-pred-D9) 0.9
   create-logistic-with-table-and-params the-species-table
     "terr-pred-velocity" (item the-spp-index mort-terr-pred-V1) 0.1 (item the-spp-index mort-terr-pred-V9) 0.9
   create-logistic-with-table-and-params the-species-table
     "terr-pred-light" (item the-spp-index mort-terr-pred-I1) 0.1 (item the-spp-index mort-terr-pred-I9) 0.9
   create-logistic-with-table-and-params the-species-table
     "terr-pred-dist-escape" (item the-spp-index mort-terr-pred-H1) 0.1 (item the-spp-index mort-terr-pred-H9) 0.9

   ; Fish predation
   create-logistic-with-table-and-params the-species-table
     "fish-pred-length" (item the-spp-index mort-fish-pred-L1) 0.1 (item the-spp-index mort-fish-pred-L9) 0.9
   create-logistic-with-table-and-params the-species-table
     "fish-pred-depth" (item the-spp-index mort-fish-pred-D1) 0.1 (item the-spp-index mort-fish-pred-D9) 0.9
   create-logistic-with-table-and-params the-species-table
     "fish-pred-density" (item the-spp-index mort-fish-pred-P1) 0.1 (item the-spp-index mort-fish-pred-P9) 0.9
   create-logistic-with-table-and-params the-species-table
     "fish-pred-light" (item the-spp-index mort-fish-pred-I1) 0.1 (item the-spp-index mort-fish-pred-I9) 0.9
   create-logistic-with-table-and-params the-species-table
     "fish-pred-temperature" (item the-spp-index mort-fish-pred-T1) 0.1 (item the-spp-index mort-fish-pred-T9) 0.9

   ; Redd mortality
   create-logistic-with-table-and-params the-species-table
    "redd-lo-temp" (item the-spp-index mort-redd-lo-temp-T1) 0.1 (item the-spp-index mort-redd-lo-temp-T9) 0.9
   create-logistic-with-table-and-params the-species-table
    "redd-hi-temp" (item the-spp-index mort-redd-hi-temp-T1) 0.1 (item the-spp-index mort-redd-hi-temp-T9) 0.9

   table:put trout-logistics-table next-species the-species-table
  ]

  if debug-survival-logistics?
  [ ; Write debug output. This is done only once.
    let survival-logistics-file-name (word "Debug-survival-logistics-" outfile-name-base)

    ; Create a list of lists that will be output, add header info
    let out-list (list (list "Debug output for survival logistic functions"))
    set out-list lput (list "Created:" date-and-time) out-list
    set out-list lput (list "Logistic function values for inputs from 0.0 to 50.0") out-list

    ; Create a table of logistic function output, for inputs from 0 to 10.
    ; A separate table is produced for each species.

    let a-species-list table:keys trout-logistics-table
    foreach a-species-list [ next-species ->
      set out-list lput (list) out-list                     ; Output a blank line
      set out-list lput (list next-species) out-list        ; Output the species name
      let a-function-list table:keys (table:get trout-logistics-table next-species)
      set out-list lput (sentence "Input" a-function-list) out-list
      let the-input 0
      while [the-input <= 50]
      [
        let row-list (list the-input)
        foreach a-function-list [ next-function ->
          set row-list lput (evaluate-logistic next-function next-species the-input) row-list
        ]
        set the-input the-input + 1
        set out-list lput row-list out-list
      ]  ; while the-input
    ]    ; foreach a-species-list

    csv:to-file survival-logistics-file-name out-list
  ] ; if debug-survival-logistics?

end

to create-logistic-with-table-and-params [a-table a-name x1 p1 x2 p2]
  ; An observer reporter to initialize a logistic function.
  ; p1 and p2 are the probabilities associated with inputs x1 and x2

  ; First some defensive programming
  if x1 = x2 [ error (word "Tried to create logistic function " a-name " with X1 = X2 = " x1) ]
  if p1 = p2 [ error (word "Tried to create logistic function " a-name " with P1 = P2 = " p1) ]

  ; For convenience in test output, make sure X1 is less than X2
  if x1 > x2
  [
    let a-num x2
    set x2 x1
    set x1 a-num
    set a-num p2
    set p2 p1
    set p1 a-num
  ]

  ; Calculate the internal parameters using method on p. 216 of Railsback & Grimm 2012
  let D ln (p1 / (1 - p1))
  let C ln (p2 / (1 - p2))
  let logistic-b (D - C) / (x1 - x2)
  let logistic-a D - (logistic-b * x1)

  let a-subtable table:make
  table:put a-subtable "logistic-b" logistic-b
  table:put a-subtable "logistic-a" logistic-a

  table:put a-table a-name a-subtable

end

to-report evaluate-logistic [ a-logistic-name a-species an-input ]
  ; An observer (trout, cell) reporter to report the value of a survival logistic
  ; It is ~25% faster to get the logistic table first ("let the-table") than to use:
  ; let the-logistic-a table:get (table:get (table:get trout-logistics-table a-species) a-logistic-name) "logistic-a"
  ; let the-logistic-b table:get (table:get (table:get trout-logistics-table a-species) a-logistic-name) "logistic-b"

  let the-table (table:get (table:get trout-logistics-table a-species) a-logistic-name)
  let Z (table:get the-table "logistic-a") + ((table:get the-table "logistic-b") * an-input)
  ; Defensive programming to avoid over/underflow runtime errors
  if Z < -200 [ report 0.0 ]
  if Z >   35 [ report 1.0 ]

  report (exp Z) / (1.0 + exp Z)   ; Calculating exp(Z) twice does not slow execution -- really.

end

;to-report evaluate-logistic-from-to-over-time [ a-logistic-name a-species a-start-input an-end-input ]
;  ; An observer (trout, cell) reporter to report the average value of a logistic function
;  ; as the input changes from one value to a second, as in calculating starvation survival
;  ; over a time horizon. This reporter does *not* factor in the time between start to end;
;  ; to get survival as (for example) condition changes from K1 to K2, raise the result of this
;  ; reporter to the power of the number of days between t1 and t2.
;
;  ; If the starting input almost equals the ending input,
;  ; then the logistic function can be used without averaging (and must be to avoid division by zero).
;  if abs ((a-start-input - an-end-input) / a-start-input ) < 0.01
;  [ report evaluate-logistic a-logistic-name a-species a-start-input ]
;
;  let the-table (table:get (table:get trout-logistics-table a-species) a-logistic-name)
;  let the-a (table:get the-table "logistic-a")
;  let the-b (table:get the-table "logistic-b")
;
;  let inner-frac (1.0 + exp (the-b * an-end-input + the-a)) / (1.0 + exp (the-b * a-start-input + the-a))
;  report (1.0 / the-b) * ln (inner-frac) / (an-end-input - a-start-input)
;
;end

to spawn
  ; A trout procedure for spawning

  ; Test output for spawning readiness.
  if debug-spawning-readiness?
  [
    let the-input [reach-time-series-input] of patches-reach
    let ready-output-file-name (word "Debug-spawn-ready-" outfile-name-base)
    if not file-exists? ready-output-file-name
    [
      file-open ready-output-file-name
      file-print (word "InSTREAM-7 spawning readiness test output file, Created " date-and-time)
      file-type "date,phase,reach,trout-species,trout-age,trout-length,trout-condition,spawned-this-year,min-temperature,max-temperature,"
      file-print "max-flow-1-day,min-flow-2-day,max-flow-2-day,ready?,reason"
      file-close
    ]

    let list-of-temps ts-get-range the-input (time:plus sim-time -1 "days") sim-time "temperature"
    if empty? list-of-temps [ set list-of-temps (list [reach-temperature] of patches-reach)]
    let list-of-1-day-flows ts-get-range the-input (time:plus sim-time -1 "days") sim-time "flow"
    if empty? list-of-1-day-flows [ set list-of-1-day-flows (list [reach-flow] of patches-reach)]
    let list-of-2-day-flows ts-get-range the-input (time:plus sim-time -2 "days") sim-time "flow"
    if empty? list-of-2-day-flows [ set list-of-2-day-flows (list [reach-flow] of patches-reach)]

    file-open ready-output-file-name
    file-print csv:to-row (list
      formatted-sim-time
      light-phase
      [reach-name] of patches-reach
      trout-species
      trout-age
      trout-length
      trout-condition
      trout-spawned-this-season?
      min list-of-temps
      max list-of-temps
      max list-of-1-day-flows
      min list-of-2-day-flows
      max list-of-2-day-flows
      ready-to-spawn?
      why-not-spawn
      )
  ]
  ; End of test output code

  ; First, determine if the trout spawns, which depends on meeting all the criteria and a random draw
  if ready-to-spawn? and (random-float 1.0 < (item trout-spp-index trout-spawn-prob))
  [ ; now spawn
    ; First, move to spawning cell, which is identified by reporter spawn-cell
    move-to one-of [cells-patches] of spawn-cell

    ; Second, create a redd and initialize it
    hatch-redds 1
    [
      set redd-species [trout-species] of myself              ; The redd's species (same as trout species)
      set redd-spp-index [trout-spp-index] of myself          ; The redd's species index (same as trout species)
      set redd-num-eggs  [num-viable-eggs] of myself          ; The number of viable eggs in the redd (integer reporter)
      set redd-frac-developed  0.0                            ; Fraction of fully developed (0.0 to 1.0)
      set redd-emerge-days 0                                  ; The number of days on which the redd has been producing new trout
      set shape "target"
      save-event (list "created" redd-num-eggs)

      set redd-date-created time:show sim-time "MM-dd-yyyy"
      set eggs-initial redd-num-eggs
      set eggs-lo-temp 0
      set eggs-hi-temp 0
      set eggs-dewatering 0
      set eggs-scour 0
      set eggs-superimp 0
      set eggs-emerged 0
    ]

    ; Cause superimposition
    let other-redds-here other (redds with [patches-cell = [patches-cell] of myself])
    if any? other-redds-here [ask other-redds-here [ be-superimposed-by myself]]

    ; select a male mate and make it spawn
    let potential-mates trout with
    [
      trout-sex = "male" and
      trout-species = [trout-species] of myself and
      patches-reach = [patches-reach] of myself and
      trout-length >= (item trout-spp-index trout-spawn-min-length) and
      trout-age >= (item trout-spp-index trout-spawn-min-age) and
      not trout-spawned-this-season?
    ]
    if any? potential-mates
    [
      ask max-one-of potential-mates [trout-length]
      [
        set trout-spawned-this-season? true
        set trout-weight trout-weight * (1.0 - (item trout-spp-index trout-spawn-wt-loss-fraction))
      ] ; ask mate
    ] ; if any

    ; incur weight loss and finish
    set trout-weight trout-weight * (1.0 - (item trout-spp-index trout-spawn-wt-loss-fraction))
    set trout-spawned-this-season? true
    save-event "spawned"
  ]

end

to-report ready-to-spawn?
  ; A trout reporter that reports true only if *all* criteria for spawning are met
  ; The go procedure executes this only for females

  if trout-sex != "female" [ error (word "Illegal sex " trout-sex " in ready-to-spawn?") ]

  ; First, minimum age, length, and condition criteria
  if trout-age       < (item trout-spp-index trout-spawn-min-age)    [ set why-not-spawn "age"       report false ]
  if trout-length    < (item trout-spp-index trout-spawn-min-length) [ set why-not-spawn "length"    report false ]
  if trout-condition < (item trout-spp-index trout-spawn-min-cond)   [ set why-not-spawn "condition" report false ]

  ; Second, not spawned this season
  if trout-spawned-this-season? [ set why-not-spawn "spawned this season" report false ]

  ; Third, date window
  ; Unfortunately time:is-between does not currently handle day value ranges that span 1/1
  let spawn-start-day time:copy (item trout-spp-index trout-spawn-start-day)
  let spawn-end-day   time:copy (item trout-spp-index trout-spawn-end-day)
  let this-day        time:create (time:show sim-time "MM-dd")
  ifelse time:is-before? spawn-start-day spawn-end-day
  [ ; spawning season starts and ends in the same calendar year
    if not time:is-between? this-day spawn-start-day spawn-end-day    [ set why-not-spawn "date"  report false ]
  ]
  [ ; spawning season starts before 1/1 and ends after it
    if (time:is-after? this-day spawn-end-day) and (time:is-before? this-day spawn-start-day)
      [ set why-not-spawn "date"  report false ]
  ]

  ; Fourth, temperature
  ; Use mean of all input temperatures over previous 1.0 day and current day
  let the-input [reach-time-series-input] of patches-reach
  let list-of-temps ts-get-range the-input (time:plus sim-time -1 "days") sim-time "temperature"
  if empty? list-of-temps [ set list-of-temps (list [reach-temperature] of patches-reach)]
  if mean list-of-temps < (item trout-spp-index trout-spawn-min-temp) [ set why-not-spawn "min temp" report false ]
  if mean list-of-temps > (item trout-spp-index trout-spawn-max-temp) [ set why-not-spawn "max temp" report false ]

  ; Fifth, flow limit
  let list-of-flows ts-get-range the-input (time:plus sim-time -1 "days") sim-time "flow"
  if empty? list-of-flows [ set list-of-flows (list [reach-flow] of patches-reach)]
  if max list-of-flows > ([reach-max-spawn-flow] of patches-reach) [ set why-not-spawn "max flow" report false ]

  ; Sixth, steady flows
  set list-of-flows ts-get-range the-input (time:plus sim-time -2 "days") sim-time "flow"
  if length list-of-flows > 1  ; Skip this criterion if there are not at least 2 flow inputs in 2 days
  [
    let q-min min list-of-flows
    let frac-flow-change abs (((max list-of-flows) - q-min) / q-min)
    if frac-flow-change > (item trout-spp-index trout-spawn-max-flow-change) [ set why-not-spawn "unsteady flow" report false ]
  ]

  ; All criteria are met, so report true
  set why-not-spawn "can spawn"
  report true

end

to-report spawn-cell
  ; A trout procedure to identify the cell for spawning
  ; First, identify which cells to consider, using exactly the same method as habitat selection

    ; Convert radius from cm to patch-widths
    ; Remember that the fish doing this may not be exactly at its cell while
    ; distances are measured from the cell center; so "cells in-radius" does not work
    let my-cell patches-cell
    let hab-sel-radius (item trout-spp-index trout-move-radius-max) * (table:get trout-logistic-table "hab-sel-radius") ; in cm
    ; In following statement, "distance my-cell" is used instead of "in-radius" because
    ; distance must be measured from the cell, not the trout.
    let dest-cells cells with [(distance my-cell) <= (hab-sel-radius / world-resolution )]

    ; If necessary, add adjacent cells to destinations
    if hab-sel-radius < [max-dist-to-adjacent-cell] of my-cell
    [ set dest-cells (turtle-set dest-cells ([adjacent-cells] of my-cell)) ]

    ; Exclude dry cells
    set dest-cells dest-cells with [depth > 0]

  ; Exclude cells with gravel area less than the species' redd area, if there are any
  let the-redd-area (item trout-spp-index redd-area)
  let suitable-cells dest-cells with [(cell-area * cell-frac-spawn) >= the-redd-area]
  if not any? suitable-cells [ set suitable-cells dest-cells ]

  ; Calculate maximum spawning suitability of cells
  let the-species trout-species  ; Create a local variable for use by cells below
  let best-suitability max
  [
    (interpolate spawn-depth-interpolation-table the-species depth) *
    (interpolate spawn-vel-interpolation-table the-species velocity)
  ] of suitable-cells

  ifelse best-suitability <= 0.0  ; Cope with possibility that no cells have suitability > 0
  [
    report my-cell  ; If no suitable cells, stay in current cell
  ]
  [
    ; Identify cells with suitability within acceptable range
    let min-suitability (1.0 - (item trout-spp-index trout-spawn-suitability-tol)) * best-suitability
    set suitable-cells suitable-cells with
    [
      ((interpolate spawn-depth-interpolation-table the-species depth) *
      (interpolate spawn-vel-interpolation-table the-species velocity)) > min-suitability
    ]
  ]

  ; Finally, pick and report one of the suitable cells randomly, weighted by gravel area
  ; This will be an unweighted draw if no cells have gravel
  report rnd:weighted-one-of suitable-cells [ cell-area * cell-frac-spawn ]

end

to-report num-viable-eggs
  ; A trout procedure to calculate the number of viable eggs a female lays

  report round ((item trout-spp-index trout-spawn-fecund-mult) *
    (trout-length ^ (item trout-spp-index trout-spawn-fecund-exp)) *
    (item trout-spp-index trout-spawn-egg-viability))

end

to develop-eggs
  ; A redd procedure that updates development status
  let temperature [reach-temperature] of patches-reach
  set redd-frac-developed redd-frac-developed +
    (step-length * ((item redd-spp-index redd-devel-A) +
      ((item redd-spp-index redd-devel-B) * temperature) +
      ((item redd-spp-index redd-devel-C) * temperature * temperature)))

end

to redd-emergence
  ; A redd procedure that models emergence of new trout
  ; The go procedure calls this only for redds with redd-frac-developed >= 1.0
  ; and only once per day

  ; First, record when emergence starts
  if redd-emerge-days = 0 [ save-event "started emergence" ]

  ; Update days of emergence
  set redd-emerge-days redd-emerge-days + 1

  ; Determine how many trout to create and reduce the number of remaining eggs
  let num-eggs-to-emerge ceiling (redd-num-eggs * (redd-emerge-days / 10))
  set redd-num-eggs redd-num-eggs - num-eggs-to-emerge
  set eggs-emerged eggs-emerged + num-eggs-to-emerge
  save-event (list "eggs emerged" num-eggs-to-emerge) ; Un-comment this to output number of eggs emerged per day

  ; Consider superindividuals in determining how many trout to create
  let num-new-trout num-eggs-to-emerge
  let last-trout-rep 0  ; superindividual size of last trout created
  let superind-rep (item redd-spp-index trout-superind-max-rep)

  if superind-rep > 1
  [
    set last-trout-rep remainder num-new-trout superind-rep ; This statement must be before the 2nd
    set num-new-trout floor (num-new-trout / superind-rep)
  ]

  ; Create the trout that have superindividual size = maximum
  let my-species redd-species
  let my-index redd-spp-index
  hatch-trout num-new-trout
  [
    let the-length random-triangular (item my-index trout-emerge-length-min)
      (item my-index trout-emerge-length-mode)
      (item my-index trout-emerge-length-max)

    initialize-trout-with my-species 0 the-length superind-rep
    save-event "emerged"
    move-to one-of [cells-patches] of patches-cell
  ]

  ; Now create the last superindividual, which represents fewer individuals
  if last-trout-rep > 0
  [
    hatch-trout 1
    [
      let the-length random-triangular (item my-index trout-emerge-length-min)
        (item my-index trout-emerge-length-mode)
        (item my-index trout-emerge-length-max)

      initialize-trout-with my-species 0 the-length last-trout-rep
      save-event "emerged"
      move-to one-of [cells-patches] of patches-cell
    ]
  ]

  ; If all eggs are gone, then write output and die -- THIS MUST BE LAST
  if redd-num-eggs < 1 [ redd-die ]

end

to redd-die
  ; A redd procedure to write file output and die

  set redd-date-emptied time:show sim-time "MM-dd-yyyy"

  set redd-summary-list lput (csv:to-row (list behaviorspace-run-number
    who
    redd-species
    [reach-name] of patches-reach
    patches-cells-ID
    redd-date-created
    redd-date-emptied
    eggs-initial
    eggs-lo-temp
    eggs-hi-temp
    eggs-dewatering
    eggs-scour
    eggs-superimp
    eggs-emerged
    )) redd-summary-list

  save-event "emptied"

  die

end

to separate-superindividual
  ; A trout procedure to separate a superindividual into new individual trout
  hatch trout-superind-rep
  [
    ; New individuals inherit all characteristics of the superindividual except these.
    set trout-superind-rep 1
    set trout-tracked? (random-float 1.0 < frac-trout-to-output)  ; This reports TRUE or FALSE
    save-event "Separated from superindividual"
  ]

  ; Record the event and die
  save-event "Superindividual separated"
  die

end

to-report random-triangular [a-min a-mode a-max]
  ; Return a random value from a triangular distribution
  ; Method from https://en.wikipedia.org/wiki/Triangular_distribution#Generating_Triangular-distributed_random_variates
  ; Obtained 2015-11-27

  if (a-min > a-mode) or (a-mode > a-max) or (a-min >= a-max)
   [ error (word "Random-triangular received illegal parameters (min, mode, max): " a-min " " a-mode " " a-max) ]

  let a-rand random-float 1.0
  let F (a-mode - a-min) / (a-max - a-min)

  ifelse a-rand < F
  [ report a-min + sqrt (a-rand * (a-max - a-min) * (a-mode - a-min)) ]
  [ report a-max - sqrt ((1 - a-rand) * (a-max - a-min) * (a-max - a-mode)) ]

end

to show-GIS-properties
  ; Observer procedure called from Interface button to see what is in the GIS file
  let GIS-dataset gis:load-dataset GIS-file-name
  let the-property-names gis:property-names GIS-dataset
  let some-polygons n-of 5 gis:feature-list-of GIS-dataset

  foreach the-property-names
  [ the-property ->
    show (word the-property " " (gis:property-value (item 0 some-polygons) the-property)
     " " (gis:property-value (item 1 some-polygons) the-property)
     " " (gis:property-value (item 2 some-polygons) the-property)
     " " (gis:property-value (item 3 some-polygons) the-property)
     " " (gis:property-value (item 4 some-polygons) the-property))
  ]

end

to shade-reaches-patches
  ; Reach procedure to color the reach's patches by patch variables

  ; Skip shading if desired; this can save a lot of execution time
  if shade-variable = "off" [ stop ]

  let cell-color green  ; local variable re-used to color each patch

  ask reaches-cells
  [
    set cell-color green  ; the color for dry cells
    if depth > 0  ; cell agent gets depth and vel from patch it is on
    [ ; Cell is below water so shade it by depth or velocity
      ; Velocity is shaded yellow-red
      ifelse shade-variable = "depth"
      [ set cell-color scale-color blue depth max-shade-depth  0]
      [
        ifelse shade-variable = "velocity"
        [ set cell-color (list 255 (255 * (1 - (min (list velocity max-shade-velocity) / max-shade-velocity))) 0)]
        [ ifelse shade-variable = "light"
          [ set cell-color scale-color grey light 0 max-shade-light ] ; This makes patches white during day, grey in twilight.
          [ error (word "Illegal shade variable: " shade-variable) ]
        ]
      ]
    ]

      ask cells-patches [ set pcolor cell-color ]
  ]

  ; Shading patches outside the channel according to light intensity does not work
  ; because this procedure is only executed when flow (not light) changes when
  ; shade-variable is not set to "light".

end

to shade-by-depth
  ; Observer procedure to change patch shading to depth, called by interface button
  clear-drawing  ; Get rid of the background image if there was one; it hides patch colors
  gis:draw polygons-dataset 1  ; Re-draw the cell boundaries
  set shade-variable "depth"
  ask reaches [shade-reaches-patches]  ; Update the shading immediately
end

to shade-by-vel
  ; Observer procedure to change patch shading to velocity, called by interface button
  clear-drawing  ; Get rid of the background image if there was one; it hides patch colors
  gis:draw polygons-dataset 1  ; Re-draw the cell boundaries
  set shade-variable "velocity"
  ask reaches [shade-reaches-patches]  ; Update the shading immediately
end

to shade-by-light
  ; Observer procedure to change patch shading to light, called by interface button
  clear-drawing  ; Get rid of the background image if there was one; it hides patch colors
  gis:draw polygons-dataset 1  ; Re-draw the cell boundaries
  set shade-variable "light"
  ask reaches [shade-reaches-patches]  ; Update the shading immediately so this works when execution is paused
end

to stop-shading
  ; Observer procedure to stop shading and display background image, called by interface button
  ; Display the background image if there is one
  ask patches [set pcolor black]
  if (background-image-file != "None") and (file-exists? background-image-file)
  [ import-drawing background-image-file ]
  gis:draw polygons-dataset 1  ; Re-draw the cell boundaries
  set shade-variable "off"
end

to check-for-missing-input
  ; Reach procedure to check time-series input for missing data--when the
  ; time between input rows exceeds the value of parameter max-days-between-time-series-inputs
  ; This must be executed *after* time is set up and sim-time initialized
  show "Checking time-series input for missing values"
  let input-copy but-first reach-time-series-input

  ; Zeroeth, check variable names, which are in the first item of the time series input list
  let var-names first reach-time-series-input
  if (position "flow" var-names = false) or (position "temperature" var-names = false) or (position "turbidity" var-names = false)
  [ error (word "Time series input file " reach-time-series-input-file " does not have legal column labels. "
    "They must be exactly flow temperature turbidity, all lower case, no quotation marks.") ]

  ; First, ignore input before model start time
  while [time:is-before? (first (first input-copy)) sim-time]
  [
    set input-copy but-first input-copy
  ]

  ; Create a list of problem dates
  let problem-list (list)

  ; Now check values until end time
  while [(time:difference-between (first (first input-copy)) sim-end-time "days") > max-days-between-time-series-inputs]
  [
    if (time:difference-between (first (first input-copy)) (first (item 1 input-copy)) "days") > max-days-between-time-series-inputs
    [
      set problem-list lput (first (first input-copy)) problem-list
    ]
    set input-copy but-first input-copy
  ]

  ; Check for and report any problem dates
  if not empty? problem-list
  [
    foreach problem-list [next-time ->
      show (word "Input file " reach-time-series-input-file " has insufficient records after: " (time:show next-time "M/d/yyyy HH:mm"))
    ]
    user-message (word "Missing time series input for reach " reach-name ". See Command Center for details.")
  ]

end

;; NOTE: the following ts- procedures are copied from the file time-series.nls distributed
;; with the Time extension for NetLogo 6.1.1
;; They should not be modified.

;; Implementation of what were primitives from the time-extension time-series tool
to-report ts-create [col-name-list]
  report  (list sentence "LOGOTIME" col-name-list)
end

to-report ts-add-row [ltime-series row]

  (ifelse
    length ltime-series = 1 or time:is-after? (__time-of row) (__time-of last ltime-series) [
      report lput (__row-copy row) ltime-series
    ]
    time:is-before? (__time-of row) (__time-of item 1 ltime-series) [
      report insert-item 1 ltime-series (__row-copy row)
    ]
    [ ;else
      let b&a __indices-before-and-after ltime-series (__time-of row)
      report insert-item (last b&a) ltime-series (__row-copy row)
    ]
  )
end

to-report __row-copy [row]
  report fput (time:copy first row) (butfirst row)
end

to-report ts-get [ltime-series logotime col-name]
  let b&a ts-get-before-and-after ltime-series logotime
  let columns (item 0 ltime-series)
  (ifelse
    first b&a = nobody [report __col-of-row columns (last b&a) col-name] ; need this explicitly so we don't try to get __time-of first b&a iin the next conditional when first b&a is nobody
    last b&a = nobody or
    (abs(time:difference-between logotime (__time-of first b&a) "seconds") <= abs(time:difference-between logotime (__time-of last b&a) "seconds"))[
      report __col-of-row columns (first b&a) col-name   ; logotime is closer to the item before than after
    ]  [ ; else
    report __col-of-row columns (last b&a) col-name
    ]
  )
end

to-report ts-get-before-and-after [ltime-series logotime]
  let b&a __indices-before-and-after ltime-series logotime
  let before ifelse-value (first b&a = nobody) [nobody] [item (first b&a) ltime-series]
  let after ifelse-value (last b&a = nobody) [nobody] [item (last b&a) ltime-series]
  report list before after
end

to-report ts-get-interp [ltime-series logotime col-name]
  ;; if col-name is a specific column, return just the linearly interpolated value
  ;; if col-name is "all" then return a whole interpolated row, including an interpolated logotime

  let before-and-after ts-get-before-and-after ltime-series logotime
  let col-names first ltime-series

  ifelse col-name = "all" or col-name = "ALL" [
    report fput logotime map [col -> __interp before-and-after logotime col-names col] butfirst col-names
  ] [
    report __interp before-and-after logotime col-names col-name
  ]
end

to-report ts-get-exact [ltime-series logotime col-name]
  let closest-row ts-get ltime-series logotime "all"
  ifelse time:is-equal? (item 0 closest-row) logotime [
    report __col-of-row (item 0 ltime-series) closest-row col-name
  ] [
    error word "There is no row in this timeseries with exactly the time " logotime
  ]

end

to-report ts-get-range [ltime-series logotime1 logotime2 col-name]
  ; reports a subset of the time-series between the two logotimes.

; reverse the times if logotime2 is before logotime1
  if time:is-before? logotime2 logotime1 [
    let temp1 logotime1
    set logotime1 logotime2
    set logotime2 temp1
  ]

  let max-index (length ltime-series)

  ; get the start index of the range
  let b&a-t1 __indices-before-and-after ltime-series logotime1
  let start-index nobody
  ifelse first b&a-t1 = last b&a-t1 [ ; if b&a are the same, there was an exact match. So, need to get the index of the first row with that time
    set start-index first b&a-t1
    while [start-index > 1 and time:is-equal? logotime1 (__time-of item (start-index - 1) ltime-series)] [set start-index start-index - 1]
  ] [ ; if there isn't an exact match, then the after index will be the start of the returned range.
    set start-index __after-index (last b&a-t1) max-index
  ]

  ; get the end index of the range
  let b&a-t2 __indices-before-and-after ltime-series logotime2
  let end-index nobody
  ifelse (first b&a-t2 = last b&a-t2) [ ; if b&a are the same, there was an exact match. So need to get the index of the first row after that time.
    set end-index last b&a-t2
    while [end-index < max-index and time:is-equal? logotime2 (__time-of item end-index ltime-series)] [set end-index end-index + 1]
  ] [ ; if there isn't an exact match, then the after index will be the end of the returned range
    set end-index __after-index (last b&a-t2) max-index
  ]

  ; report the columns of the sublist between aftere-t1 and after-t2
  let columns (item 0 ltime-series)
  report map [row -> __col-of-row columns row col-name] sublist ltime-series start-index end-index
end

to-report ts-load [filepath]
  report ts-load-with-format filepath nobody
end

to-report ts-load-with-format [filepath format-string]
  let temp-ts __remove-comment-lines (csv:from-file filepath ",")

  let columns fput "LOGOTIME" butfirst item 0 temp-ts ; replace the first column with "LOGOTIME"

  report fput columns __sort-by-date map [ row ->
    __convert-row-with-format row format-string
  ] (butfirst temp-ts)
end

to-report ts-has-repeat-times? [ltime-series]
  foreach (range 1 (length ltime-series - 1)) [i ->
    if time:is-equal? (__time-of item i ltime-series) (__time-of item (i + 1) ltime-series) [report true]]
  report false
end

to-report ts-has-repeat-of-time? [ltime-series logotime]
  let b&a __indices-before-and-after ltime-series logotime
  let i1 first b&a
  let i2 last b&a
  ; if the indices are the same, that just means there was an exact match -> still need to check above and below
  ifelse i1 = i2 [
    (ifelse
      i1 > 1 and time:is-equal? (__time-of item (i1 - 1) ltime-series) (__time-of item i1 ltime-series) [
        report true
      ]
      i2 < (length ltime-series - 1) and time:is-equal? (__time-of item i2 ltime-series) (__time-of item (i2 + 1) ltime-series) [
        report true
      ]
      [report false]
    )
  ] [
    ; if the i1 != i2, there isn't even an exact match, let alone a repeat.
    report false
  ]
end

to ts-write [ltime-series filepath]
  csv:to-file filepath fput (first ltime-series) map [row ->
    fput (__show-default (first row)) (butfirst row)]
    butfirst ltime-series
end

;**********************************
;********Helper Procedures for ts- procedures *********
;**********************************
to-report __interp [before-and-after logotime col-names col-name]
  let before-val __col-of-row col-names (item 0 before-and-after) col-name
  let after-val __col-of-row col-names (item 1 before-and-after) col-name

  (ifelse
    before-val = after-val [
      report before-val]
    before-val = nobody [
      report after-val]
    after-val = nobody [
      report before-val]
    [ ; else
      let before-time item 0 item 0 before-and-after
      let after-time item 0 item 1 before-and-after
      let before-after-time-diff (time:difference-between before-time after-time "seconds")
      let slope (after-val - before-val) / before-after-time-diff
      let before-logotime-diff (time:difference-between before-time logotime "seconds")

      report before-val + slope * before-logotime-diff
    ]
  )
end

to-report __indices-before-and-after [timeseries logotime]
  ; return the indexes of the rows in timeseries that have logotimes above and below the input logotime.
  ; if there is an exact match, return the index of the exact row twice.
  ; the column names are included so that you can just use the index on the timeseries list and not butfirst timeseries list.
  ; This is equivalent to treating ltime-series objects here as starting at index 1
  let i round ((length timeseries) / 2)
  let max-i length timeseries - 1
  let min-i 1
  while [true] [
    (ifelse
      time:is-equal? logotime (__time-of item i timeseries) [report list i i]
      (time:is-before? logotime (__time-of item i timeseries)) and i = 1 [report list nobody i]
      (time:is-after? logotime (__time-of item i timeseries)) and i = (length timeseries - 1) [report list i nobody]
      (time:is-after? logotime (__time-of item i timeseries)) and (time:is-before? logotime (__time-of item (i + 1) timeseries)) [report list i (i + 1)]
      time:is-before? logotime (__time-of item i timeseries) [
        set max-i i
        set i floor ((i + min-i) / 2)
      ]
      __is-after-or-equal? logotime (__time-of item (i + 1) timeseries) [
        set min-i i
        set i ceiling ((i + max-i) / 2)
      ]
    )
  ]
end

to-report __time-of [row]
  report item 0 row
end

to-report __sort-by-date [timeseries]
  report sort-by [[row1 row2] -> time:is-before? (item 0 row1) (item 0 row2)] timeseries
end

to-report __remove-comment-lines [data]
  while [(first (first (first data)) = ";")]  ;; remove lines that start with a semicolon
      [set data but-first data]
  report data
end

to-report __convert-row-with-format [row format-string] ; helper
  let new-time ifelse-value format-string = nobody [time:create item 0 row] [time:create-with-format (item 0 row) format-string]
  ;report fput new-time (map [i -> ifelse-value is-string? i [read-from-string i] [i]] butfirst row)
  report fput new-time butfirst row
end

to-report __convert-row [row] ; helper
  report __convert-row-with-format row nobody
end

to-report __col-of-row [columns row col-name]
  (ifelse
    col-name = "all" or col-name = "ALL" [report row]
    col-name = "LOGOTIME" or col-name = "logotime" [report item 0 row]
    [ ;else
      let pos position col-name columns
      if pos = false [error (word "'" col-name "' is not a column in this time series")]
      report item pos row
    ]
  )
end

to-report __is-before-or-equal? [logotime1 logotime2]
  report time:is-before? logotime1 logotime2 or time:is-equal? logotime1 logotime2
end

to-report __is-after-or-equal? [logotime1 logotime2]
  report time:is-after? logotime1 logotime2 or time:is-equal? logotime1 logotime2
end

to-report __logotime-variety [logotime]
  let variety nobody
  carefully [
    let temp time:is-equal? logotime time:create "2000-12-01 1"
    set variety "DATETIME"
  ] [
    carefully [
      let temp time:is-equal? logotime time:create "2000-12-01"
      set variety "DATE"
    ] [
      let temp time:is-equal? logotime time:create "12-01"
      set variety "DAY"
    ]
  ]
  report variety
end

to-report __show-default [logotime]
  ; Like time:show, but it shows the default format depending on the variety
  ; of the logotime .
  let variety __logotime-variety logotime
  (ifelse
    variety = "DATETIME" [report time:show logotime "yyyy-MM-dd HH:mm:ss.SSS"]
    variety = "DATE" [report time:show logotime "yyyy-MM-dd"]
    variety = "DAY" [report time:show logotime "MM-dd"]
  )
end

to-report __after-index [after-i max-index]
  ; if the after-index exists, report it, otherwise report the max-index
  report ifelse-value after-i = nobody [max-index] [after-i]
end

;; END OF TIME-SERIES PROCEDURES

to build-output-file-named [a-file-name]
  ; An observer procedure to initialize an output file.
  ; The parameter a-file-name is the global variable for the file name.

  ; Create the brief population output file.
  if a-file-name = "b-p-o-n" ; This is the value of the uninitialized file name
  [
    set brief-pop-outfile-name (word "BriefPopOut-" outfile-name-base)
    if file-exists? brief-pop-outfile-name [ file-delete brief-pop-outfile-name ]
    file-open brief-pop-outfile-name
    file-print (word "InSTREAM-7 brief population output file, Created " date-and-time)
    file-print "BehavSp-Run,End of time step,IsCensus?,Light phase,Reach,Flow,Temperature,Turbidity,Species,Age class,Count,Mean length,Mean weight,Mean condition,FractionDriftFeeding,FractionSearchFeeding,FractionHiding"
    file-close
  ]

  ; Create the detailed population output file.
  if a-file-name = "d-p-o-n" ; This is the value of the uninitialized file name
  [
    set detailed-pop-outfile-name (word "DetailedPopOut-" outfile-name-base)
    if file-exists? detailed-pop-outfile-name [ file-delete detailed-pop-outfile-name ]
    file-open detailed-pop-outfile-name
    file-print (word "InSTREAM-7 detailed population output file, Created " date-and-time)
    file-print "BehavSp-Run,End of time step,IsCensus?,Light phase,Reach,Flow,Temperature,Turbidity,Species,Age class,Activity,Count,Mean length,Mean weight,Mean condition"
    file-close
  ]

  ; Create the fish events output file.
  if a-file-name = "f-e-o-n" ; This is the value of the uninitialized file name
  [
    set fish-events-outfile-name (word "FishEventsOut-" outfile-name-base)
    if file-exists? fish-events-outfile-name [ file-delete fish-events-outfile-name ]
    ; There can be events on the fish-events-list when file is created (from fish initialization)
    ; so these header lines must be put at the *start* of the list. Use fput with header
    ; lines in reverse order.
    set fish-events-list fput "BehavSp-Run,End of time step,Light phase,Reach,Species,ID,Cell,Age,Length,Weight,Condition,Superind-rep,Event" fish-events-list
    set fish-events-list fput (word "InSTREAM-7 fish events output file, Created " date-and-time) fish-events-list
  ]

  ; Create the redd events output file.
  if a-file-name = "r-e-o-n" ; This is the value of the uninitialized file name
  [ ; See comment above for fish-events file
    set redd-events-outfile-name (word "ReddEventsOut-" outfile-name-base)
    if file-exists? redd-events-outfile-name [ file-delete redd-events-outfile-name ]
    set redd-events-list fput "BehavSp-Run,End of time step,Light phase,Reach,Species,ID,Cell,Event" redd-events-list
    set redd-events-list fput (word "InSTREAM-7 redd events output file, Created " date-and-time) redd-events-list
  ]

  ; Create the redd summary output file.
  if a-file-name = "r-s-o-n" ; This is the value of the uninitialized file name
  [
    set redd-summary-outfile-name (word "ReddSummaryOut-" outfile-name-base)
    if file-exists? redd-summary-outfile-name [ file-delete redd-summary-outfile-name ]
    file-open redd-summary-outfile-name
    file-print (word "InSTREAM-7 redd summary output file, Created " date-and-time)
    file-print "BehavSp-Run,ReddID,Species,Reach,Cell,DateCreated,DateEmptied,InitialEggs,EggsDiedLowTemp,EggsDiedHiTemp,EggsDiedDewatering,EggsDiedScour,EggsDiedSuperimp,EggsEmerged"
    file-close
  ]

  ; Create the individual fish output file.
  if a-file-name = "f-o-n" ; This is the value of the uninitialized file name
  [
    set fish-outfile-name (word "IndFishOut-" outfile-name-base)
    if file-exists? fish-outfile-name [ file-delete fish-outfile-name ]
    file-open fish-outfile-name
    file-print (word "InSTREAM-7 individual fish output file, Created " date-and-time)
    file-print "Growth and survival reported here were calculated using length and weight from end of previous time step!"
    file-print "BehavSp-Run,End of time step,Light phase,Reach,FishID,Species,Age,Length,Weight,Condition,SuperindRep,Activity,Growth,Survival,Cell,Cell-X,Cell-Y,Temperature,Depth,Velocity,Light,TotalHidingPlaces,AvailHidingPlaces"
    file-close
  ]

  ; Create the individual cell output file.
  if a-file-name = "c-o-n" ; This is the value of the uninitialized file name
  [
    set cell-outfile-name (word "CellOut-" outfile-name-base)
    if file-exists? cell-outfile-name [ file-delete cell-outfile-name ]
    file-open cell-outfile-name
    file-print (word "InSTREAM-7 cell output file, Created " date-and-time)
    let output-list (list "BehavSp-Run" "End of time step" "Light phase" "Reach" "Flow" "CellID" "Area" "NumHidingPlaces" "DistToHide" "FracVelShelter" "FracSpawnGravel"
      "Temperature" "Depth" "Velocity" "Light")
    foreach age-class-list [ next-age ->
      ifelse next-age = last age-class-list
      [ set output-list (sentence output-list (word "NumAge" next-age "+fish")) ]
      [ set output-list (sentence output-list (word "NumAge" next-age "fish")) ]
    ] ; foreach age-class-list
    file-print csv:to-row output-list
    file-close
  ]

end

to save-event [an-event-type]
  ; A turtle (trout, redd, cell) procedure to save events for the event output files.
  ; an-event-type is a character string that says what event happened.
  ; Save the event only if events output is turned on.

  if events-output?
  [
    if breed = trout
    [
      set fish-events-list lput (csv:to-row (list
        behaviorspace-run-number
        formatted-sim-time
        light-phase
        [reach-name] of patches-reach
        trout-species
        who
        patches-cells-ID
        trout-age
        trout-length
        trout-weight
        trout-condition
        trout-superind-rep
        an-event-type
      )) fish-events-list
    ]

    if breed = redds
    [
      set redd-events-list lput (csv:to-row (sentence
        behaviorspace-run-number
        formatted-sim-time
        light-phase
        [reach-name] of patches-reach
        redd-species
        who
        patches-cells-ID
        an-event-type
      )) redd-events-list
    ]
  ]

end

to-report is-census?
  ; Utility reporter to identify census date-times for output analysis

  if time:is-before? sim-time census-start-time                              [ report false ]
  if prev-time = 0 [ report false ] ; This can happen at start of simulation when census-years-to-skip is zero
  if (time:get "dayofyear" sim-time) = (time:get "dayofyear" prev-time)      [ report false ]  ; census is first time step of day
  let this-day time:show sim-time "M/d"
  ifelse member? this-day census-days
  [ report true ]
  [ report false ]

end

to update-output
  ; Observer procedure to update graphics and write file output

  ; Display current time, light phase
  output-print (word formatted-sim-time " Phase: " light-phase)

  ; Determine if it is a census time; if so, force an output
  ; Create a local variable to avoid calling is-census? many times
  let is-this-a-census-date? is-census?

  ; Output files are created when first used, instead of in setup,
  ; to keep new output files from being created if setup is executed but go is not.
  ; This code also allows output files to be turned on and used after a simulation has started.

  ; Write file output if it is time for the next output
  ; Use "not is-before" to include is-equal or is-after
  if (not time:is-before? sim-time next-output-time) or is-this-a-census-date?
  [
    ; First, schedule next file output, if this output was not forced by the census date
    if not time:is-before? sim-time next-output-time
    [
      set next-output-time time:plus next-output-time file-output-frequency file-output-units
    ]

    ; Do the brief population output
    if brief-pop-output?
    [
      if brief-pop-outfile-name = "b-p-o-n" [ build-output-file-named brief-pop-outfile-name ]
      file-open brief-pop-outfile-name

      foreach reach-names [ next-reach-name ->
        let next-reach one-of reaches with [reach-name = next-reach-name]
        let the-flow [reach-flow] of next-reach
        let the-temperature [reach-temperature] of next-reach
        let the-turbidity [reach-turbidity] of next-reach

        foreach species-list [ next-species ->
          foreach age-class-list [ next-age ->
           let last-age? (next-age = last age-class-list)  ; last-age? is true if this is the last age on age-class-list
           let age-label (word "Age-" next-age)
           let the-fish nobody
           ifelse last-age?
           [
             set age-label (word age-label "+")
             set the-fish trout with [ (patches-reach = next-reach) and (trout-species = next-species) and (trout-age >= next-age) ]
           ]
           [ set the-fish trout with [ (patches-reach = next-reach) and (trout-species = next-species) and (trout-age = next-age) ]
           ]

           ; These statements are to make output work when there are no fish in the category
           ; (because "mean" raises an error if there are no values).
           let the-count 0.0
           let the-length 0.0
           let the-weight 0.0
           let the-condition 0.0
           let the-frac-drift 0.0
           let the-frac-search 0.0
           let the-frac-hide 0.0

            if any? the-fish
            [
              set the-count (sum [trout-superind-rep] of the-fish)
              set the-length (mean [trout-length] of the-fish)
              set the-weight (mean [trout-weight] of the-fish)
              set the-condition (mean [trout-condition] of the-fish)
              set the-frac-drift (count the-fish with [ trout-activity = "drift" ] / count the-fish)
              set the-frac-search (count the-fish with [ trout-activity = "search" ] / count the-fish)
              set the-frac-hide (count the-fish with [ trout-activity = "hide" ] / count the-fish)
            ]

              file-print csv:to-row (list behaviorspace-run-number formatted-sim-time is-this-a-census-date? light-phase next-reach-name the-flow the-temperature the-turbidity next-species age-label
                 the-count
                 the-length
                 the-weight
                 the-condition
                 the-frac-drift
                 the-frac-search
                 the-frac-hide
                ) ; End of output list
           ] ; foreach age-class
        ] ; foreach species-list
      ] ; foreach reach-names

     file-close
    ] ; if brief-pop-output?

    ; Do the detailed population output
    if detailed-pop-output?
    [
      if detailed-pop-outfile-name = "d-p-o-n" [ build-output-file-named detailed-pop-outfile-name ]
      file-open detailed-pop-outfile-name

      foreach reach-names [ next-reach-name ->
        let next-reach one-of reaches with [reach-name = next-reach-name]
        let the-flow [reach-flow] of next-reach
        let the-temperature [reach-temperature] of next-reach
        let the-turbidity [reach-turbidity] of next-reach

        foreach species-list [ next-species ->
          foreach age-class-list [ next-age ->
            let last-age? next-age = last age-class-list
            foreach (list "drift" "search" "hide") [ next-activity ->
              let age-label (word "Age-" next-age)
              let the-fish nobody
              ifelse last-age?
              [
                set age-label (word age-label "+")
                set the-fish trout with
                  [ (patches-reach = next-reach) and (trout-species = next-species) and (trout-age >= next-age) and (trout-activity = next-activity) ]
              ]
              [ set the-fish trout with
                  [ (patches-reach = next-reach) and (trout-species = next-species) and (trout-age = next-age) and (trout-activity = next-activity) ]
              ]

             ; These statements are to make output work when there are no fish in the category
             ; (because "mean" raises an error if there are no values).
             let the-count 0.0
             let the-length 0.0
             let the-weight 0.0
             let the-condition 0.0

             if any? the-fish
             [
               set the-count (sum [trout-superind-rep] of the-fish)
               set the-length (mean [trout-length] of the-fish)
               set the-weight (mean [trout-weight] of the-fish)
               set the-condition (mean [trout-condition] of the-fish)
             ]

             file-print csv:to-row (list behaviorspace-run-number formatted-sim-time is-this-a-census-date? light-phase next-reach-name the-flow the-temperature the-turbidity next-species age-label next-activity
                the-count
                the-length
                the-weight
                the-condition
               ) ; End of output list
            ] ; foreach activity
          ] ; foreach age-class
        ] ; foreach species
      ] ; foreach reach-names

     file-close
    ] ; if detailed-pop-output?

    ; Do the individual fish output
    ; This does not work until after the first time step
    ; because trout-output-list is not initialized.
    ; It cannot report on a trout on the time step is is born because its trout-output-list is also not yet initialized.
    if individual-fish-output?
    [
      ; Check to make sure the file was initialized
      if fish-outfile-name = "f-o-n" [ build-output-file-named fish-outfile-name ]
      file-open fish-outfile-name
      ; Some outputs are fish variables, some are patch variables, and some variables of fish's cell
      ; The output list is written in the grow procedure
      ; The output is easier to use if in order of fish length
      foreach sort-on [-1 * trout-length] trout
      [ next-trout -> ask next-trout
        [ if is-list? trout-output-list [file-print csv:to-row trout-output-list ]]
      ]
      file-close
    ]

    ; Do the individual cell output
    ; Note that dry cells are skipped, to reduce file size. This is easily changed.
    ; Check to make sure the file was opened
    if cell-output?
    [
      if cell-outfile-name = "c-o-n" [ build-output-file-named cell-outfile-name ]
      file-open cell-outfile-name
      foreach sort-on [cell-ID] (cells with [depth > 0.0])  ; Delete "with [depth > 0.0]" to include dry cells
      [ next-cell -> ask next-cell
        [
          let output-list (list behaviorspace-run-number formatted-sim-time light-phase ([reach-name] of patches-reach) ([reach-flow] of patches-reach) cell-ID cell-area cell-num-hiding-places cell-dist-escape
            cell-frac-vel-shelter cell-frac-spawn ([reach-temperature] of patches-reach) depth velocity light)
          foreach age-class-list [ next-age ->
            ifelse next-age = last age-class-list
            [ set output-list (sentence output-list (count trout with [(patches-cell = myself) and (trout-age >= next-age)]))]
            [ set output-list (sentence output-list (count trout with [(patches-cell = myself) and (trout-age = next-age)]))]
          ] ; foreach age-class
          file-print csv:to-row output-list
        ]
      ]
      file-close
    ]

  ]  ; if not time:is-before  ** End of file output regulated by file-output-frequency

  ; The following outputs are produced each tick
  ; Do the redd summary output
  if redd-output? and (not empty? redd-summary-list)
  [
    if redd-summary-outfile-name = "r-s-o-n" [ build-output-file-named redd-summary-outfile-name ]
    file-open redd-summary-outfile-name
    foreach redd-summary-list [ the-line -> file-print the-line ]  ; Write each line of output to the file
    file-close
    set redd-summary-list (list)                ; Clear the output list
  ] ; if redd-output?

  ; Update events output file
  if events-output?
  [
    if not empty? fish-events-list
    [
      if fish-events-outfile-name = "f-e-o-n" [ build-output-file-named fish-events-outfile-name ]
      file-open fish-events-outfile-name
      foreach fish-events-list [ the-event -> file-print the-event ]  ; Write each event to the file
      file-close
      set fish-events-list (list)                ; Clear the event list
    ]

    if not empty? redd-events-list
    [
      if redd-events-outfile-name = "r-e-o-n" [ build-output-file-named redd-events-outfile-name ]
      file-open redd-events-outfile-name
      foreach redd-events-list [ the-event -> file-print the-event ]  ; Write each event to the file
      file-close
      set redd-events-list (list)                ; Clear the event list
    ]
  ]

  ; Just to be safe & tidy
  file-close-all

  ; Update plots.
  if update-plots?
  [
     set-current-plot "Population"
     set-current-plot-pen "Age 0"
     plot sum [trout-superind-rep] of (trout with [trout-age = 0])
     set-current-plot-pen "Age 1"
     plot sum [trout-superind-rep] of (trout with [trout-age = 1])
     set-current-plot-pen "Age 2+"
     plot sum [trout-superind-rep] of (trout with [trout-age >= 2])
  ]

  ; Update BehaviorSpace outputs, if this is a BehaviorSpace experiment
  if behaviorspace-run-number > 0 [ update-behaviorspace-outputs ]

  ; Write the display to a file if wanted, to make movies of a simulation
  ; Replace "export-interface" with "export-view" to get just the World instead of whole interface
  if write-frames? [ export-interface (word "InSTREAM-frame-" (10000 + ticks) ".png") ]

end

to update-behaviorspace-outputs
  ; Observer procedure to update global variables used as outputs in BehaviorSpace
  ; All the variables in "set" statements here must be defined in the globals [] statement
  ; at the top of this Code tab.
  ; To change species names or age classes of these variables, you must modify/add variables
  ; both in the globals [] statement and here.

  ; This procedure is only updated if the model run is part of a BehaviorSpace experiment!
  ; Therefore, you cannot use these variables for other outputs such as plots or output files.

  let RT-age-0s trout with [trout-age = 0 and trout-species = "Rainbow"]      ; Age 0 rainbow trout
  let RT-age-1s trout with [trout-age = 1 and trout-species = "Rainbow"]      ; Age 1 trout rainbow trout
  let RT-age-2s trout with [trout-age >= 2 and trout-species = "Rainbow"]     ; Age 2 and older rainbow trout
  let BT-age-0s trout with [trout-age = 0 and trout-species = "Brown"]      ; Age 0 brown trout
  let BT-age-1s trout with [trout-age = 1 and trout-species = "Brown"]      ; Age 1 trout brown trout
  let BT-age-2s trout with [trout-age >= 2 and trout-species = "Brown"]     ; Age 2 and older brown trout
  let CT-age-0s trout with [trout-age = 0 and trout-species = "Cutthroat"]      ; Age 0 cutthroat trout
  let CT-age-1s trout with [trout-age = 1 and trout-species = "Cutthroat"]      ; Age 1 trout cutthroat trout
  let CT-age-2s trout with [trout-age >= 2 and trout-species = "Cutthroat"]     ; Age 2 and older cutthroat trout

  set RT-age-0-abund sum [trout-superind-rep] of RT-age-0s      ; Abundance of age 0 rainbow trout
  set RT-age-1-abund sum [trout-superind-rep] of RT-age-1s      ; Abundance of age 1 rainbow trout
  set RT-age-2+-abund sum [trout-superind-rep] of RT-age-2s     ; Abundance of rainbow trout age 2 and older
  set RT-age-0-length ifelse-value (any? RT-age-0s) [mean [trout-length] of RT-age-0s] [""]     ; Mean length of age 0 rainbow trout; blank if there are none
  set RT-age-1-length ifelse-value (any? RT-age-1s) [mean [trout-length] of RT-age-1s] [""]     ; Mean length of age 1 rainbow trout; blank if there are none
  set RT-age-2+-length ifelse-value (any? RT-age-2s) [mean [trout-length] of RT-age-2s] [""]    ; Mean length of age 2 and older rainbow trout; blank if there are none
  set BT-age-0-abund sum [trout-superind-rep] of BT-age-0s      ; Abundance of age 0 brown trout
  set BT-age-1-abund sum [trout-superind-rep] of BT-age-1s      ; Abundance of age 1 brown trout
  set BT-age-2+-abund sum [trout-superind-rep] of BT-age-2s     ; Abundance of brown trout age 2 and older
  set BT-age-0-length ifelse-value (any? BT-age-0s) [mean [trout-length] of BT-age-0s] [""]     ; Mean length of age 0 brown trout; blank if there are none
  set BT-age-1-length ifelse-value (any? BT-age-1s) [mean [trout-length] of BT-age-1s] [""]     ; Mean length of age 1 brown trout; blank if there are none
  set BT-age-2+-length ifelse-value (any? BT-age-2s) [mean [trout-length] of BT-age-2s] [""]    ; Mean length of age 2 and older brown trout; blank if there are none
  set CT-age-0-abund sum [trout-superind-rep] of CT-age-0s      ; Abundance of age 0 cutthroat trout
  set CT-age-1-abund sum [trout-superind-rep] of CT-age-1s      ; Abundance of age 1 cutthroat trout
  set CT-age-2+-abund sum [trout-superind-rep] of CT-age-2s     ; Abundance of cutthroat trout age 2 and older
  set CT-age-0-length ifelse-value (any? CT-age-0s) [mean [trout-length] of CT-age-0s] [""]     ; Mean length of age 0 cutthroat trout; blank if there are none
  set CT-age-1-length ifelse-value (any? CT-age-1s) [mean [trout-length] of CT-age-1s] [""]     ; Mean length of age 1 cutthroat trout; blank if there are none
  set CT-age-2+-length ifelse-value (any? CT-age-2s) [mean [trout-length] of CT-age-2s] [""]    ; Mean length of age 2 and older cutthroat trout; blank if there are none

end
@#$#@#$#@
GRAPHICS-WINDOW
499
16
1918
355
-1
-1
1.1
1
10
1
1
1
0
0
0
1
0
1282
0
299
1
1
1
ticks
30.0

BUTTON
17
31
119
64
NIL
setup
NIL
1
T
OBSERVER
NIL
NIL
NIL
NIL
1

BUTTON
15
74
78
107
NIL
go
T
1
T
OBSERVER
NIL
NIL
NIL
NIL
0

BUTTON
15
118
78
151
step
go
NIL
1
T
OBSERVER
NIL
NIL
NIL
NIL
0

SWITCH
12
232
166
265
brief-pop-output?
brief-pop-output?
1
1
-1000

SWITCH
10
271
138
304
events-output?
events-output?
1
1
-1000

PLOT
11
421
479
610
Population
Tick
Number of trout
0.0
0.0
0.0
0.0
true
true
"" ""
PENS
"Age 0" 1.0 0 -16777216 true "" ""
"Age 1" 1.0 0 -2674135 true "" ""
"Age 2+" 1.0 0 -13345367 true "" ""

SWITCH
181
352
313
385
update-plots?
update-plots?
0
1
-1000

SWITCH
9
311
157
344
detailed-pop-output?
detailed-pop-output?
1
1
-1000

SWITCH
9
352
137
385
redd-output?
redd-output?
1
1
-1000

INPUTBOX
193
233
311
293
random-number-seed
0.0
1
0
Number

BUTTON
338
344
484
377
NIL
show-GIS-properties
NIL
1
T
OBSERVER
NIL
NIL
NIL
NIL
1

BUTTON
14
166
103
199
NIL
shade-by-depth
NIL
1
T
OBSERVER
NIL
NIL
NIL
NIL
1

BUTTON
114
166
195
199
NIL
shade-by-vel
NIL
1
T
OBSERVER
NIL
NIL
NIL
NIL
1

BUTTON
207
167
299
200
NIL
shade-by-light
NIL
1
T
OBSERVER
NIL
NIL
NIL
NIL
1

BUTTON
305
167
408
200
NIL
stop-shading
NIL
1
T
OBSERVER
NIL
NIL
NIL
NIL
1

@#$#@#$#@
## InSTREAM-7.3

This NetLogo file implements version 7.3 of inSTREAM, the Individual-based Stream Trout Research and Environmental Assessment Model. 

The model, its software, and its use are described in the report: _InSTREAM 7 User Manual: Model Description, Software Guide, and Application Guide_ by S.F. Railsback, B.C. Harvey, and D. Ayllón. This report is currently in production as a US Forest Service General Technical Report from the Pacific Southwest Research Station. (Version 7.3 is described in releases of this report made in August, 2021; previous releases of the report documented previous versions of inSTREAM 7.)

InSTREAM 7 can be referenced via either or both of these citations:

> Railsback, S. F., B.C. Harvey, and D. Ayllón. In preparation. _InSTREAM 7 User Manual:  Model Description, Software Guide, and Application Guide_ General Technical Report,  Pacific Southwest Research Station, US Forest Service, Albany, California. Available at: https://ecomodel.humboldt.edu.
> 
> Railsback, S. F., D. Ayllón, and B. C. Harvey. 2021. InSTREAM 7: Instream flow assessment and management model for stream trout. _River Research and Applications_ 37:1294–1302.

This code is compatible with version 6.2 (and, presumably, later versions) of NetLogo, but not with versions before 6.2.

## Revision log

  * 9 Aug 2022: Add BehaviorSpace run number as column in main output files
  * 9 Aug 2022: Add BehaviorSpace run number to output file names, before date (in set-globals)
  * 9 Aug 2022: Modify read-hydraulics to check for mis-ordered or duplicate flows in input.
  * 11 Jan 2022: Allow more than one census date per year.
    * Change global variable census-day to census-days
    * Change parameter file to make census-days a list of day values
    * Revisions to procedures: set-up-time, is-census?
  * 11 Oct 2021: Fix bug in select-activity-and-habitat. The bug caused fish selecting search feeding to move to best *drift-feeding* cell.
  * 14 Sept 2021: Replace read-hydraulics and update-hydraulics-for to simplify them and better handle extreme low flows. Add new reporter interpolate-hydraulics.
    * Use separate depth and velocity lookup tables for each cell, as table objects. Multiple flows with zero depth/velocity are removed, and a new table point is added for the flow at which depth/vel reach zero or the depth at zero flow.
    * Add new debug file for cell hydraulics, separating it from the cell habitat variables debug file.
    * Revise test procedures for depths and velocities to produce more output at lower flows and for near-zero flows.

  * 9 Sept 2021: Revise read-hydraulics and update-hydraulics-for to better handle flows between zero and lowest in input table.
  * 9 Sept 2021: Revise update-logistics to set reach-pisciv-trout-density to zero if no wet cells.
  * 8 Sept 2021: Revise redd events output to report number of eggs as separate column
  * 27 Aug 2021: Update from v. 7.2 to 7.3
    * Corrected rainbow trout redd development parameter values;  
    * save-event only saves event if events output is on;
    * Implement v. 7.3's expected length function in fitness measure (proc. fitness-for), including new parameter trout-fitness-length.
    * Modify update-hydraulics-for so it extrapolates depths when flow is zero.

  * 30 July 2021: Add missing save-event for redd superimposition.
  * 25 June 2021: Modified check-for-missing-input. Add check for illegal column labels. Revision at "; Now check values until end time" so it does not cause a run-time error if the last input time is before sim-end-time.
  * 28 May 2021: Modified update-day-length to improve error statement for day length <0 or >1. Add checks on phase lengths to update-time-and-habitat to allow use in near-arctic latitudes (per User Manual sect. 9.7).
  * 27 May 2021: Modified shade-by procedures so they update shading immediately instead of at next change in flow/light.
  * 14 May 2021: Added procedure to test parameters for correct number of values.
  * May 2021: Version 7.2 modified from version 7.1. Key differences are: habitat selection radius is limited to a maximum (proc. select-activity-and-habitat), expected fitness is set to zero for cells where swimming speed would exceed trout's maximum sustainable speed (proc. fitness-for), high velocity mortality is removed, and the standard value of parameter trout-capture-R1 is changed from 1.7 to 1.3. A software-only change gives each trout a unique color shade (proc. initialize-trout-with).
  * January 2021: Version 7.1 forked from development code.


## Software Copyright

Copyright 2022 by Lang Railsback & Associates, Arcata, California USA
info@langrailsback.com

## Contact and Information

Information and support for inSTREAM is available from its web site at Humboldt State University:
https://ecomodel.humboldt.edu/instream-and-insalmo-overview

Enquiries about model use, training, and support should be sent to:
Steve Railsback: steven.railsback@humboldt.edu
Bret Harvey: bret.harvey@usda.gov
Daniel Ayllón: daniel.ayllon@bio.ucm.es

## License

InSTREAM-7 is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

InSTREAM-7 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received [a copy of the GNU General Public License
along with InSTREAM-7] (file:License.txt).  If not, see <https://www.gnu.org/licenses/>.
@#$#@#$#@
default
true
0
Polygon -7500403 true true 150 5 40 250 150 205 260 250

airplane
true
0
Polygon -7500403 true true 150 0 135 15 120 60 120 105 15 165 15 195 120 180 135 240 105 270 120 285 150 270 180 285 210 270 165 240 180 180 285 195 285 165 180 105 180 60 165 15

arrow
true
0
Polygon -7500403 true true 150 0 0 150 105 150 105 293 195 293 195 150 300 150

box
false
0
Polygon -7500403 true true 150 285 285 225 285 75 150 135
Polygon -7500403 true true 150 135 15 75 150 15 285 75
Polygon -7500403 true true 15 75 15 225 150 285 150 135
Line -16777216 false 150 285 150 135
Line -16777216 false 150 135 15 75
Line -16777216 false 150 135 285 75

bug
true
0
Circle -7500403 true true 96 182 108
Circle -7500403 true true 110 127 80
Circle -7500403 true true 110 75 80
Line -7500403 true 150 100 80 30
Line -7500403 true 150 100 220 30

butterfly
true
0
Polygon -7500403 true true 150 165 209 199 225 225 225 255 195 270 165 255 150 240
Polygon -7500403 true true 150 165 89 198 75 225 75 255 105 270 135 255 150 240
Polygon -7500403 true true 139 148 100 105 55 90 25 90 10 105 10 135 25 180 40 195 85 194 139 163
Polygon -7500403 true true 162 150 200 105 245 90 275 90 290 105 290 135 275 180 260 195 215 195 162 165
Polygon -16777216 true false 150 255 135 225 120 150 135 120 150 105 165 120 180 150 165 225
Circle -16777216 true false 135 90 30
Line -16777216 false 150 105 195 60
Line -16777216 false 150 105 105 60

car
false
0
Polygon -7500403 true true 300 180 279 164 261 144 240 135 226 132 213 106 203 84 185 63 159 50 135 50 75 60 0 150 0 165 0 225 300 225 300 180
Circle -16777216 true false 180 180 90
Circle -16777216 true false 30 180 90
Polygon -16777216 true false 162 80 132 78 134 135 209 135 194 105 189 96 180 89
Circle -7500403 true true 47 195 58
Circle -7500403 true true 195 195 58

circle
false
0
Circle -7500403 true true 0 0 300

circle 2
false
0
Circle -7500403 true true 0 0 300
Circle -16777216 true false 30 30 240

cow
false
0
Polygon -7500403 true true 200 193 197 249 179 249 177 196 166 187 140 189 93 191 78 179 72 211 49 209 48 181 37 149 25 120 25 89 45 72 103 84 179 75 198 76 252 64 272 81 293 103 285 121 255 121 242 118 224 167
Polygon -7500403 true true 73 210 86 251 62 249 48 208
Polygon -7500403 true true 25 114 16 195 9 204 23 213 25 200 39 123

cylinder
false
0
Circle -7500403 true true 0 0 300

dot
false
0
Circle -7500403 true true 90 90 120

face happy
false
0
Circle -7500403 true true 8 8 285
Circle -16777216 true false 60 75 60
Circle -16777216 true false 180 75 60
Polygon -16777216 true false 150 255 90 239 62 213 47 191 67 179 90 203 109 218 150 225 192 218 210 203 227 181 251 194 236 217 212 240

face neutral
false
0
Circle -7500403 true true 8 7 285
Circle -16777216 true false 60 75 60
Circle -16777216 true false 180 75 60
Rectangle -16777216 true false 60 195 240 225

face sad
false
0
Circle -7500403 true true 8 8 285
Circle -16777216 true false 60 75 60
Circle -16777216 true false 180 75 60
Polygon -16777216 true false 150 168 90 184 62 210 47 232 67 244 90 220 109 205 150 198 192 205 210 220 227 242 251 229 236 206 212 183

fish
false
0
Polygon -1 true false 44 131 21 87 15 86 0 120 15 150 0 180 13 214 20 212 45 166
Polygon -1 true false 135 195 119 235 95 218 76 210 46 204 60 165
Polygon -1 true false 75 45 83 77 71 103 86 114 166 78 135 60
Polygon -7500403 true true 30 136 151 77 226 81 280 119 292 146 292 160 287 170 270 195 195 210 151 212 30 166
Circle -16777216 true false 215 106 30

flag
false
0
Rectangle -7500403 true true 60 15 75 300
Polygon -7500403 true true 90 150 270 90 90 30
Line -7500403 true 75 135 90 135
Line -7500403 true 75 45 90 45

flower
false
0
Polygon -10899396 true false 135 120 165 165 180 210 180 240 150 300 165 300 195 240 195 195 165 135
Circle -7500403 true true 85 132 38
Circle -7500403 true true 130 147 38
Circle -7500403 true true 192 85 38
Circle -7500403 true true 85 40 38
Circle -7500403 true true 177 40 38
Circle -7500403 true true 177 132 38
Circle -7500403 true true 70 85 38
Circle -7500403 true true 130 25 38
Circle -7500403 true true 96 51 108
Circle -16777216 true false 113 68 74
Polygon -10899396 true false 189 233 219 188 249 173 279 188 234 218
Polygon -10899396 true false 180 255 150 210 105 210 75 240 135 240

house
false
0
Rectangle -7500403 true true 45 120 255 285
Rectangle -16777216 true false 120 210 180 285
Polygon -7500403 true true 15 120 150 15 285 120
Line -16777216 false 30 120 270 120

leaf
false
0
Polygon -7500403 true true 150 210 135 195 120 210 60 210 30 195 60 180 60 165 15 135 30 120 15 105 40 104 45 90 60 90 90 105 105 120 120 120 105 60 120 60 135 30 150 15 165 30 180 60 195 60 180 120 195 120 210 105 240 90 255 90 263 104 285 105 270 120 285 135 240 165 240 180 270 195 240 210 180 210 165 195
Polygon -7500403 true true 135 195 135 240 120 255 105 255 105 285 135 285 165 240 165 195

line
true
0
Line -7500403 true 150 0 150 300

line half
true
0
Line -7500403 true 150 0 150 150

pentagon
false
0
Polygon -7500403 true true 150 15 15 120 60 285 240 285 285 120

person
false
0
Circle -7500403 true true 110 5 80
Polygon -7500403 true true 105 90 120 195 90 285 105 300 135 300 150 225 165 300 195 300 210 285 180 195 195 90
Rectangle -7500403 true true 127 79 172 94
Polygon -7500403 true true 195 90 240 150 225 180 165 105
Polygon -7500403 true true 105 90 60 150 75 180 135 105

plant
false
0
Rectangle -7500403 true true 135 90 165 300
Polygon -7500403 true true 135 255 90 210 45 195 75 255 135 285
Polygon -7500403 true true 165 255 210 210 255 195 225 255 165 285
Polygon -7500403 true true 135 180 90 135 45 120 75 180 135 210
Polygon -7500403 true true 165 180 165 210 225 180 255 120 210 135
Polygon -7500403 true true 135 105 90 60 45 45 75 105 135 135
Polygon -7500403 true true 165 105 165 135 225 105 255 45 210 60
Polygon -7500403 true true 135 90 120 45 150 15 180 45 165 90

sheep
false
15
Circle -1 true true 203 65 88
Circle -1 true true 70 65 162
Circle -1 true true 150 105 120
Polygon -7500403 true false 218 120 240 165 255 165 278 120
Circle -7500403 true false 214 72 67
Rectangle -1 true true 164 223 179 298
Polygon -1 true true 45 285 30 285 30 240 15 195 45 210
Circle -1 true true 3 83 150
Rectangle -1 true true 65 221 80 296
Polygon -1 true true 195 285 210 285 210 240 240 210 195 210
Polygon -7500403 true false 276 85 285 105 302 99 294 83
Polygon -7500403 true false 219 85 210 105 193 99 201 83

square
false
0
Rectangle -7500403 true true 30 30 270 270

square 2
false
0
Rectangle -7500403 true true 30 30 270 270
Rectangle -16777216 true false 60 60 240 240

star
false
0
Polygon -7500403 true true 151 1 185 108 298 108 207 175 242 282 151 216 59 282 94 175 3 108 116 108

target
false
0
Circle -7500403 true true 0 0 300
Circle -16777216 true false 30 30 240
Circle -7500403 true true 60 60 180
Circle -16777216 true false 90 90 120
Circle -7500403 true true 120 120 60

tree
false
0
Circle -7500403 true true 118 3 94
Rectangle -6459832 true false 120 195 180 300
Circle -7500403 true true 65 21 108
Circle -7500403 true true 116 41 127
Circle -7500403 true true 45 90 120
Circle -7500403 true true 104 74 152

triangle
false
0
Polygon -7500403 true true 150 30 15 255 285 255

triangle 2
false
0
Polygon -7500403 true true 150 30 15 255 285 255
Polygon -16777216 true false 151 99 225 223 75 224

truck
false
0
Rectangle -7500403 true true 4 45 195 187
Polygon -7500403 true true 296 193 296 150 259 134 244 104 208 104 207 194
Rectangle -1 true false 195 60 195 105
Polygon -16777216 true false 238 112 252 141 219 141 218 112
Circle -16777216 true false 234 174 42
Rectangle -7500403 true true 181 185 214 194
Circle -16777216 true false 144 174 42
Circle -16777216 true false 24 174 42
Circle -7500403 false true 24 174 42
Circle -7500403 false true 144 174 42
Circle -7500403 false true 234 174 42

turtle
true
0
Polygon -10899396 true false 215 204 240 233 246 254 228 266 215 252 193 210
Polygon -10899396 true false 195 90 225 75 245 75 260 89 269 108 261 124 240 105 225 105 210 105
Polygon -10899396 true false 105 90 75 75 55 75 40 89 31 108 39 124 60 105 75 105 90 105
Polygon -10899396 true false 132 85 134 64 107 51 108 17 150 2 192 18 192 52 169 65 172 87
Polygon -10899396 true false 85 204 60 233 54 254 72 266 85 252 107 210
Polygon -7500403 true true 119 75 179 75 209 101 224 135 220 225 175 261 128 261 81 224 74 135 88 99

wheel
false
0
Circle -7500403 true true 3 3 294
Circle -16777216 true false 30 30 240
Line -7500403 true 150 285 150 15
Line -7500403 true 15 150 285 150
Circle -7500403 true true 120 120 60
Line -7500403 true 216 40 79 269
Line -7500403 true 40 84 269 221
Line -7500403 true 40 216 269 79
Line -7500403 true 84 40 221 269

wolf
false
0
Polygon -16777216 true false 253 133 245 131 245 133
Polygon -7500403 true true 2 194 13 197 30 191 38 193 38 205 20 226 20 257 27 265 38 266 40 260 31 253 31 230 60 206 68 198 75 209 66 228 65 243 82 261 84 268 100 267 103 261 77 239 79 231 100 207 98 196 119 201 143 202 160 195 166 210 172 213 173 238 167 251 160 248 154 265 169 264 178 247 186 240 198 260 200 271 217 271 219 262 207 258 195 230 192 198 210 184 227 164 242 144 259 145 284 151 277 141 293 140 299 134 297 127 273 119 270 105
Polygon -7500403 true true -1 195 14 180 36 166 40 153 53 140 82 131 134 133 159 126 188 115 227 108 236 102 238 98 268 86 269 92 281 87 269 103 269 113

x
false
0
Polygon -7500403 true true 270 75 225 30 30 225 75 270
Polygon -7500403 true true 30 75 75 30 270 225 225 270
@#$#@#$#@
NetLogo 6.2.2
@#$#@#$#@
@#$#@#$#@
@#$#@#$#@
<experiments>
  <experiment name="experiment" repetitions="5" runMetricsEveryStep="true">
    <setup>setup</setup>
    <go>go</go>
    <metric>formatted-sim-time</metric>
    <metric>light-phase</metric>
    <metric>is-census?</metric>
    <metric>RT-age-0-abund</metric>
    <metric>RT-age-1-abund</metric>
    <metric>RT-age-2+-abund</metric>
    <metric>RT-age-0-length</metric>
    <metric>RT-age-1-length</metric>
    <metric>RT-age-2+-length</metric>
    <metric>BT-age-0-abund</metric>
    <metric>BT-age-1-abund</metric>
    <metric>BT-age-2+-abund</metric>
    <metric>BT-age-0-length</metric>
    <metric>BT-age-1-length</metric>
    <metric>BT-age-2+-length</metric>
    <metric>CT-age-0-abund</metric>
    <metric>CT-age-1-abund</metric>
    <metric>CT-age-2+-abund</metric>
    <metric>CT-age-0-length</metric>
    <metric>CT-age-1-length</metric>
    <metric>CT-age-2+-length</metric>
    <enumeratedValueSet variable="brief-pop-output?">
      <value value="false"/>
    </enumeratedValueSet>
    <enumeratedValueSet variable="detailed-pop-output?">
      <value value="false"/>
    </enumeratedValueSet>
    <enumeratedValueSet variable="shade-variable">
      <value value="&quot;off&quot;"/>
    </enumeratedValueSet>
    <enumeratedValueSet variable="update-plots?">
      <value value="false"/>
    </enumeratedValueSet>
    <enumeratedValueSet variable="events-output?">
      <value value="false"/>
    </enumeratedValueSet>
  </experiment>
</experiments>
@#$#@#$#@
@#$#@#$#@
default
0.0
-0.2 0 0.0 1.0
0.0 1 1.0 0.0
0.2 0 0.0 1.0
link direction
true
0
Line -7500403 true 150 150 90 180
Line -7500403 true 150 150 210 180
@#$#@#$#@
0
@#$#@#$#@
